/* boot.c */

/* Gujin is a bootloader, it loads a Linux kernel from cold boot or DOS.
 * Copyright (C) 1999-2001 Etienne Lorrain, fingerprint (2D3AF3EA):
 *   2471 DF64 9DEE 41D4 C8DB 9667 E448 FF8C 2D3A F3EA
 * E-Mail: etienne.lorrain@masroudeau.com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

/*
 * Take care of 'library.h' here, the start of this file shall
 * not depend on configuration, and we do not want to use
 * the 'movs' GAS macroes in library.h because of size
 * restrictions, and because this bootsector shall work
 * up to displaying a message on a 8086 and 80286.
 * So the definition of BOOT_C.
 * Can someone test the 8086 and 80286 behavior please?
 */
#define BOOT_C
#include "make.h"
#include "library.h"	/* for farptr and UPDATE_BOOTPARAM */
#include "boot.h"
#include "debug.h"	/* for applications calling boot2 interface */

/*
 * The (false) compilation switch "DOSPATCH" is there to show
 * what to patch to have a DOS bootloader (boot.com), it should not
 * be defined (I do not know how to remove optimisation
 * "byte constant coded on 1 byte instead of 2"
 */

/**
 ** This is the beginning of the boot sector
 **/

asm (
".text								\n"
"								\n"
".org 0		# start of MSDOS reserved area			\n"
#ifndef DOSPATCH  /* NO cli for floppy disks, crashes my BIOS ! */
"dospatch1adr = .						\n"
"	jmp	start	# a *relative* jump !			\n"
"	nop							\n"
#else
"	cli							\n"
"	jmp	start	# a *relative* jump !			\n"
#endif
	);

/* initialised by external software (see instboot.c) : */
/* "const" to have the good section attribute for .text */
static const bootbefore_type bootbefore
	 __attribute__ ((section (".text"), unused)) = {};

asm (
".text									\n"
".org 0x3E	# end of MSDOS reserved area, target of the jmp		\n"
"start:									\n"
"	# Do not know what is CS and IP, usual hardware address 0007C3E,\n"
"	# i.e. 07C0:003E or 0000:7C3E for cold boot; but if we are	\n"
"	# building a DOS bootloader we cannot use absolute values...	\n"
"	# The stack segment will be defined once for the whole loader	\n"
"	# here to be able to save everything, the stack offset will be	\n"
"	# set to zero to get a maximum working space.			\n"
"	# N.B.  push: predecrement and write, pop: read and postincr	\n"
"									\n"
"	cli								\n"
"#	mov	%ss,%dx	# if you are curious... remove serial line	\n"
"#	mov	%sp,%ax	# initialisation and uncomment those.		\n"
"	mov	%cs,%sp							\n"
"									\n"
#ifndef DOSPATCH
"									\n"
"dospatch2adr = . + 2							\n"
"	addw	$0x840+deltaseg,%sp	# far away, even if %cs is null	\n"
"									\n"
#else
"									\n"
"	# do not really move the software (debugger):			\n"
"	# HERE (0x10+deltaseg) has to be a TWO BYTES constant,		\n"
"	# even if (deltaseg == 0)!					\n"
"#	addw	$0x10+deltaseg,%sp	# ONLY for boot.com files	\n"
"	addw	$deltaseg,%sp		# ONLY for boot.exe files	\n"
"									\n"
#endif
"									\n"
"	mov	%sp,%ss							\n"
"	xor	%sp,%sp							\n"
"	sti								\n"
"									\n"
"	# save everything in 16 bit, top of stack with 'iretw' format	\n"
"									\n"
"	pushfw								\n"
"	pushw	%cs	# 'pushw %cs; callw puship' = 'lcallw puship'	\n"
"	callw	puship	# push	%ip					\n"
"puship:								\n"
"	pushw	%ds							\n"
"	pushw	%es							\n"
"	pushaw	   # push ax, cx, dx, bx, initsp, bp, di, si (16 bytes)	\n"
"									\n"
"	# normalise %cs and setup %ds,%es				\n"
"	mov	%sp,%bp	# %sp cannot be indexed (%bp used with %ss)	\n"
"	subw	$puship,20(%bp)	# if not modified, 'iretw' will restart all.\n"
"	mov	20(%bp),%ax	# 20(bp) is the pushed %ip (callw nextip)\n"
"	shr	$4,%ax		# segment/offset for 80x86 real mode	\n"
"	add	22(%bp),%ax	# the pushed %cs			\n"
"	mov	%ax,%ds							\n"
"	mov	%ax,%es		# just for 'read_disk'			\n"
"	push	%ax							\n"
"	pushw	$normalise_jump						\n"
"	lretw								\n"
"	inc	%bp							\n"
"	dec	%sp							\n"
"normalise_jump:							\n"
"	# What is the output system (serial/screen) ?			\n"
"	movw	$serialconf,%si						\n"
"	lodsw	%ds:(%si),%ax		# %ah should be zero (int $0x14)\n"
"	cmp	$0,%al			# config reg cannot be zero if serial\n"
"	je	noserialbios						\n"
"	movw	(%si),%dx						\n"
"	mov	%dx,%bp			# save it, %bp used only if serial\n"
"	int	$0x14			# init serial line		\n"
"noserialbios:								\n"
"	# Here, if %bp hold a stack pointer (>0xFF00), the local	\n"
"	# screen is used, else %bp holds the serial port number.	\n"
"									\n"
"	cld				# needed for lods[bw] (postincr)\n"
"	movw    $simple_puts,%di	# to save few bytes, %di constant\n"
"									\n"
"	movw	$crlf_msg,%si		# parameter in %ds:%si		\n"
"	callw	*%di			# callw	simple_puts		\n"
"	movw	$boot1_msg,%si						\n"
"	callw	*%di			# callw	simple_puts		\n"
"									\n"
"	xor	%cx,%cx							\n"
"	movw	$start,%si		# i.e. 0x3E			\n"
"boot1_checksum_loop:							\n"
"	lodsw	%ds:(%si),%ax						\n"
"	add	%ax,%cx							\n"
"	cmpw	$bootend,%si						\n"
"	jl	boot1_checksum_loop					\n"
"	jcxz	boot1_checksum_ok					\n"
"	movw	$checksum_msg,%si					\n"
"	callw	*%di			# callw	simple_puts		\n"
"	callw	*%di		# print the string declared just after	\n"
"boot1_checksum_ok:							\n"
"					# %es already setup		\n"
"	movw	$loadnextsector,%bx	# load just after boot1		\n"
"	pushw	$bootloader2						\n"
"	movw	$comma_msg,%si						\n"
"	callw	*%di			# callw	simple_puts		\n"
"	mov	$boot2_msg,%si						\n"
"	jmp	loop_loader_entry					\n"
"									\n"
"loop_loader:								\n"
"	pushw	(%si)			# .next				\n"
"	lodsw	%ds:(%si),%ax		# same as 'add $2,%si'		\n"
"	lodsw	%ds:(%si),%ax						\n"
"	push	%ax			# .checksum			\n"
"	lodsw	%ds:(%si),%ax						\n"
"	push	%ax			# .nbword			\n"
"	push	%di			# save it (*di = simple_puts)	\n"
"									\n"
#ifndef DOSPATCH
"									\n"
"dospatch3adr = . + 1							\n"
"	callw	read_disk						\n"
"									\n"
#else
"									\n"
"	callw	read_disk-1	# which is a ret (carry cleared here)	\n"
"									\n"
#endif
"									\n"
"	pop	%di							\n"
"	pop	%cx							\n"
"	pop	%dx							\n"
"	jc	error_disk		# I: %si, %es:%bx /O: %es:%bx	\n"
"	mov	%bx,%si							\n"
"boot2_checksum_loop:							\n"
"	lodsw	%es:(%si),%ax						\n"
"	add	%ax,%dx							\n"
"	loop	boot2_checksum_loop	# this do not modify flags	\n"
"	mov	%si,%bx			# this do not modify flags	\n"
"	jne	error_checksum						\n"
"	testw	%bx,%bx							\n"
"	jnz	noincseg						\n"
"	mov	%es,%dx							\n"
"	addb	$0x10,%dh						\n"
"	mov	%dx,%es							\n"
"noincseg:								\n"
"	movw	$dot_msg,%si						\n"
"loop_loader_entry:							\n"
"	callw	*%di			# callw	simple_puts		\n"
"	pop	%si							\n"
"	testw	%si,%si			# clears carry			\n"
"	jnz	loop_loader						\n"
"									\n"
"	movw	$go_msg,%si						\n"
"	callw	*%di			# callw	simple_puts		\n"
"	callw	bootloader2_entrypoint	# adjust return address b4 the ret!\n"
"									\n"
"retoffsetboot2 = retry_all-error_checksum				\n"
"retoffsetboot2_error = retry_all-error_disk				\n"
"									\n"
"error_checksum:							\n"
"	movw	$checksum_msg,%si					\n"
"	callw	*%di			# callw	simple_puts		\n"
"error_disk:								\n"
"	movw	$error_msg,%si						\n"
"	callw	*%di			# callw	simple_puts		\n"
"	pop	%si			# realign stack			\n"
"retry_all:								\n"
"									\n"
"									\n"
#ifndef DOSPATCH
"									\n"
"	# 4 bytes, that is 0xCF1F0761					\n"
"dospatch4adr = .							\n"
"	popaw	# take care of popa bug on i386				\n"
"	popw	%es							\n"
"	popw	%ds							\n"
"	iretw								\n"
"									\n"
#else
"									\n"
"	# 4 bytes, that is 0x21CD4CB4					\n"
"	mov	$0x4C,%ah						\n"
"	int	$0x21							\n"
"									\n"
#endif
"									\n"
"									\n"
"# register %di is clearly modified by the BIOS of 3DLabs Permedia 2,	\n"
"# VIPER II AGP Rev-2.1.0 product P2-85 in VESA modes... (hwinfo.exe)	\n"
"# Result is the same as Trident/SVGA video mode/%ds : crash when scroll\n"
"#									\n"
"# The video BIOSes are just crap these days, execute a 'pusha', but	\n"
"# after having realigned the stack which is currently only on two	\n"
"# bytes boundary.							\n"
"simple_puts:		# parameter: zero ended string at %ds:%si	\n"
"	lodsb	%ds:(%si),%al						\n"
"	cmp	$0,%al		# instruction coded on 2 bytes		\n"
"	je	simple_puts_end						\n"
"	pushw	%ds							\n"
"	pushaw			# %di, %bp and %bx would be a minimum	\n"
"	mov	%bp,%dx		# %bp has been set before, serial or screen\n"
"	testb	%dh,%dh		# %bp was a stack pointer ?		\n"
"	jnz	screen_output						\n"
"	mov	$0x01,%ah						\n"
"	int	$0x14							\n"
"	jmp	was_serial_output					\n"
"screen_output:								\n"
"	mov	$0x07,%bx	# display page 0, color 7 if graphic mode\n"
"	.GLOBAL color_overwrite	# used by inlined function elsewhere	\n"
"color_overwrite = . - 2;						\n"
"	mov	$0x0E,%ah						\n"
"	int	$0x10		# int 0x10 may destroy %bp if scroll...	\n"
"was_serial_output:							\n"
"	popaw	# take care of popa bug on i386				\n"
"	popw	%ds							\n"
"	jmp	simple_puts						\n"
"simple_puts_end:							\n"
"	retw								\n"
"									\n"
"# Parameters: %es:%bx points to buffer, %ds:%si points to the union	\n"
"read_disk:			# BIOS INT13 %ah=0x02			\n"
"	lodsw	%ds:(%si),%ax	# .reg_dx / .ide_dcr_address		\n"
"	mov	%ax,%dx							\n"
"	lodsw	%ds:(%si),%ax	# .reg_ax				\n"
"	mov	(%si),%cx	# .reg_cx / .nb_word_to_read (unused 13/42)\n"
"	cmp	$0xF0,%ah						\n"
"	je	ide_hard						\n"
"	#cmp	$0x42,%ah	We do not really care to modify memory	\n"
"	#jne	proceed_int13	at 6(%si) and 4(%si) for int1302...	\n"
"	#								\n"
"	#  hope this INT13 %ah=42 keeps %es:%bx				\n"
"	#  here  %si==&(bootloader2.int1342.cst_16)			\n"
"	mov	%es,6(%si)	# .segment	(0x8C,0x44,0x06)	\n"
"	mov	%bx,4(%si)	# .offset	(0x89,0x5C,0x04)	\n"
"#proceed_int13:							\n"
"	int	$0x13							\n"
"	jnc	int13_ok						\n"
"init_bios_disk:							\n"
"	xor	%ax,%ax							\n"
"	int	$0x13		# reset disks (known by BIOS)		\n"
"	stc			# shall remember the error...		\n"
"int13_ok:								\n"
"	retw								\n"
"									\n"
"#*****************************************************************	\n"
"#* What you were looking for - REMEMBER IT IS PROTECTED BY GPL - *	\n"
"#* see top of this file. Copyrighted by Etienne LORRAIN, 1999.   *	\n"
"#*                                                               *	\n"
"#*   YOU ARE NOT AUTORISED TO READ THAT CODE BEFORE HAVING READ  *	\n"
"#*    AND UNDERSTOOD THE GPL, i.e. the General Public License.   *	\n"
"#* See http://www.gnu.org/copyleft/gpl.html to read the GPL or   *	\n"
"#*              its translation in your language.                *	\n"
"#*****************************************************************	\n"
"# Reading IDE status:							\n"
"# b0: general error, b1: 0, b2: data successfully corrected,		\n"
"# b3: data request, b4: disk ready, b5: write fault, b6: IDE ready,	\n"
"# b7: IDE busy (no bit valid)						\n"
"# More info on Linux file hd.c						\n"
"ide_hard:			# access IDE directly			\n"
"	mov	%bx,%di		# %bx should not be modified in this fct\n"
"	outb	%al,%dx		# %al = 0x0A, %dx = 0x3F6 (for instance)\n"
"	lodsw	%ds:(%si),%ax	# .base_ide_plus7 (that is already in	\n"
"				# cx, but it also increment %si and	\n"
"				# initialise %ah=1 !)			\n"
"	mov	%ax,%dx		# for instance 0x1F7 (cmd/status register)\n"
"	xor	%cx,%cx		# 65536*1us (approx) = 65 ms		\n"
"waitnotbusy:								\n"
"	inb	%dx,%al							\n"
"	test	$0x80,%al	# busy					\n"
"	loopnz	waitnotbusy						\n"
"	jcxz	ide_error	# timeout: %ah=MSB(ideaddr), often 1	\n"
"	dec	%dx							\n"
"	outsb	%ds:(%si),%dx	# .lba_head				\n"
"	inc	%dx							\n"
"waitdiskready:								\n"
"	dec	%cx							\n"
"	jcxz	ide_error	# timeout BUSY or !DRDY or !RDY: %ah = 1\n"
"	inb	%dx,%al							\n"
"	test	$0xD0,%al	# clear carry				\n"
"	jle	waitdiskready	# Don't bother, I understood that for you!\n"
"	jnp	waitdiskready						\n"
"	push	%dx		# %dx points to status/cmd register	\n"
"	dec	%dx							\n"
"	dec	%dx							\n"
"	outsb	%ds:(%si),%dx	# .cylinder_high			\n"
"	dec	%dx							\n"
"	outsb	%ds:(%si),%dx	# .cylinder_low				\n"
"	dec	%dx							\n"
"	outsb	%ds:(%si),%dx	# .sector				\n"
"	dec	%dx							\n"
"	outsb	%ds:(%si),%dx	# .nb_sect				\n"
"	pop	%dx							\n"
"	outsb	%ds:(%si),%dx	# .ide_command				\n"
"	lodsw	%ds:(%si),%ax	# .nb_read_sect, %ah = 0 (but if 256 sectors)\n"
"	mov	%ax,%cx							\n"
"loopnextsector:							\n"
"	dec	%ah		# debug only				\n"
"	push	%cx							\n"
"	xor	%cx,%cx							\n"
"waitcommandcomplete:							\n"
"	inb	%dx,%al							\n"
"	test	$0x81,%al	# clear carry				\n"
"	loopnz	waitcommandcomplete					\n"
"	jcxz	ide_error_popcx	# timeout/error/no DRQ: %ah = -nbsectortoread\n"
"	test	$0x08,%al						\n"
"	jz	waitcommandcomplete					\n"
"	mov	(%si),%cx						\n"
"	push	%dx							\n"
"	sub	$7,%dx		# point to data register and clear carry\n"
"	rep insw %dx,%es:(%di)						\n"
"	pop	%dx							\n"
"	pop	%cx							\n"
"	loop	loopnextsector						\n"
"	retw								\n"
"ide_error_popcx:							\n"
"	pop	%cx							\n"
"ide_error:								\n"
"	stc								\n"
"	retw								\n"
	);

static char crlf_msg[]
	__attribute__ ((section (".text"), aligned(1), unused)) = "\r\n";

static char comma_msg[]
	__attribute__ ((section (".text"), aligned(1), unused)) = ", ";

static char checksum_msg[]
	__attribute__ ((section (".text"), aligned(1), unused)) = ": checksum";
	 /* position dependancy with next msg */

static char error_msg[]
	__attribute__ ((section (".text"), aligned(1), unused)) = " ERROR!";
	 /* position dependancy with previous msg */


/*
 * If you are using a debugger to get the checksum, take care:
 * it calculates the checksum with the soft breakpoint instruction added!
 * The checksum is set so as to have sum of words between (included)
 * 0x03E and (included) 0x1B6 equals zero, and is updated by 'instboot.c'
 * to take care of 'deltaseg' which is a 32 bit number known at link time.
 */
asm (" .org 0x196 ");
/* initialised by external software: */
static const boot1param_type boot1param
	 __attribute__ ((section (".text"), unused)) = {
	 boot1_checksum: 0x7849,
	 serialconf: { port: 0xFF00 }
	 };

asm (
".org 0x1B6	# Max end is 0x1B6 for some kind of password ?		\n"
"									\n"
"bootend:								\n"
".org 0x1B8	# Max end is 0x1B8 for Windows NT marker + partition tables\n"
	);

static bootafter_type bootafter __attribute__ ((section (".text"), unused))
	= { Signature0xAA55: 0xAA55};

/**
 ** This is the beginning of the second level bootloader:
 **/

asm (
".text									\n"
".org 0x200	# no more, that is the end of boot1 / start of boot2.	\n"
"loadnextsector:							\n"
"instboot_info_ptr:	# address hardcoded in instboot.c		\n"
"	.long	instboot_info						\n"
"bootloader2_entrypoint:						\n"
"	# readjust the return address to bootloader1			\n"
"	popw	%ax							\n"
"	addw	$retoffsetboot2,%ax					\n"
"	mov	%ax,return_address_bootloader1				\n"
"	# some output to know boot2 has been reached			\n"
"	movw	$runloader2_msg,%si					\n"
"	callw	*%di		# callw	simple_puts			\n"
"	jmp	bootloader2_continue					\n"
"									\n"
"return_address_bootloader1:						\n"
"	.long	0							\n"
"runloader2_msg:							\n"
"	.byte	' ', 'v', '0', '.', '5', ' '		\n" /*VERSION*/
"runloader2_endmsg:							\n"
"	sub	%al,0x29(%bp,%di)					\n"
"	and	%al,0x74(%di)						\n"
"	imul	$0x656E,0x6E(%di),%sp					\n"
"	and	%cl,0x4F(%si)						\n"
"	push	%dx							\n"
"	push	%dx							\n"
"	inc	%cx							\n"
"	dec	%cx							\n"
"	dec	%si							\n"
"	and	%dh,(%bp,%si)						\n"
"	xor	%dh,(%bx,%si)						\n"
"	xor	%bp,(0x0A0D)						\n"
"	add	%al,(%bx,%si)						\n"
	);

/*
 * The variable up to 'gujin_param' could be modified by the boot level 2,
 * to save default state when outside of any operating system,
 * because it is in THE sector described in boot1 (easy access).
 * Its size should be a multiple of 2.
 */
struct gujin_param_str gujin_param	/* aligned for checksum calculus */
	__attribute__ ((section (".text"), aligned (2))) = {
    magic:   16980327,
    version: 0x0005,	/*VERSION*/ /* to abort ./instboot on a wrong version */
    dot_msg: ".",	/* 4 bytes */
    go_msg:  " go!",	/* 8 bytes, a space before */
    /* The following can be moved without changing checksum of BOOT1: */
    default_video_mode: 0xFFFF,
    default_text_video_mode: 0xFFFF,
    default_graphic_video_mode: 0xFFFF,
    attrib: {
	/* Can be changed and saved at execution: */
#define Apply(field, deflt, grey)	field : deflt,
	APPLY_ATTRIBUTE_ORDER(Apply)
#undef Apply

	read_only_param:	0,
	/* The compile time constants: */
	vga_support : !!(USER_SUPPORT & VGA_SUPPORT),
	big_malloc: !!(SETUP & BIG_MALLOC),
	bios_only: !!(SETUP & BIOS_ONLY),
	},

#if USER_SUPPORT & VGA_SUPPORT
    VGA_valid_mode: {
	0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
	0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF
	},
    vga_mode: {{0}},
#endif

    extra_cmdline: "",
    partial_checksum: 0
    };
 /* This is the end of config variables */

/*
 * The first element has to be in the first sector loaded of
 * the bootloader, Fat/Root blocks are not loaded.
 *
 * Take care, some BIOS can not read across track boundary... and
 * some may be limited (to what?) in the number of sectors they can
 * read at once.
 *
 * There is also the problem of 64K crossing DMA: if we are loading
 * boot2 sectors just after boot1 (at 0x0000:0x7C00), we can load
 * (0x10000 - 0x7C00)/512 = 66 sectors before the DMA cross.
 *
 * Note that the DOS bootloader will not load boot2, the operating
 * system has already loaded everything. Note also that the
 * relocation to segment 0x0840 is done after the loading phase, and
 * can move only 64 K in real mode.
 *
 * For BIOS, the sector 0 does not exists, sector 1 is the boot sector;
 * LBA starts at 0. We already have read 1 sectors, see 'boot1param',
 * so the (beginning of the) following array is now in memory:
 */
/* initialised by external software, size defined at compilation:
   There isn't any trick on the size - just define it as small as
   possible considering compiler version and floppy format (720 K)
   supported, and recompile: */
static const __attribute__ ((section (".text")))
bootloader2_type bootchain[8
#if defined (INCLUDE_E2FS) || (USER_SUPPORT != 0) || BUILD_TESTER()
			   + 25
#endif
				] = {};

/*
 * With GCC, we cannot access structure with asm(""), only with asm("": : :)
 * which has to be in a function, so do NOT call the following:
 * Why do I need to say section .text for this one?
 */
static void __attribute__((unused, section (".text")))
		dummy_do_not_call (void)
  {
  asm volatile (" serialconfport = %c0 " : : "i" (&boot1param.serialconf.port));
  asm volatile (" serialconf = %c0 "     : : "i" (&boot1param.serialconf));
  asm volatile (" bootloader2 = %c0 "    : : "i" (&boot1param.bootloader2));
  asm volatile (" boot1_msg = %c0 "      : : "i" (bootbefore.String));
  asm volatile (" boot2_msg = %c0 "      : : "i" (bootbefore.Volumelabel));
  asm volatile (" dot_msg = %c0 "        : : "i" (gujin_param.dot_msg) );
  asm volatile (" go_msg = %c0 "         : : "i" (gujin_param.go_msg) );
  asm volatile (" bootloader2_diskcmd = %c0 "
			: : "i" (&boot1param.bootloader2.bootloader2_cmd) );
  }

asm (
".text									\n"
"bootloader2_continue:							\n"
"	# Detect < 386 processors:					\n"
"	pushfw								\n"
"	popw	%ax							\n"
"	testw	$0x8000,%ax						\n"
"	jz	processor_not_8086					\n"
"	movw	$processor_8086_msg,%si					\n"
"	callw	*%di		# callw	simple_puts			\n"
"	jmp	.							\n"
"processor_not_8086:							\n"
"	testw	$0x7000,%ax						\n"
"	jnz	processor_sure_80386					\n"
"	orw	$0x2000,%ax						\n"
"	pushw	%ax							\n"
"	popfw								\n"
"	pushfw								\n"
"	popw	%ax							\n"
"	testw	$0x7000,%ax						\n"
"	jnz	processor_80386						\n"
"	movw	$processor_80286_msg,%si				\n"
"	callw	*%di		# callw	simple_puts			\n"
"	jmp	.							\n"
"processor_80386:							\n"
"	andw	$~0x2000,%ax						\n"
"	pushw	%ax							\n"
"	popfw								\n"
"processor_sure_80386:							\n"
"	# Clear NT and IOPL flags					\n"
"	pushf								\n"
"	popw	%ax							\n"
"	andw	$0x0FFF,%ax						\n"
"	pushw	%ax							\n"
"	popf								\n"
"	# restore 16 bit registers and store 32 bit registers as	\n"
"	# structure parameter to 'main()'				\n"
"	popaw	# take care of popa bug on i386				\n"
"	pushw	$0x454C		# Align stack to go faster ...		\n"
"	pushw	%fs							\n"
"	pushw	%gs							\n"
"	pushal								\n"
"	# Now it is time to move out of the way to be able		\n"
"	# to load another bootloader at physical address 0x7C00		\n"
"	# By the way, align %ds and %ss to be able to do in C:		\n"
"	#	void fct (int *a) {(*a)++;}				\n"
"	#	void caller (void) { int x; fct(&x); }			\n"
"									\n"
"	# clear the stack to be able to measure its used part,		\n"
"	# and clear BSS:						\n"
"	mov	%ss,%ax							\n"
"	mov	%ax,%es							\n"
"	xorl	%eax,%eax						\n"
"	std			# top to bottom				\n"
"	mov	%sp,%di							\n"
"	and	$0xFFFC,%di						\n"
"	mov	$_sbss,%cx						\n"
"	neg	%cx							\n"
"	sub	$4,%di							\n"
"	add	%di,%cx							\n"
"	shr	$2,%cx							\n"
"	rep stosl %eax,%es:(%di)					\n"
"									\n"
"	# Copy .rodata and .data (and .text if no CODE_SEGMENT):	\n"
"	mov	%ds,%ax							\n"
"	add	$deltaseg,%ax	# not necessary if no CODE_SEGMENT	\n"
"	mov	%ax,%ds							\n"
"	mov	%di,%si							\n"
"	mov	%di,%cx							\n"
"	shr	$2,%cx							\n"
"	rep movsl %ds:(%si),%es:(%di)					\n"
"	nop								\n"
"	sub	$deltaseg,%ax	# not necessary if no CODE_SEGMENT	\n"
"	mov	%ax,%ds							\n"
"									\n"

#if SETUP & EXTRA_SEGMENT
"									\n"
"	# Copy .extra							\n"
"	mov	$extraseg,%cx						\n"
"	jcxz	extraseg_null						\n"
"	mov	%ss,%ax							\n"
"	sub	$deltaseg,%ax						\n"
"	add	%cx,%ax							\n"
"	mov	%ax,%es							\n"
"	mov	%ds,%ax							\n"
"	add	%cx,%ax							\n"
"	mov	%ax,%ds							\n"
"	mov	$_eextra - 4,%si					\n"
"	mov	%si,%di							\n"
"	mov	$_eextra,%cx						\n"
"	shr	$2,%cx							\n"
"	rep movsl %ds:(%si),%es:(%di)					\n"
"	nop								\n"
"	sub	$extraseg,%ax						\n"
"	mov	%ax,%ds							\n"
"extraseg_null:								\n"
"									\n"
#endif /* EXTRA_SEGMENT */

#if SETUP & CODE_SEGMENT
"									\n"
"	# Copy .text: (if not copied with .data/.rodata)		\n"
"	mov	%ss,%ax							\n"
"	sub	$deltaseg,%ax						\n"
"	mov	%ax,%es							\n"
"	mov	$_etext,%si	# _etext is aligned on 4 bytes, and is < 64K\n"
"	mov	%si,%di							\n"
"	mov	%si,%cx							\n"
"	shr	$2,%cx							\n"
"	rep movsl %ds:(%si),%es:(%di)					\n"
"	nop								\n"
"									\n"
#endif /* CODE_SEGMENT */

"									\n"
"	# Realign code:							\n"
"	pushw	%es							\n"
"	pushw	$align_mover						\n"
"	lretw								\n"
"									\n"
"	# Interrupts / exceptions stuff:				\n"
"									\n"
"									\n"

#ifdef TREAT_COUNTER_TRACE_FLAG
"									\n"
"	.GLOBAL	TFflagHandler, TFflagCounter				\n"
"	.align 4							\n"
"TFflagCounter:								\n"
"	.long   0							\n"
"TFflagHandler:	# This ignores any other debugging interrupt...		\n"
"	incl	%cs:TFflagCounter					\n"
"	iretw								\n"
"									\n"
#endif /* TREAT_COUNTER_TRACE_FLAG */

#ifdef TREAT_TIMER
"									\n"
"	.GLOBAL timer_handler_ptr, timer_handler_oldptr			\n"
"	.GLOBAL timer_handler_relay, timer_counter			\n"
"	.align 4							\n"
"timer_counter:								\n"
"	.long	0							\n"
"timer_handler_ptr:	# set WITHOUT segment				\n"
"	.long	0							\n"
"timer_handler_oldptr:							\n"
"	.long	0							\n"
"timer_handler_relay:							\n"
"	cmpl	$0,%cs:timer_counter					\n"
"	je	timer_no_treat						\n"
"	cmpl	$0,%cs:timer_handler_ptr				\n"
"	je	timer_no_treat						\n"
"	decl	%cs:timer_counter					\n"
"	jne	timer_no_treat						\n"
"									\n"
"	xchgw	4(%esp),%bp						\n"
"	pushw	%bp							\n"
"	xchgw	6(%esp),%bp						\n"
"									\n"
"	pushw	%cs							\n"
"	pushw	$timer_handler_treat					\n"
"timer_no_treat:							\n"
"	ljmpw	*%cs:timer_handler_oldptr				\n"
"timer_handler_treat:							\n"
"	pushfw		# align stack					\n"
"	pushw	%gs							\n"
"	pushw	%fs							\n"
"	pushw	%ds							\n"
"	pushw	%es							\n"
"	pushl	%edx							\n"
"	pushl	%ecx							\n"
"	pushl	%eax							\n"
"	movw	%cs,%ax							\n"
"	addw	$deltaseg,%ax   # not necessary if no CODE_SEGMENT	\n"
"	movw	%ax,%ds							\n"
"	movw	%ax,%es							\n"
"	calll	*%cs:timer_handler_ptr	# set WITHOUT segment		\n"
"	popl	%eax		# timer_handler_ptr shall not use %ebp	\n"
"				# as a pointer (%ss not set)		\n"
"	popl	%ecx							\n"
"	popl	%edx							\n"
"	popw	%es							\n"
"	popw	%ds							\n"
"	popw	%fs							\n"
"	popw	%gs							\n"
"	popfw		# align stack					\n"
"	iretw								\n"
"									\n"
#endif /* TREAT_TIMER */

#ifdef TREAT_EXCEPTION
"									\n"
"	# handler for all exceptions:					\n"
"	# function 'exceptions' accept parameters			\n"
"	.GLOBAL exceptions						\n"
"									\n"

#if !(SETUP & BIOS_ONLY)
"									\n"
"	.GLOBAL control_break						\n"
"control_break:								\n"
"	iretw								\n"
"									\n"
#endif

"									\n"
"	.GLOBAL divide_error						\n"
"divide_error:								\n"
"	pushl	$0							\n"
"common_exception:							\n"
"	pushl	%es							\n"
"	pushl	%ds							\n"
"	pushal								\n"
"	movw	%cs,%ax							\n"
"	addw	$deltaseg,%ax	# not necessary if no CODE_SEGMENT	\n"
"	movw	%ax,%ds							\n"
"	movw	%ax,%es							\n"
"	calll	exceptions	# it shall not use %ebp as a pointer	\n"
"				# because %ss not set			\n"
"	popal			# take care of popa bug on i386		\n"
"	popl	%ds							\n"
"	popl	%es							\n"
"	add	$4,%sp							\n"
"	iretw								\n"
"									\n"

#if DEBUG & (DEBUG_ADDR | DEBUG_STACK)
"									\n"
"	.GLOBAL	bound_error						\n"
"bound_error:								\n"
"	pushl	$5							\n"
"	jmp	common_exception					\n"
"									\n"
#endif	/* DEBUG_ADDR | DEBUG_STACK */

"									\n"
"	.GLOBAL instruction_invalid					\n"
"instruction_invalid:							\n"
"	pushl	$6							\n"
"	jmp	common_exception					\n"
"	.GLOBAL stack_segment_fault					\n"
"stack_segment_fault:							\n"
"	pushl	$12							\n"
"	jmp	common_exception					\n"
"	.GLOBAL general_protection					\n"
"general_protection:	# NO error_code IN REAL MODE !			\n"
"	pushl	$13							\n"
"	jmp	common_exception					\n"
"									\n"

#endif /* TREAT_EXCEPTION */

"									\n"
"	# Back to the initialisation treatments:			\n"
"									\n"
"align_mover:								\n"
"	cld								\n"
"	movw	%ds,%dx							\n"
"	movw	%es,%cx							\n"
"	movw	%dx,%es							\n"
"	sub	%dx,%cx							\n"
"	shl	$2,%cx							\n"
"	xorl	%eax,%eax						\n"
"	xorw	%di,%di							\n"
"	rep stosl %eax,%es:(%di)					\n"
"	movw	%ss,%ax							\n"
"	movw	%ax,%ds							\n"
"	movw	%ax,%es							\n"
"									\n"
"	# For optimisation reasons, following variables are declared	\n"
"	# extern const in C, so we shall initialise them in assembler	\n"
"	# to prevent warnings:						\n"
"	# (initialise BEFORE calling install_exceptions)		\n"
"									\n"
"	.comm copy_boot1param_serialconf, 4				\n"
"	movl	%cs:serialconf,%eax					\n"
"	movl	%eax,copy_boot1param_serialconf				\n"
"									\n"

#if !(SETUP & BIOS_ONLY)
"									\n"
"	.comm dos_running, 1						\n"
"	movl	%cs:dospatch4adr,%eax					\n"
"	cmpl	$0x21CD4CB4,%eax					\n"
"	sete	dos_running						\n"
"									\n"
#endif

"									\n"
"	.comm dataseg_adr, 4						\n"
"	mov	%ds,dataseg_adr+2 # lower part is already zero because	\n"
"	.comm codeseg_adr, 4						\n"
"	mov	%cs,codeseg_adr+2 # .common is initialised like .bss	\n"
"									\n"

#if SETUP & EXTRA_SEGMENT
"									\n"
"	.comm extraseg_adr, 4						\n"
"	mov	%cs,%ax							\n"
"	addw	$extraseg,%ax						\n"
"	mov	%ax,extraseg_adr+2	#				\n"
"									\n"
#endif

#if ASSEMBLY_TYPE == ASSEMBLY_DSES
"									\n"
"	# easy one: (call long because of its retl)			\n"
"	calll	initial_open_door					\n"
"	testl	%eax,%eax	# value returned			\n"
"	jz	1f							\n"
"	pushw	%cs							\n"
"	popw	%ds							\n"
"	jmp	2f	# with DOS, when cannot open segs doors		\n"
"	# now it is time to do serious things				\n"
"	1:								\n"
"									\n"
#endif

#ifdef TREAT_EXCEPTION
"									\n"
"	# another easy one: (call long because of its retl)		\n"
"	calll	setup_vectors						\n"
"									\n"
#endif

"									\n"
"	xorl	%ebp,%ebp						\n"
"	xorl	%edi,%edi						\n"
"	xorl	%esi,%esi						\n"
"	xorl	%edx,%edx						\n"
"	xorl	%ecx,%ecx						\n"
"	xorl	%ebx,%ebx						\n"
"	mov	%esp,%eax						\n"
"	pushl	$0x0000352E						\n"
"	pushl	$0x30205620						\n"
"	pushl	$0x4E494152						\n"
"	pushl	$0x524F4C20						\n"
"	pushl	$0x656E6E65						\n"
"	pushl	$0x69744520						\n"
"	pushl	$0x31303032						\n"
"	pushl	$0x20294328						\n"
"	pushl	%eax		# main's parameter			\n"
"	calll	main							\n"
"	pushw	%cs							\n"
"	popw	%ds							\n"
"	addl	$36,%esp						\n"
"	testl	%eax,%eax						\n"
"	2:								\n"
"	popal	# take care of popa bug on i386				\n"
"	popw	%gs							\n"
"	popw	%fs							\n"
"	popw	%es	# crashes %es, to pop alignment word		\n"
"	pushaw								\n"
"	mov	return_address_bootloader1,%ax				\n"
"	jz	no_display_error					\n"
"	sub	$retoffsetboot2_error,%ax				\n"
"	cld								\n"
"	pushw	%si			# it will be restored		\n"
"	movw	$simple_puts,%di	# it has been erased,		\n"
"					# so set it again		\n"
"	movw	serialconfport,%bp	# also erased before		\n"
"no_display_error:							\n"
"	pushw	%ax							\n"
"	retw								\n"
	);

static const char processor_8086_msg[]  __attribute__ ((section (".text")))
	= "\r\nwill not run on 8086!\r\n";
static const char processor_80286_msg[]  __attribute__ ((section (".text")))
	= "\r\nwill not run on 80286!\r\n";

extern unsigned char start[], bootend[];
extern unsigned char dospatch1adr[], dospatch2adr[],
		     dospatch3adr[], dospatch4adr[];

/*
 * This is not static because its address is passed to
 * the kernel at startup (%cs relative):
 */
const struct instboot_info_str instboot_info
		 __attribute__ ((section (".text"))) = {
    start_adr: (unsigned) start,
    boot1param_adr: (unsigned) &boot1param,
    bootend_adr: (unsigned) bootend,
    bootchain_adr: (unsigned) bootchain,
    bootchain_nbblock: (unsigned) nbof (bootchain),
    gujin_param_adr: (unsigned) &gujin_param,
    patch1adr: (unsigned) dospatch1adr,
    patch2adr: (unsigned) dospatch2adr,
    patch3adr: (unsigned) dospatch3adr,
    patch4adr: (unsigned) dospatch4adr,
    sbss: (unsigned) _sbss,
    edata: (unsigned) _edata,
    deltaseg: (unsigned) deltaseg,	// not necessary if no CODE_SEGMENT
#if SETUP & EXTRA_SEGMENT /* else not initialised */
    extraseg: (unsigned) extraseg
#endif
    };

/*
 * The BOOT1 "library" functions:
 */
GENERATE_VOID_EXTRA2CODE_STUB(BOOT1_putstr, const char *str)
void BOOT1_putstr (const char *str)
  {
  unsigned short dummy1;

  asm volatile (
"	pushw	%%bp		# only 16 bits, aligned stack for int10	\n"
"	mov	%%dx,%%bp						\n"
"	callw	simple_puts						\n"
"	popw	%%bp							\n"
	: "=a" (dummy1), "+S" (str)
	: "d" (BOOT1_COM_port())
	);
  }

GENERATE_VOID_EXTRA2CODE_STUB(BOOT1_initdisks, unsigned char drive)
void BOOT1_initdisks (unsigned char drive)
  {
  unsigned short dummy;
  asm volatile (" callw init_bios_disk "
	: "=a" (dummy)
	: "d" (drive)	/* if drive & 0x80, all disks, else only floppys */
	: "cc" );	/* no return test, carry has been set manually */
  }

GENERATE_EXTRA2CODE_STUB(BOOT1_diskread, unsigned char,
				union diskcmd *str, farptr buffer)
unsigned char BOOT1_diskread (union diskcmd *str, farptr buffer)
  {
  unsigned short status, dummy;
  unsigned char  returned;

  asm volatile (
"	movw	%%es,%%ax						\n"
"	pushl	%4							\n"
"	popw	%%bx							\n"
"	popw	%%es							\n"
"	pushw	%%ax							\n"
"	callw	read_disk						\n"
"	popw	%%es							\n"
"	setc	%%dl	# set dest to 1 if carry, else clear dest	\n"
	: "=&a" (status), "=d" (returned), "=S" (dummy)
	: "S" (str), "g" (buffer)
	: "ebx", "edi", "ecx"
	);
  if (returned != 0)
      BDBG ((" [Error carry in 'diskread', ax=0x%X] ", status));
  return returned;
  }

#if (SETUP & UPDATE_BOOTPARAM) || defined (WRITE_ENABLE)
#ifndef WRITE_ENABLE
static inline
#else
GENERATE_EXTRA2CODE_STUB(BOOT1_diskwrite, unsigned char,
				union diskcmd *str, farptr buffer)
#endif
unsigned char
BOOT1_diskwrite (union diskcmd *str, farptr buffer)
  {
  unsigned short status, dummy;
  unsigned char  returned;

  asm volatile (
"	movw	%%es,%%ax						\n"
"	pushl	%4							\n"
"	popw	%%bx							\n"
"	popw	%%es							\n"
"	pushw	%%ax							\n"
"	callw	write_disk						\n"
"	popw	%%es							\n"
"	setc	%%dl	# set dest to 1 if carry, else clear dest	\n"
	: "=&a" (status), "=d" (returned), "=S" (dummy)
	: "S" (str), "g" (buffer)
	: "ebx", "edi", "ecx"
	);
  if (returned != 0)
      BDBG ((" [Error carry in 'diskwrite', ax=0x%X] ", status));
  return returned;
  }
#endif

#if SETUP & UPDATE_BOOTPARAM
asm (
"	.GLOBAL	write_disk						\n"
"write_disk:			# BIOS INT13 %ah=0x03			\n"
"	lodsw	%ds:(%si),%ax	# .reg_dx / .ide_dcr_address		\n"
"	mov	%ax,%dx							\n"
"	lodsw	%ds:(%si),%ax	# .reg_ax				\n"
"	mov	(%si),%cx	# .reg_cx / .nb_word_to_write (unused 13/43)\n"
"	cmp	$0xF0,%ah						\n"
"	je	write_ide_hard						\n"
"	mov	%es,6(%si)						\n"
"	mov	%bx,4(%si)						\n"
"	int	$0x13							\n"
"	retw								\n"
"									\n"
"# Copyrighted  Etienne LORRAIN, 1999-2000.				\n"
"# If you are reading that code before reading the GPL license,		\n"
"# you are actually fucking me.						\n"
"write_ide_hard:			# access IDE directly		\n"
"	mov	%bx,%di		# %bx should not be modified in this fonction\n"
"	outb	%al,%dx		# %al = 0x0A, %dx = 0x3F6 (for instance)\n"
"	lodsw	%ds:(%si),%ax	# .base_ide_plus7 (that is already in cx, but\n"
"				# it also increment %si and initialise %ah=1 !)\n"
"	mov	%ax,%dx		# for instance 0x1F7 (command/status register)\n"
"	xor	%cx,%cx		# 65536*1us (approx) = 65 ms		\n"
"write_waitnotbusy:							\n"
"	inb	%dx,%al							\n"
"	test	$0x80,%al	# busy					\n"
"	loopnz	write_waitnotbusy					\n"
"	jcxz	write_ide_error	# timeout: %ah=MSB(ideaddr), often 1	\n"
"	dec	%dx							\n"
"	outsb	%ds:(%si),%dx	# .lba_head				\n"
"	inc	%dx							\n"
"write_waitdiskready:							\n"
"	dec	%cx							\n"
"	jcxz	write_ide_error	# timeout BUSY or no DRDY or no RDY: %ah = 1\n"
"	inb	%dx,%al							\n"
"	test	$0xD0,%al	# clear carry				\n"
"	jle	write_waitdiskready	# I have understood that for you!\n"
"	jnp	write_waitdiskready					\n"
"	push	%dx		# %dx points to status/cmd register	\n"
"	dec	%dx							\n"
"	dec	%dx							\n"
"	outsb	%ds:(%si),%dx	# .cylinder_high			\n"
"	dec	%dx							\n"
"	outsb	%ds:(%si),%dx	# .cylinder_low				\n"
"	dec	%dx							\n"
"	outsb	%ds:(%si),%dx	# .sector				\n"
"	dec	%dx							\n"
"	outsb	%ds:(%si),%dx	# .nb_sect				\n"
"	pop	%dx							\n"
"	outsb	%ds:(%si),%dx	# .ide_command				\n"
"	lodsw	%ds:(%si),%ax	# .nb_read_sect, %ah = 0 (but if 256 sectors)\n"
"	mov	%ax,%cx							\n"
"write_loopnextsector:							\n"
"	dec	%ah		# debug only				\n"
"	push	%cx							\n"
"	xor	%cx,%cx							\n"
"write_waitcommandcomplete:						\n"
"	inb	%dx,%al							\n"
"	test	$0x81,%al	# clear carry				\n"
"	loopnz	write_waitcommandcomplete				\n"
"	jcxz	write_ide_error_popcx	# timeout/error/no DRQ: %ah = -nbsectortowrite\n"
"	test	$0x08,%al						\n"
"	jz	write_waitcommandcomplete				\n"
"	mov	(%si),%cx						\n"
"	push	%dx							\n"
"	sub	$7,%dx		# point to data register and clear carry\n"
"	rep outsw %es:(%di),%dx						\n"
"	pop	%dx							\n"
"	pop	%cx							\n"
"	loop	write_loopnextsector					\n"
"	retw								\n"
"write_ide_error_popcx:							\n"
"	pop	%cx							\n"
"write_ide_error:							\n"
"	stc								\n"
"	retw								\n"
	);

GENERATE_EXTRA2CODE_STUB(BOOT1_update_bootparam, unsigned char, void)
unsigned char BOOT1_update_bootparam (void)
  {
  unsigned short cpt, delta;
  extern char bootloader2_diskcmd[];
  struct gujin_param_str *param;
  union diskcmd cmdblock;	/* in current data/stack segment */
  unsigned char space[1024], *sector;
  farptr sector_fptr;

  /* This set segment bit 4..11 to zero, each page is 64Kb: */
  sector = space;
  sector_fptr = linear2farptr (farptr2linear (stack_adr(sector)));
  if ((sector_fptr & 0xFFFFU) + 512 > 0x10000) { /* DMA barrier */
      sector = &space[512];
      sector_fptr = linear2farptr (farptr2linear (stack_adr(sector)));
      }

  if (sizeof(bootsector_type) != 512)
      __ERROR();
  if (sizeof(struct gujin_param_str) % 2 != 0)
      __ERROR();

  if (get_gujin_param_attrib().read_only_param) {
      BDBG ((" [%s: Writing has been disabled] ", __FUNCTION__));
      return 1;
      }
  for (cpt = 0; cpt < sizeof (cmdblock); cpt++)
      ((char *)&cmdblock)[cpt] = codeseg_peekb (bootloader2_diskcmd + cpt);

  if (cmdblock.int1302.reg_ax != 0x0201) {
      BDBG ((" [%s: disable writing, ", __FUNCTION__));
      BDBG (("not 0x0201] "));
      disable_update_bootparam();
      return 2;
      }

  if ((cmdblock.int1302.reg_dx & 0xFF) != 0x00) {
      BDBG ((" [%s: disable writing, ", __FUNCTION__));
      BDBG (("not floppy] "));
      disable_update_bootparam();
      return 3;
      }

  cpt = 4;
  for (;;) {
      if (BOOT1_diskread (&cmdblock, sector_fptr) == 0)
	  break;
      if (--cpt == 0) {
	  BDBG ((" [%s: disable writing, ", __FUNCTION__));
	  BDBG (("error reading] "));
	  disable_update_bootparam();
	  return 4;
	  }
      }

  param = (struct gujin_param_str *)
		((unsigned)sector + (unsigned)&gujin_param - 512);

  if (param->magic != 16980327) {	/* changed floppy disk ? */
      BDBG ((" [%s: disable writing, ", __FUNCTION__));
      BDBG (("bad magic] "));
      disable_update_bootparam();
      return 5;
      }

  for (cpt = 0, delta = 0; cpt < sizeof(struct gujin_param_str)/2; cpt++) {
      unsigned short tmp
		= codeseg_peekw ((char *)&gujin_param + 2*cpt);

      delta += ((unsigned short *)param)[cpt] - tmp;
      ((unsigned short *)param)[cpt] = tmp;
      }
  param->partial_checksum += delta;

  cmdblock.int1302.reg_ax = 0x0301;
  cpt = 4;
  for (;;) {
      if (BOOT1_diskwrite (&cmdblock, sector_fptr) == 0)
	  break;
      if (--cpt == 0) {
	  BDBG ((" [%s: error writing] ", __FUNCTION__));
	  disable_update_bootparam();
	  return 6;
	  }
      }
  return 0;
  }
#endif /* UPDATE_BOOTPARAM */

#if SETUP & EXTRA_SEGMENT
__attribute__ ((section (".extra"), noreturn))
void extraseg_never_call_address_zero (void)
  {
  unsigned short i;
  for (i = 0; i < 5; i++)
      extraseg_BOOT1_putstr ("PANIC: address zero in extra segment called!"
				"\007\r\n");
#if 0
  {
  unsigned *stkptr = (unsigned *)getesp();
  int extraseg_printf (const char *format, ...);

  extraseg_printf ("stack: 0x%X, 0x%X, 0x%X, 0x%X, 0x%X, 0x%X, 0x%X, 0x%X\r\n",
			stkptr[0], stkptr[1], stkptr[2], stkptr[3],
			stkptr[4], stkptr[5], stkptr[6], stkptr[7]);
  }
#endif

  for (;;) ;
  }
#endif /* EXTRA_SEGMENT */
