/* boot.ld */

/* Gujin is a bootloader, it loads a Linux kernel from cold boot or DOS.
 * Copyright (C) 1999-2001 Etienne Lorrain, fingerprint (2D3AF3EA):
 *   2471 DF64 9DEE 41D4 C8DB 9667 E448 FF8C 2D3A F3EA
 * E-Mail: etienne.lorrain@masroudeau.com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "make.h"

/*
 * The linker file (man/info ld):
 */
MEMORY { ram : ORIGIN = 0, LENGTH = 64K }

/* Alignment of 32 bytes, i.e. 0x20 is forced by ld after output sections */

SECTIONS {
	.text 0 : AT (0) {
		_stext = . ;
		/*
		 * These two are not working (or I did not understand
		 * how to use them)
		 *
		STARTUP(boot.o)		=> gives double definition
		*(EXCLUDE_FILE ( libext2fs.a ) .text* )
		 */

		*(EXCLUDE_FILE (
fileio.o bmap.o closefs.o dir_iterate.o freefs.o inode.o openfs.o read_bb.o
valid_blk.o alloc.o badblocks.o bitops.o block.o dir_iterate.o alloc.o
rw_bitmaps.o bitmaps.o gen_bitmap.o dirblock.o swapfs.o bb_compat.o
				 ) .text* )
		KEEP (boot.o(.text));

		__sizeof_gujin_code_in_text = . - _stext ;

#if !(SETUP & EXTRA_SEGMENT)
#ifdef INCLUDE_E2FS
		_stexte2fs = . ;
		libext2fs.a(.text*)
		__sizeof_e2fs_code_in_text = . - _stexte2fs ;
#endif

		. = ALIGN (32) ;
		_etext = .;
		} > ram = 0

	__sizeof_all_code = SIZEOF(.text);
#else /* EXTRA_SEGMENT */
		. = ALIGN (4) ;
		} > ram = 0

	.text_extra : AT(SIZEOF(.text)) {
		*(.Xtext2extra*)
		. = ALIGN (32) ;
		_etext = .;
		} > ram = 0

	.extra 0 : AT(SIZEOF(.text) + SIZEOF(.text_extra)) {
		_sextra = . ;
		*(.extra*)
		__sizeof_gujin_code_in_extra = . - _sextra ;

#ifdef INCLUDE_E2FS
		_sextrae2fs = . ;
		libext2fs.a(.text*)
		__sizeof_e2fs_code_in_extra = . - _sextrae2fs ;
#endif

		. = ALIGN (4) ;
		} > ram = 0

	.extra_text : AT(SIZEOF(.text) + SIZEOF(.text_extra) + SIZEOF(.extra)) {
		*(.Xextra2text*)
		. = ALIGN (32) ;
		_eextra = . ;
		} > ram = 0

	__sizeof_inter_section_code = SIZEOF(.text_extra) + SIZEOF(.extra_text) ;
	__sizeof_all_code = SIZEOF(.text) + SIZEOF(.text_extra)
			  + SIZEOF(.extra) + SIZEOF(.extra_text);
#endif /* EXTRA_SEGMENT */

#if SETUP & CODE_SEGMENT
	.data 0 : AT(__sizeof_all_code) {
		. = 0x10 ; /* test null pointer dereference */
#else  /* CODE_SEGMENT */
	.data : {
#endif /* CODE_SEGMENT */
		_srodata = . ;
		*(.rodata)
		. = ALIGN (32) ;
		_erodata = . ;

		__sizeof_constants = _erodata - _srodata ;

		_end = .;
		_sdata = . ;
		*(.data)
		. = ALIGN (32);
		_sbss = .;

		__sizeof_inited_data = _sbss - _sdata ;

		} > ram = 0

	.bss (NOLOAD) : {
		*(.bss) *(COMMON)
		. = ALIGN (4);
		_edata = .;
		} > ram = 0

		__sizeof_zeroed_data = SIZEOF (.bss) ;

	__sizeof_all_data = __sizeof_constants + __sizeof_inited_data
					+ __sizeof_zeroed_data;

	.note (NOLOAD) : {
		*(.note)
		}
	.comment (NOLOAD) : {
		*(.comment)
		}
	.stab (NOLOAD) : {
		*(.stab)
		}
	.stabstr (NOLOAD) : {
		*(.stabstr)
		}
	}

#if SETUP & EXTRA_SEGMENT
NOCROSSREFS (.text .extra);
NOCROSSREFS (.extra_text .text_extra);
EXTERN (extraseg_never_call_address_zero); /* or KEEP() it */

#ifdef MALLOC_IN_EXTRASEG
library_malloc	= malloc ;
library_free	= free ;
library_realloc	= realloc ;
#else
library_malloc	= extraseg_malloc ;
library_free	= extraseg_free ;
library_realloc	= extraseg_realloc ;
#endif

library_printf	= extraseg_printf ;
library_sprintf	= extraseg_sprintf ;


deltaseg = (SIZEOF(.text) + SIZEOF (.text_extra)
		+ SIZEOF(.extra) + SIZEOF (.extra_text)) >> 4 ;
extraseg =  (SIZEOF (.text) + SIZEOF (.text_extra)) >> 4 ;

#else /* EXTRA_SEGMENT */

library_malloc	= malloc ;
library_free	= free ;
library_realloc	= realloc ;

library_printf	= printf ;
library_sprintf	= sprintf ;


#if SETUP & CODE_SEGMENT
deltaseg = SIZEOF(.text) >> 4 ;
#else
deltaseg = 0 ;
#endif

#endif /* EXTRA_SEGMENT */
