/* debug.h */
#ifndef DEBUG_H
#define DEBUG_H

/* Gujin is a bootloader, it loads a Linux kernel from cold boot or DOS.
 * Copyright (C) 1999-2001 Etienne Lorrain, fingerprint (2D3AF3EA):
 *   2471 DF64 9DEE 41D4 C8DB 9667 E448 FF8C 2D3A F3EA
 * E-Mail: etienne.lorrain@masroudeau.com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "make.h"

#if DEBUG & DEBUG_OUTPUT

#if (SETUP & EXTRA_SEGMENT) && defined (CALLING_FROM_EXTRASEG)
#define ASMNAME(x)	asm (STRING(extraseg_ ## x))
#else
#define ASMNAME(x)	/* call in same segment */
#endif

void dbg_init (void) ASMNAME(dbg_init);
#define DBG_INIT()	dbg_init()

void dbg_end (void) ASMNAME(dbg_end);
#define DBG_END()	dbg_end()

void dbg_printf (const char *format, ...) ASMNAME(dbg_printf);
#define DBG(X)		dbg_printf X /* to be used as DBG (("%d", 4)); */

#undef ASMNAME

#if DEBUG & DEBUG_BOOT
#define BDBG DBG
#else
#define BDBG(X) /* */
#endif

#if DEBUG & DEBUG_DISK
#define DDBG DBG
#else
#define DDBG(X) /* */
#endif

#if DEBUG & DEBUG_VIDEO
#define VDBG DBG
#else
#define VDBG(X) /* */
#endif

#if DEBUG & DEBUG_SERIAL
#define SDBG DBG
#else
#define SDBG(X) /* */
#endif

#if DEBUG & DEBUG_MALLOC
#define MDBG DBG
#else
#define MDBG(X) /* */
#endif

#if DEBUG & DEBUG_UTIL
#define UDBG DBG
#else
#define UDBG(X) /* */
#endif

#if DEBUG & DEBUG_MOUSE
#define PDBG DBG
#else
#define PDBG(X) /* */
#endif

#if DEBUG & DEBUG_LOADER
#define ZDBG DBG
#else
#define ZDBG(X) /* */
#endif

#if DEBUG & DEBUG_CACHE
#define CDBG DBG
#else
#define CDBG(X) /* */
#endif

#if DEBUG & DEBUG_IOCHANNEL
#define ODBG DBG
#else
#define ODBG(X) /* */
#endif

#if DEBUG & DEBUG_A20
#define ADBG DBG
#else
#define ADBG(X) /* */
#endif

#else	/* DEBUG_OUTPUT */

#define DBG_INIT()	/* */
#define DBG_END()	/* */
#define DBG(X)		/* */

#define BDBG(X) /* */
#define DDBG(X) /* */
#define VDBG(X) /* */
#define SDBG(X) /* */
#define MDBG(X) /* */
#define UDBG(X) /* */
#define PDBG(X) /* */
#define ZDBG(X) /* */
#define CDBG(X) /* */
#define ODBG(X) /* */
#define ADBG(X) /* */

#endif	/* DEBUG_OUTPUT */

/*
 * The stack and bad pointer protection:
 */

#if DEBUG & DEBUG_STACK
/*
 * limit included - but add 2 to high limit for reg16, and 4 for reg32
 * if not in bound, exception #BR generated (INT5).
 * iret from INT5 will retry the bound instruction.
 *
 * the structure are declared and initialised in library.c
 */
typedef struct {
  signed short low_limit;
  signed short high_limit;
  } __attribute__ ((packed)) short_bound_type;

extern short_bound_type stack_limit;	/* stack < 32 Kbytes */

extern inline void bound_stack (void)
  {
  asm volatile (" bound %%sp,%0 " : : "m" (stack_limit) );
  }

#else /* DEBUG_STACK */

extern inline void bound_stack (void) { }

#endif /* DEBUG_STACK */

#if DEBUG & DEBUG_ADDR

typedef struct {
  signed  low_limit;
  signed  high_limit;
  } __attribute__ ((packed)) long_bound_type;

extern long_bound_type data_limit;
extern volatile const char *BoundMsg;

extern inline void checkptr (const void *ptr, const char *msg)
  {
  BoundMsg = msg;
  asm volatile (" boundl %1,%0 " : : "m" (data_limit), "r" (ptr) );
  BoundMsg = 0;
  }

#else /* DEBUG_ADDR */

/* Using this would leave the constant message in the code:
extern inline void checkptr (const void *ptr, char *msg) {}
   so we simply:
 */
#define checkptr(ptr, msg) /* */

#endif /* DEBUG_ADDR */

#endif /* DEBUG_H */
