/* ide.h */
#ifndef IDE_H
#define IDE_H

/* Gujin is a bootloader, it loads a Linux kernel from cold boot or DOS.
 * Copyright (C) 1999-2001 Etienne Lorrain, fingerprint (2D3AF3EA):
 *   2471 DF64 9DEE 41D4 C8DB 9667 E448 FF8C 2D3A F3EA
 * E-Mail: etienne.lorrain@masroudeau.com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

extern inline unsigned short
_IDE_guess_altadr (unsigned short ideIOadr)
  {
  /*
   * This works with (someone get better?):
   * ide0=0x01F0,0x03F6,14
   * ide1=0x0170,0x0376,15
   * ide2=0x01e8,0x03ee,11
   * ide3=0x0168,0x036e,10
   * ide4=0xe820,0xe836,5
   * ide5=0xe828,0xe83e,5
   */
  if (ideIOadr < 0x206)
      return ideIOadr + 0x206;
    else
      return ideIOadr + 0x016;
  }

extern inline void
_IDE_set_drivehead (unsigned short ideIOadr, unsigned char value)
  {
  outb (ideIOadr + 6, value);
  }

extern inline unsigned char
_IDE_get_drivehead (unsigned short ideIOadr)
  {
  return inb (ideIOadr + 6);
  }

extern inline unsigned char
_IDE_get_status (unsigned short ideIOadr)
  {
  return inb (ideIOadr + 7);
  }

extern inline unsigned char
_IDE_get_alternate_status (unsigned short ideIOadr)
  {
  return inb (_IDE_guess_altadr (ideIOadr));
  }

extern inline unsigned char
_IDE_get_card_addr_reg (unsigned short ideIOadr)
  {
  return inb (_IDE_guess_altadr (ideIOadr) + 1);
  }

extern inline unsigned short
_IDE_request_sense (unsigned short ideIOadr)
  {
  unsigned short error_sense;

  asm ASMREAD (
"	inc	%%dx					\n"
"	inb	%%dx,%%al				\n"
"	mov	%%al,%%ah				\n"
"	add	$6,%%dx					\n"
"	mov	$0x03,%%al	# request sense		\n"
"	outb	%%al,%%dx				\n"
"	xor	%%cx,%%cx				\n"
"looprequest:						\n"
"	inb	%%dx,%%al				\n"
"	test	$0x80,%%al				\n"
"	loopnz	looprequest				\n"
"	sub	$6,%%dx					\n"
"	inb	%%dx,%%al				\n"
"	dec	%%dx					\n"
	: "=a" (error_sense)
	: "d" (ideIOadr)
	: "cx"
	);
  return error_sense;
  }

extern inline unsigned char
_IDE_calibratedrive (unsigned char drive, unsigned short ideIOadr,
		     unsigned short *timeout)
  {
  unsigned char status;

  /* ATA3 : C/H/S should point to first sector */
  asm volatile (
"	add	$2,%%dx							\n"
"	outb	%%al,%%dx	# %%al == 0, sector count		\n"
"	inc	%%dx							\n"
"	inc	%%al							\n"
"	outb	%%al,%%dx	# not in LBA mode, sector = 1		\n"
"	dec	%%al							\n"
"	inc	%%dx							\n"
"	outb	%%al,%%dx	# cylinder low				\n"
"	inc	%%dx							\n"
"	outb	%%al,%%dx	# cylinder high				\n"
"	inc	%%dx							\n"
"	mov	%%ah,%%al						\n"
"	outb	%%al,%%dx	# lba/head				\n"
"	inc	%%dx							\n"
"	mov	$0x10,%%al	# calibrate drive			\n"
"	outb	%%al,%%dx						\n"
"	xor	%%cx,%%cx						\n"
"loop_waitcalibrate:							\n"
"	inb	%%dx,%%al						\n"
"	test	$0x80,%%al						\n"
"	loopnz	loop_waitcalibrate					\n"
"	jcxz    error_waitcalibrate					\n"
"	test    $0x50,%%al						\n"
"	jz      loop_waitcalibrate					\n"
"	jnp     loop_waitcalibrate					\n"
"error_waitcalibrate:							\n"
"	sub	$7,%%dx							\n"
	: "=c" (*timeout), "=a" (status)
	: "d" (ideIOadr), "a" ((0xA0|((drive)? 0x10:0) << 8))
	);

  return status;
  }

extern inline unsigned char
_IDE_Initialisedrive (unsigned char drive, unsigned short ideIOadr,
		      unsigned short nbcylinder,
		      unsigned char nbhead,
		      unsigned char nbsectorpertrack,
		      unsigned short *timeout)
  {
  unsigned char status;

  asm volatile (
"	add	$2,%%dx							\n"
"	outb	%%al,%%dx						\n"
"	inc	%%dx							\n"
"	outb	%%al,%%dx	# just in case...			\n"
"									\n"
"	inc	%%dx		#					\n"
"	mov	%%cl,%%al	# just in case...			\n"
"	outb	%%al,%%dx	#					\n"
"	inc	%%dx		#					\n"
"	mov	%%ch,%%al	#					\n"
"	outb	%%al,%%dx	#					\n"
"									\n"
"	inc	%%dx							\n"
"	mov	%%ah,%%al						\n"
"	outb	%%al,%%dx						\n"
"	inc	%%dx							\n"
"	mov	$0x91,%%al	# initialise drive nbhead & nbsector	\n"
"	xor	%%cx,%%cx						\n"
"loop_waitinitialise:							\n"
"	inb	%%dx,%%al						\n"
"	test	$0x80,%%al						\n"
"	loopnz	loop_waitinitialise					\n"
"	jcxz	error_waitinitialise					\n"
"	test    $0x50,%%al						\n"
"	jz      loop_waitinitialise					\n"
"	jnp     loop_waitinitialise					\n"
"error_waitinitialise:							\n"
"	sub	$7,%%dx							\n"
	: "=c" (*timeout), "=a" (status) /* status can be OK even if error */
	: "d" (ideIOadr), "c" (nbcylinder),
	  "a" (((0xA0 | ((drive)? 0x10 : 0) | ((nbhead-1) & 0x0F)) << 8)
					    | nbsectorpertrack)
	);

  return status;
  }

extern inline void
_IDE_flush (unsigned short IOadr)
  {
  /* command could be read multiple, with multiple up to 64: */
  asm volatile (
"loop_flush:						\n"
"	inw	%%dx,%%ax				\n"
"	jcxz	.		# wait a little bit	\n"
"	jcxz	.					\n"
"	loop	loop_flush				\n"
	: : "d" (IOadr), "c" (64*256) : "ax"
	);
  }

extern inline unsigned char
_IDE_reset (unsigned short ideIOadr, unsigned *timeout)
  {
  unsigned char status;
  unsigned twoadr = ideIOadr + 7;

  twoadr = (twoadr << 16) | _IDE_guess_altadr (ideIOadr);
  /*
   * to be called with *timeout = 0xF0000,
   * returns a "timeout" if *timeout = 0
   */
  asm volatile (
"	pushl	%%ecx						\n"
"	outb	%%al,%%dx					\n"
"loop_waitreset:						\n"
"	inb	%%dx,%%al	# wait a little bit		\n"
"	loopl	loop_waitreset					\n"
"	mov	%%ah,%%al					\n"
"	outb	%%al,%%dx					\n"
"	shrl	$16,%%edx					\n"
"	popl	%%ecx						\n"
"loop_waitbusy:							\n"
"	inb	%%dx,%%al					\n"
"	test	$0x80,%%al					\n"
"	loopnzl loop_waitbusy					\n"
	: "+c" (*timeout), "=a" (status), "+d" (twoadr)
	: "a" (0x0A0E)
	);

  return status;
  }

extern inline unsigned char
_IDE_diagnose (unsigned short ideIOadr, unsigned *timeout,
	unsigned char *diagcode)
  {
  unsigned short status_diagcode;
  /*
   * to be called with *timeout = 0x1000000,
   * returns a "timeout" if *timeout = 0
   */

  asm volatile (
"	add	$7,%%dx				\n"
"	outb	%%al,%%dx			\n"
"loop_waitdiagnostic:				\n"
"	inb	%%dx,%%al			\n"
"	test	$0x80,%%al			\n"
"	loopnzl loop_waitdiagnostic		\n"
"	mov	%%al,%%ah			\n"
"	sub	$6,%%dx				\n"
"	inb	%%dx,%%al			\n"
"	dec	%%dx				\n"
	: "=c" (*timeout), "=a" (status_diagcode)
	: "d" (ideIOadr), "a" (0x90), "c" (*timeout)
	);
  *diagcode = (unsigned char) status_diagcode;
  return status_diagcode >> 8;
  }

extern inline unsigned char
_IDE_idle (unsigned char drive, unsigned short ideIOadr,
	   unsigned *timeout, unsigned char time)
  {
  unsigned char status;
  /*
   * to be called with *timeout = 0x1000000,
   * returns a "timeout" if *timeout = 0
   */

  asm volatile (
"	add	$2,%%dx							\n"
"	outb	%%al,%%dx	# 0 disabled, up to 240 by 5 sec increment\n"
"	add	$4,%%dx							\n"
"	mov	%%ah,%%al						\n"
"	outb	%%al,%%dx						\n"
"	inc	%%dx							\n"
"#	mov	$0xE3,%%al  # Idle: 0x97 or 0xE3, later preferred	\n"
"	mov	$0xE1,%%al  # Idle immediate : 0x95 or 0xE1, later preferred\n"
"	outb	%%al,%%dx						\n"
"loop_idle:								\n"
"	inb	%%dx,%%al						\n"
"	test	$0x80,%%al						\n"
"	loopnzl loop_idle						\n"
"	jecxz	error_idle						\n"
"	test	$0x50,%%al						\n"
"	jz	loop_idle						\n"
"	jnp	loop_idle						\n"
"error_idle:								\n"
"	sub	$7,%%dx							\n"
	: "=c" (*timeout), "=a" (status)
	: "d" (ideIOadr), "c" (*timeout),
	  "a" (((0xA0 | ((drive)? 0x10 : 0)) << 8) | time)
	);
  return status;
  }

extern inline unsigned char
_IDE_seek (unsigned char drive, unsigned short ideIOadr,
	   unsigned short cylinder, unsigned char head)
  {
  unsigned short timeout;
  unsigned char status;

  asm volatile (
"	add	$2,%%dx						\n"
"	outb	%%al,%%dx	# clear it just in case...	\n"
"	inc	%%dx						\n"
"	outb	%%al,%%dx	# clear it just in case...	\n"
"	inc	%%dx						\n"
"	mov	%%cl,%%al					\n"
"	outb	%%al,%%dx					\n"
"	inc	%%dx						\n"
"	mov	%%ch,%%al					\n"
"	outb	%%al,%%dx					\n"
"	inc	%%dx						\n"
"	mov	%%ah,%%al					\n"
"	outb	%%al,%%dx					\n"
"	inc	%%dx						\n"
"	mov	$0x70,%%al	# 0x7X : seek			\n"
"	xor	%%cx,%%cx					\n"
"loop_waitseek:							\n"
"	inb	%%dx,%%al					\n"
"	test	$0x80,%%al					\n"
"	loopnz	loop_waitseek					\n"
"	jcxz	error_waitseek					\n"
"	test    $0x50,%%al					\n"
"	jz      loop_waitseek					\n"
"	jnp     loop_waitseek					\n"
"error_waitseek:						\n"
"	sub	$7,%%dx						\n"
	: "=c" (timeout), "=a" (status)
	: "d" (ideIOadr), "c" (cylinder),
	  "a" ((0xA0 | ((drive)? 0x10 : 0) | (head & 0x0F)) << 8)
	);

  return status;
  }

extern inline unsigned char
_IDE_verifysector (unsigned char drive, unsigned short ideIOadr)
  {
  unsigned short timeout;
  unsigned char status;

  asm volatile (
"	add	$2,%%dx					\n"
"	mov	$1,%%al					\n"
"	outb	%%al,%%dx	# sector count		\n"
"	inc	%%dx					\n"
"	outb	%%al,%%dx	# sector		\n"
"	inc	%%dx					\n"
"	mov	$0,%%al					\n"
"	outb	%%al,%%dx	# cylinder low		\n"
"	inc	%%dx					\n"
"	outb	%%al,%%dx	# cylinder high		\n"
"	inc	%%dx					\n"
"	mov	%%cl,%%al				\n"
"	outb	%%al,%%dx	# lba, head		\n"
"	inc	%%dx					\n"
"	mov	$0x40,%%al	# verify		\n"
"	outb	%%al,%%dx				\n"
"	xor	%%cx,%%cx				\n"
"loop_waitverify:					\n"
"	inb	%%dx,%%al				\n"
"	test	$0x80,%%al				\n"
"	loopnz	loop_waitverify				\n"
"	sub	$7,%%dx					\n"
	: "=c" (timeout), "=a" (status)
	: "d" (ideIOadr), "c" (0xA0|((drive)? 0x10:0))
	);

  return status;
  }

extern inline unsigned char
_IDE_setfeature (unsigned char drive, unsigned short ideIOadr,
		 unsigned char feature, unsigned char value)
  {
  unsigned short timeout;
  unsigned char status;

  asm volatile (
"	inc	%%dx					\n"
"	outb	%%al,%%dx				\n"
"	inc	%%dx					\n"
"	mov	%%cl,%%al				\n"
"	outb	%%al,%%dx				\n"
"	add	$4,%%dx					\n"
"	mov	%%ah,%%al				\n"
"	outb	%%al,%%dx				\n"
"	inc	%%dx					\n"
"	mov	$0xEF,%%al	# Set Feature		\n"
"	outb	%%al,%%dx				\n"
"	xor	%%cx,%%cx				\n"
"loop_waitsetfeature:					\n"
"	inb	%%dx,%%al				\n"
"	test	$0x80,%%al				\n"
"	loopnz	loop_waitsetfeature			\n"
"	sub	$7,%%dx					\n"
	: "=c" (timeout), "=a" (status)
	: "d" (ideIOadr), "a" ( (0xA0|((drive)? 0x10:0) << 8) | feature),
	  "c" (value)
	);

  return status;
  }

extern inline unsigned char
_IDE_setmultiple (unsigned char drive, unsigned short ideIOadr,
		  unsigned char value)
  {
  unsigned short timeout;
  unsigned char status;

  asm volatile (
"	add	$2,%%dx					\n"
"	outb	%%al,%%dx				\n"
"	add	$4,%%dx					\n"
"	mov	%%ah,%%al				\n"
"	outb	%%al,%%dx				\n"
"	inc	%%dx					\n"
"	mov	$0xC6,%%al	# Set Multiple		\n"
"	outb	%%al,%%dx				\n"
"	xor	%%cx,%%cx				\n"
"loop_waitsetmultiple:					\n"
"	inb	%%dx,%%al				\n"
"	test	$0x80,%%al				\n"
"	loopnz	loop_waitsetmultiple			\n"
"	sub	$7,%%dx					\n"
	: "=c" (timeout), "=a" (status)
	: "d" (ideIOadr), "a" ( (0xA0|((drive)? 0x10:0) << 8) | value)
	);

  return status;
  }

#endif /* IDE_H */
