/* libextra.h */
#ifndef LIBEXTRA_H
#define LIBEXTRA_H

/* Gujin is a bootloader, it loads a Linux kernel from cold boot or DOS.
 * Copyright (C) 1999-2001 Etienne Lorrain, fingerprint (2D3AF3EA):
 *   2471 DF64 9DEE 41D4 C8DB 9667 E448 FF8C 2D3A F3EA
 * E-Mail: etienne.lorrain@masroudeau.com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "make.h"

extern struct processor_str {
    enum processor_enum {
	proc_nocpuid = 0,
	proc_intel,
	proc_amd,
	proc_umc,
	proc_cyrix,
	proc_nexgen,
	proc_idt,
	proc_rise,
	proc_transmeta,
	proc_unknown
	} brand;
    unsigned family;
    unsigned stdflags;
    unsigned extraflags;
    char longname[3 * 4 * 4];
    /* not initialised by init_processor_str() : */
    unsigned calibrate_rdtsc;	/* cycle (rdtsc) per second */
    unsigned DOS_version;	/* 0 if on BIOS */
    enum { stdkbd, extkbd, ext122kbd, serialkbd } keyboard;
    unsigned joystick_present;
    } processor;

static inline int kbd_use_ext_fct(void)
  {
  return processor.keyboard != stdkbd;
  }

static inline unsigned joystick_present(void)
  {
  return processor.joystick_present;
  }


/*
 * The functions:
 */
#if (SETUP & EXTRA_SEGMENT) && !defined (CALLING_FROM_EXTRASEG)
#define ASMNAME(x)	asm (STRING(codeseg_ ## x))
#else
#define ASMNAME(x)	/* call in same segment */
#endif

void detect_processor(struct processor_str *proc)
				ASMNAME(detect_processor);

unsigned get_number (unsigned deflt, unsigned max)
				ASMNAME(get_number);

#undef ASMNAME

#endif /* LIBEXTRA_H */
