/* libincl.h */

/* Gujin is a bootloader, it loads a Linux kernel from cold boot or DOS.
 * Copyright (C) 1999-2001 Etienne Lorrain, fingerprint (2D3AF3EA):
 *   2471 DF64 9DEE 41D4 C8DB 9667 E448 FF8C 2D3A F3EA
 * E-Mail: etienne.lorrain@masroudeau.com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

asm (" .code16gcc ");

extern char memory[];

//#define I386_MOVS

#ifdef LONG_STR_ASM

#define SRCREG		"%%esi"
#define DSTREG		"%%edi"

#ifdef I386_MOVS
#define MOVS_PATCH	" ; addr32 nop "
#else
#define MOVS_PATCH	/**/
#endif

asm (
".macro movsl src=%ds:(%esi), dst=%es:(%edi)				\n"
"	.byte 0x67, 0x66, 0xA5	# movs.l %ds:(%esi),%es:(%edi)		\n"
"	" MOVS_PATCH "							\n"
".endm									\n"
".macro movsw src=%ds:(%esi), dst=%es:(%edi)				\n"
"	.byte 0x67, 0xA5	# movs.w %ds:(%esi),%es:(%edi)		\n"
"	" MOVS_PATCH "							\n"
".endm									\n"
".macro movsb src=%ds:(%esi), dst=%es:(%edi)				\n"
"	.byte 0x67, 0xA4	# movs.b %ds:(%esi),%es:(%edi)		\n"
"	" MOVS_PATCH "							\n"
".endm									\n"
"									\n"
".macro stosl src=%eax, dst=%es:(%edi)					\n"
"	.byte 0x67, 0x66, 0xAB	# stos.l %eax,%es:(%edi)		\n"
".endm									\n"
".macro stosw src=%ax, dst=%es:(%edi)					\n"
"	.byte 0x67, 0xAB	# stos.w %ax,%es:(%edi)			\n"
".endm									\n"
".macro stosb src=%al, dst=%es:(%edi)					\n"
"	.byte 0x67, 0xAA	# stos.b %al,%es:(%edi)			\n"
".endm									\n"
"									\n"
".macro scasl str=%es:(%edi), chr=%eax					\n"
"	.byte 0x67, 0x66, 0xAF	# scas.l %es:(%edi),%eax		\n"
".endm									\n"
".macro scasw str=%es:(%edi), chr=%ax					\n"
"	.byte 0x67, 0xAF	# scas.w %es:(%edi),%ax			\n"
".endm									\n"
".macro scasb str=%es:(%edi), chr=%al					\n"
"	.byte 0x67, 0xAE	# scas.b %es:(%edi),%al			\n"
".endm									\n"
"									\n"
".macro cmpsl src=%eax, dst=%es:(%edi)					\n"
"	.byte 0x67, 0x66, 0xA7	# cmps.l %es:(%edi),%ds:(%esi)		\n"
".endm									\n"
".macro cmpsw src=%ax, dst=%es:(%edi)					\n"
"	.byte 0x67, 0xA7	# cmps.w %es:(%edi),%ds:(%esi)		\n"
".endm									\n"
".macro cmpsb src=%al, dst=%es:(%edi)					\n"
"	.byte 0x67, 0xA6	# cmps.b %es:(%edi),%ds:(%esi)		\n"
".endm									\n"
"									\n"
".macro lodsl str=%es:(%edi), chr=%eax					\n"
"	.byte 0x67, 0x66, 0xAD	# lods.l %es:(%edi),%eax		\n"
".endm									\n"
".macro lodsw str=%es:(%edi), chr=%ax					\n"
"	.byte 0x67, 0xAD	# lods.w %es:(%edi),%ax			\n"
".endm									\n"
".macro lodsb str=%es:(%edi), chr=%al					\n"
"	.byte 0x67, 0xAC	# lods.b %es:(%edi),%al			\n"
".endm									\n"
"									\n"
"# same macros for ins[bwl], outs[bwl], I am fed up to write.		\n"
);

#else /* LONG_STR_ASM */

#define SRCREG		"%%si"
#define DSTREG		"%%di"

#ifdef I386_MOVS
#define MOVS_PATCH	" ; nop "
#else
#define MOVS_PATCH	/**/
#endif

asm (
".macro movsl src=%ds:(%si), dst=%es:(%di)				\n"
"	.byte 0x66, 0xA5	# movs.l %ds:(%si),%es:(%di)		\n"
"	" MOVS_PATCH "							\n"
".endm									\n"
".macro movsw src=%ds:(%si), dst=%es:(%di)				\n"
"	.byte 0xA5		# movs.w %ds:(%si),%es:(%di)		\n"
"	" MOVS_PATCH "							\n"
".endm									\n"
".macro movsb src=%ds:(%si), dst=%es:(%di)				\n"
"	.byte 0xA4		# movs.b %ds:(%si),%es:(%di)		\n"
"	" MOVS_PATCH "							\n"
".endm									\n"
);

#endif /* LONG_STR_ASM */

#ifndef inline_all_stringops    /* GCC-3.0 do it itself hopefully */

#define _STRING_H       1
#define __NO_STRING_INLINES	/* for "/usr/include/bits/string*.h" */

extern inline void _memset (void *buffer, int value, unsigned size)
  {
  asm (
"	cld	# _memset modify %2	\n"
"	rep stosb %%al,%%es:("DSTREG")	\n"
	: "+c" (size), "+D" (buffer), /* in fact %es:%edi */
	  "=m" (*memory)
	: "a" (value)
	: "cc");
  }

extern inline void _memcpy (void *dst, const void *src, unsigned size)
  {
  asm (
"	cld	# _memcpy modify %3	\n"
"	rep movsb %%ds:("SRCREG"),%%es:("DSTREG")" MOVS_PATCH	"\n"
	: "+c" (size), "+D" (dst), "+S" (src), "+m" (*memory)
	:
	: "cc");
  }

extern inline int _memcmp (const void *s1, const void *s2, unsigned nb)
  {
  asm (
"	cld	# _memcmp depends on %3			\n"
"	repe cmpsb %%es:("DSTREG"),%%ds:("SRCREG")	\n"
	: "+c" (nb), "+S" (s1), "+D" (s2)
	: "m" (*memory)
	: "cc" );
  ((unsigned char *)s1)--;
  ((unsigned char *)s2)--;

  if (*(unsigned char *)s1 == *(unsigned char *)s2)
      return 0;
  if (*(unsigned char *)s1 > *(unsigned char *)s2)
      return 1;
    else
      return -1;
  }

extern inline const char *_strnchr (const char *str, char c, unsigned size)
  {
  asm ("cld ; repne scasb %%es:("DSTREG"),%%al"
	: "+c" (size), "+D" (str) /* in fact %es:%edi */
	: "a" (c), "m" (*memory)
	: "cc");
  return str - 1;
  }


extern inline void *memcpy (void *dst, const void *src, unsigned size)
  {
  if (__builtin_constant_p (size)) {
      if (size == 1) {
	  ((unsigned char *)dst)[0] = ((const unsigned char *)src)[0];
	  return dst;
	  }
	else if (size == 2) {
	  ((unsigned short *)dst)[0] = ((const unsigned short *)src)[0];
	  return dst;
	  }
	else if (size == 3) {
	  ((unsigned short *)dst)[0] = ((const unsigned short *)src)[0];
	  ((unsigned char *)dst)[2] = ((const unsigned char *)src)[2];
	  return dst;
	  }
	else if (size == 4) {
	  ((unsigned *)dst)[0] = ((const unsigned *)src)[0];
	  return dst;
	  }
	else if (size == 5) {
	  ((unsigned *)dst)[0] = ((const unsigned *)src)[0];
	  ((unsigned char *)dst)[4] = ((const unsigned char *)src)[4];
	  return dst;
	  }
	else if (size == 6) {
	  ((unsigned *)dst)[0] = ((const unsigned *)src)[0];
	  ((unsigned short *)dst)[2] = ((const unsigned short *)src)[2];
	  return dst;
	  }
	else if (size == 8) {
	  ((unsigned *)dst)[0] = ((const unsigned *)src)[0];
	  ((unsigned *)dst)[1] = ((const unsigned *)src)[1];
	  return dst;
	  }
      }
  _memcpy (dst, src, size);
  return dst;
  }

extern inline void *memset (void *dst, int value, unsigned size)
  {
  if (__builtin_constant_p (size)) {
      if (size == 1) {
	  ((unsigned char *)dst)[0] = value;
	  return dst;
	  }
	else if (size == 2) {
	  unsigned short val = (value & 0xFF) | (value << 8);
	  ((unsigned short *)dst)[0] = val;
	  return dst;
	  }
	else if (size == 3) {
	  unsigned short val = (value & 0xFF) | (value << 8);
	  ((unsigned short *)dst)[0] = val;
	  ((unsigned char *)dst)[2] = val;
	  return dst;
	  }
	else if (size == 4) {
	  unsigned val = value & 0xFF;
	  val |= val << 8;
	  val |= val << 16;
	  ((unsigned *)dst)[0] = val;
	  return dst;
	  }
	else if (size == 5) {
	  unsigned val = value & 0xFF;
	  val |= val << 8;
	  val |= val << 16;
	  ((unsigned *)dst)[0] = val;
	  ((unsigned char *)dst)[4] = val;
	  return dst;
	  }
	else if (size == 6) {
	  unsigned val = value & 0xFF;
	  val |= val << 8;
	  val |= val << 16;
	  ((unsigned *)dst)[0] = val;
	  ((unsigned short *)dst)[2] = val;
	  return dst;
	  }
	else if (size == 8) {
	  unsigned val = value & 0xFF;
	  val |= val << 8;
	  val |= val << 16;
	  ((unsigned *)dst)[0] = val;
	  ((unsigned *)dst)[1] = val;
	  return dst;
	  }
      }
  _memset (dst, value, size);
  return dst;
  }

extern inline int memcmp (const void *s1, const void *s2, unsigned nb)
  {
  return _memcmp (s1, s2, nb);
  }

#define MAXSTRING (~0)

extern inline unsigned strlen (const char *str)
  {
  return _strnchr (str, '\0', MAXSTRING) - str;
  }

extern inline void _strncpy (char *dst, const char *src, unsigned nb)
  {
  unsigned len = strlen(src);
  if (len > nb)
      len = nb;
  memcpy (dst, src, len);
  dst[len] = '\0';
  }

extern inline int _strncmp (const char *s1, const char *s2, unsigned nb)
  {
  unsigned len = strlen(s1) + 1;
  if (len > nb)
      len = nb;
  return _memcmp (s1, s2, len);
  }

extern inline const char *strchr (const char *str, char c)
  {
  const char *returned = _strnchr (str, c, strlen(str));

  if (*returned == c)
      return returned;
    else
      return 0;
  }

extern inline char *strncpy (char *dst, const char *src, unsigned nb)
  {
  _strncpy (dst, src, nb);
  return dst;
  }

extern inline char *strcpy (char *dst, const char *src)
  {
  _strncpy (dst, src, MAXSTRING - 1);
  return dst;
  }

extern inline char *strcat (char *dst, const char *src)
  {
  strcpy (dst + strlen (dst), src);
  return dst;
  }

extern inline int strncmp (const char *s1, const char *s2, unsigned nb)
  {
  return _strncmp (s1, s2, nb);
  }

extern inline int strcmp (const char *s1, const char *s2)
  {
  return _strncmp (s1, s2, MAXSTRING);
  }

extern inline void *memmove (void *dst, const void *src, unsigned size)
  {
  void *returned = dst;

  if (dst < src)
    memcpy (dst, src, size);
  else {
    void *_dst = dst + size - 1;
    const void *_src = src + size - 1;

    asm volatile ("std ; rep movsb %%ds:("SRCREG"),%%es:("DSTREG") " MOVS_PATCH
		  "; cld"
	  : "+c" (size), "+D" (_dst), "+S" (_src) : : "memory");
    }
  return returned;
  }

extern inline char *strerror (int __errnum)
  {
  return "error!";
  }
#endif /* inline_all_stringops */

#ifdef _EXT2_HAVE_ASM_SWAB
extern inline unsigned short ext2fs_swab16(unsigned short val)
  {
	return (val >> 8) | (val << 8);
  }

extern inline unsigned ext2fs_swab32(unsigned val)
  {
#if 0
  // The C version spend all time writing
  // the same value over different registers:
#if 1
  union {
      unsigned l;
      unsigned short s;
      } tmp = { val };

  tmp.s = (tmp.s >> 8) | (tmp.s << 8);
  tmp.l = (tmp.l >> 16) | (tmp.l << 16);
  tmp.s = (tmp.s >> 8) | (tmp.s << 8);

  val = tmp.l;
#else
  *(unsigned short *)&val = (*(unsigned short *)&val >> 8)
	  | (*(unsigned short *)&val << 8);
  val = (val >> 16) | (val << 16);
  *(unsigned short *)&val = (*(unsigned short *)&val >> 8)
	  | (*(unsigned short *)&val << 8);

#endif
#else
 // We do that better than GCC:
#if defined(__i386__)
  asm volatile (
"	rolw	$8,%w0	# the 'w' is ignored if '+g' is a reference	\n"
"	roll	$16,%0							\n"
"	rolw	$8,%w0							\n"
	: "=g" (val)
	: "0" (val)
	);
#else
  asm volatile (" bswap	%0 "
	: "=g" (val)
	: "0" (val)
	);
#endif
#endif
  return val;
  }
#endif /* _EXT2_HAVE_ASM_SWAB */

/*
 * Those symbols are set in boot.ld, depending on EXTRASEG,
 * and on in which segment are the alloc/(s)printf functions:
 * (see MALLOC_IN_EXTRASEG define in make.h)
 */
#define malloc		library_malloc
#define free		library_free
#define realloc		library_realloc
#define sprintf		library_sprintf
#define printf		library_printf

