/* library.h */
#ifndef LIBRARY_H
#define LIBRARY_H

/* Gujin is a bootloader, it loads a Linux kernel from cold boot or DOS.
 * Copyright (C) 1999-2001 Etienne Lorrain, fingerprint (2D3AF3EA):
 *   2471 DF64 9DEE 41D4 C8DB 9667 E448 FF8C 2D3A F3EA
 * E-Mail: etienne.lorrain@masroudeau.com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifndef MAKE_H
#error I need file "make.h" included!
#endif

// just for tests:
//#define TEST_LIBINCL_H
//#define I386_MOVS

#define ASMREAD volatile  /* inline function with asm("") are const fcts */
#define ASMCONST /* asm const ("...") no more recognised by GCC-3.0 */

#define FASTCALL        __attribute__ (( regparm(1) ))

/*
 * If you do not understand this general stuff,
 * re-read your favorite C reference book...
 */
#define UINT_MAX		0xFFFFFFFFU
#define string(str)		#str
#define STRING(str)		string(str)
#define nbof(X)			(sizeof (X) / sizeof (X[0]))
#define lastof(X)		(nbof(X) - 1)
#define DIVROUNDUP(X,Y)		(((X)+(Y)-1)/(Y))
#define ROUNDUP(X,Y)		((Y) * DIVROUNDUP(X,Y))
#define ABS(X)			(((X) < 0)? -(X) : (X))
#if 1
#define MIN(X,Y)		(((X) <= (Y))? (X) : (Y))
#define MAX(X,Y)		(((X) >= (Y))? (X) : (Y))
#else
#define MIN(X,Y) \
	( (((X) < 0) && ((Y) > 0))					\
	 ? (X)								\
	 : (((X) > 0) && ((Y) < 0)) ? (Y)				\
				    : (((X) <= (Y))? (X) : (Y)))

#define MAX(X,Y) \
	( (((X) < 0) && ((Y) > 0))					\
	 ? (Y)								\
	 : (((X) > 0) && ((Y) < 0)) ? (X)				\
				    : (((X) >= (Y))? (X) : (Y))
#endif

#ifdef offsetof
#undef offsetof
#endif
#define offsetof(type, field)	((unsigned)(&(((type *)0)->field)))

#if (__GNUC__ <= 2) && (__GNUC_MINOR__ <= 97)
#define UNUSUAL(X)	X
#define USUAL(X)	X
#else
#define UNUSUAL(X)	__builtin_expect (X, 0)
#define USUAL(X)	__builtin_expect (X, 1)
#endif

/*
 * The C language is a bit subtile times to times,
 * "extern char memory[];" will not cover local variables,
 * and "extern char *memory;" can be transmitted by a register.
 */
extern char memory[];

/* ONLY to be used with cst_str as a constant inline string, i.e. "abc" */
#define STRINIT(dst, cst_str) \
    ({ memcpy (dst, cst_str, sizeof (cst_str)); dst + sizeof (cst_str) - 1; })

#define STREQL(str, cst_str) \
    memeql (str, cst_str, sizeof (cst_str))

#define STRBEGINEQL(str, cst_str) \
    memeql (str, cst_str, sizeof (cst_str) - 1)

#ifndef SETUP	/* for instboot.c */

/* Maybe you should also re-read the GCC documentation ? */
#define Memset(ptr, value, size) do { \
    union { char tmparray[size]; typeof (*ptr) noalias[0];	\
		} *unused  __attribute__((unused));		\
    *((typeof(unused)) ptr) = (typeof(*unused)) {		\
	tmparray: {[0 ... size-1] = value }};			\
    } while (0)

#define Bzero(ptr, size) do { \
    union { char tmparray[size]; typeof (*ptr) noalias[0];	\
		} *unused  __attribute__((unused));		\
    *((typeof(unused)) ptr) = ((typeof(*unused)) {});		\
    } while (0)

#define Memcpy(dst, src, size) do { \
    union { char tmparray[size]; typeof (*dst) noalias[0];	\
		} *unused  __attribute__((unused));		\
    *(typeof(unused)) dst = *(typeof(unused)) src;		\
    } while (0)

/* Unfortunately that cannot be compiled: */
#define Memcmp(dst, src, size) ({ \
    union { char tmparray[size]; typeof (*dst) noalias[0];	\
		} *unused  __attribute__((unused));		\
    *(typeof(unused)) dst != *(typeof(unused)) src;		\
    })

/* do not compile instboot.c on a 64 bit machine, or take care... */
typedef unsigned farptr;

#else /* SETUP, instboot.c */

#ifndef inline_all_stringops    /* gcc-2.96 will handle that */

#ifdef TEST_LIBINCL_H
#include "libincl.h"
#else

#if SETUP & BIG_MALLOC

#define SRCREG		"%%esi"
#define DSTREG		"%%edi"

#ifdef I386_MOVS
#define MOVS_PATCH	" ; addr32 nop "
#else
#define MOVS_PATCH	/**/
#endif

#ifndef BOOT_C
asm (
".macro movsl src=%ds:(%esi), dst=%es:(%edi)			\n"
"	.byte 0x67, 0x66, 0xA5	# movs.l %ds:(%esi),%es:(%edi)	\n"
"	" MOVS_PATCH "						\n"
".endm								\n"
".macro movsw src=%ds:(%esi), dst=%es:(%edi)			\n"
"	.byte 0x67, 0xA5	# movs.w %ds:(%esi),%es:(%edi)	\n"
"	" MOVS_PATCH "						\n"
".endm								\n"
".macro movsb src=%ds:(%esi), dst=%es:(%edi)			\n"
"	.byte 0x67, 0xA4	# movs.b %ds:(%esi),%es:(%edi)	\n"
"	" MOVS_PATCH "						\n"
".endm								\n"
"								\n"
".macro stosl src=%eax, dst=%es:(%edi)				\n"
"	.byte 0x67, 0x66, 0xAB	# stos.l %eax,%es:(%edi)	\n"
".endm								\n"
".macro stosw src=%ax, dst=%es:(%edi)				\n"
"	.byte 0x67, 0xAB	# stos.w %ax,%es:(%edi)		\n"
".endm								\n"
".macro stosb src=%al, dst=%es:(%edi)				\n"
"	.byte 0x67, 0xAA	# stos.b %al,%es:(%edi)		\n"
".endm								\n"
"								\n"
".macro scasl str=%es:(%edi), chr=%eax				\n"
"	.byte 0x67, 0x66, 0xAF	# scas.l %es:(%edi),%eax	\n"
".endm								\n"
".macro scasw str=%es:(%edi), chr=%ax				\n"
"	.byte 0x67, 0xAF	# scas.w %es:(%edi),%ax		\n"
".endm								\n"
".macro scasb str=%es:(%edi), chr=%al				\n"
"	.byte 0x67, 0xAE	# scas.b %es:(%edi),%al		\n"
".endm								\n"
"								\n"
".macro cmpsl src=%eax, dst=%es:(%edi)				\n"
"	.byte 0x67, 0x66, 0xA7	# cmps.l %es:(%edi),%ds:(%esi)	\n"
".endm								\n"
".macro cmpsw src=%ax, dst=%es:(%edi)				\n"
"	.byte 0x67, 0xA7	# cmps.w %es:(%edi),%ds:(%esi)	\n"
".endm								\n"
".macro cmpsb src=%al, dst=%es:(%edi)				\n"
"	.byte 0x67, 0xA6	# cmps.b %es:(%edi),%ds:(%esi)	\n"
".endm								\n"
"								\n"
".macro lodsl str=%es:(%edi), chr=%eax				\n"
"	.byte 0x67, 0x66, 0xAD	# lods.l %es:(%edi),%eax	\n"
".endm								\n"
".macro lodsw str=%es:(%edi), chr=%ax				\n"
"	.byte 0x67, 0xAD	# lods.w %es:(%edi),%ax		\n"
".endm								\n"
".macro lodsb str=%es:(%edi), chr=%al				\n"
"	.byte 0x67, 0xAC	# lods.b %es:(%edi),%al		\n"
".endm								\n"
"								\n"
"# same macros for ins[bwl], outs[bwl]				\n"
"								\n"
);
#endif /* BOOT_C */

#else /* BIG_MALLOC */

#define SRCREG		"%%si"
#define DSTREG		"%%di"

#ifdef I386_MOVS
#define MOVS_PATCH	" ; nop "
#else
#define MOVS_PATCH	/**/
#endif

#ifndef BOOT_C
asm (
" .macro movsl src=%ds:(%si), dst=%es:(%di)			\n"
"	.byte 0x66, 0xA5	# movs.l %ds:(%si),%es:(%di)	\n"
"	" MOVS_PATCH "						\n"
".endm								\n"
".macro movsw src=%ds:(%si), dst=%es:(%di)			\n"
"	.byte 0xA5		# movs.w %ds:(%si),%es:(%di)	\n"
"	" MOVS_PATCH "						\n"
".endm								\n"
".macro movsb src=%ds:(%si), dst=%es:(%di)			\n"
"	.byte 0xA4		# movs.b %ds:(%si),%es:(%di)	\n"
"	" MOVS_PATCH "						\n"
".endm								\n"
"								\n"
);
#endif /* BOOT_C */

#endif /* BIG_MALLOC */

/*
 * The gas macroes are not expanded if they appear after a
 * semicolon in functions memcpy,... later in this file.
 */

extern inline void _memset (void *buffer, int value, unsigned size)
  {
#if ASSEMBLY_TYPE == ASSEMBLY_NONE
  size++;
  while (--size)
      *((unsigned char *)buffer)++ = value;
#else
  asm (
"	cld	# _memset modify %2	\n"
"	rep stosb %%al,%%es:("DSTREG")	\n"
	: "+c" (size), "+D" (buffer), /* in fact %es:%edi */
	  "=m" (*memory)
	: "a" (value)
	: "cc");
#endif
  }

extern inline void _memcpy (void *dst, const void *src, unsigned size)
  {
#if ASSEMBLY_TYPE == ASSEMBLY_NONE
  size++;
  while (--size)
      *((unsigned char *)dst)++ = *((unsigned char *)src)++;
#else
  asm (
"	cld	# _memcpy modify %3	\n"
"	rep movsb %%ds:("SRCREG"),%%es:("DSTREG")" MOVS_PATCH	"\n"
	: "+c" (size), "+D" (dst), "+S" (src), "+m" (*memory)
	:
	: "cc");
#endif
  }

extern inline int _memcmp (const void *s1, const void *s2, unsigned nb)
  {
#if ASSEMBLY_TYPE == ASSEMBLY_NONE
  nb++;
  while (--nb != 0) {
      if (*(unsigned char *)s1 != *(unsigned char *)s2)
	  break;
      ((unsigned char *)s1)++;
      ((unsigned char *)s2)++;
      }
#else
  asm (
"	cld	# _memcmp depends on %3			\n"
"	repe cmpsb %%es:("DSTREG"),%%ds:("SRCREG")	\n"
	: "+c" (nb), "+S" (s1), "+D" (s2)
	: "m" (*memory)
	: "cc" );
  ((unsigned char *)s1)--;
  ((unsigned char *)s2)--;
#endif
  if (*(unsigned char *)s1 == *(unsigned char *)s2)
      return 0;
  if (*(unsigned char *)s1 > *(unsigned char *)s2)
      return 1;
    else
      return -1;
  }

extern inline int _memeql (const void *s1, const void *s2, unsigned nb)
  {
#if ASSEMBLY_TYPE == ASSEMBLY_NONE
  nb++;
  while (--nb != 0) {
      if (UNUSUAL(*(unsigned char *)s1 != *(unsigned char *)s2))
	  break;
      ((unsigned char *)s1)++;
      ((unsigned char *)s2)++;
      }
#else
  asm ("cld; repe cmpsb %%es:("DSTREG"),%%ds:("SRCREG") "
	: "+c" (nb), "+S" (s1), "+D" (s2)
	: "m" (*memory));
#endif
  return !nb;
  }

extern inline void memcpy_door_open (void *dst, const void *src, unsigned size)
  {
#if ASSEMBLY_TYPE == ASSEMBLY_NONE
  size ++;
  while (--size)
      *((unsigned char *)dst)++ = *((unsigned char *)src)++;
#else
  asm (
"	cld	# memcpy_door_open modify %3	\n"
"	rep movsb %%ds:(%%esi),%%es:(%%edi)	\n"
"	addr32 nop				\n"
	: "+c" (size), "+D" (dst), "+S" (src), "+m" (*memory)
	:
	: "cc");
#endif
  }

extern inline const char *_strnchr (const char *str, char c, unsigned size)
  {
#if ASSEMBLY_TYPE == ASSEMBLY_NONE
  while (*str++ != c && size--) ;
#else
  asm ("cld ; repne scasb %%es:("DSTREG"),%%al"
	: "+c" (size), "+D" (str) /* in fact %es:%edi */
	: "a" (c), "m" (*memory)
	: "cc");
#endif
  return str - 1;
  }

extern inline void *memcpy (void *dst, const void *src, unsigned size)
  {
  if (__builtin_constant_p (size)) {
      if (size == 1) {
	  ((unsigned char *)dst)[0] = ((const unsigned char *)src)[0];
	  return dst;
	  }
	else if (size == 2) {
	  ((unsigned short *)dst)[0] = ((const unsigned short *)src)[0];
	  return dst;
	  }
	else if (size == 3) {
	  ((unsigned short *)dst)[0] = ((const unsigned short *)src)[0];
	  ((unsigned char *)dst)[2] = ((const unsigned char *)src)[2];
	  return dst;
	  }
	else if (size == 4) {
	  ((unsigned *)dst)[0] = ((const unsigned *)src)[0];
	  return dst;
	  }
	else if (size == 5) {
	  ((unsigned *)dst)[0] = ((const unsigned *)src)[0];
	  ((unsigned char *)dst)[4] = ((const unsigned char *)src)[4];
	  return dst;
	  }
	else if (size == 6) {
	  ((unsigned *)dst)[0] = ((const unsigned *)src)[0];
	  ((unsigned short *)dst)[2] = ((const unsigned short *)src)[2];
	  return dst;
	  }
	else if (size == 8) {
	  ((unsigned *)dst)[0] = ((const unsigned *)src)[0];
	  ((unsigned *)dst)[1] = ((const unsigned *)src)[1];
	  return dst;
	  }
      }
  _memcpy (dst, src, size);
  return dst;
  }

/* This returns !memcmp(), so not the -1/0/1 system: */
extern inline unsigned memeql (const void *dst, const void *src, unsigned size)
  {
  if (__builtin_constant_p (size)) {
      if (size == 1) {
	  return ((const unsigned char *)dst)[0] == ((const unsigned char *)src)[0];
	  }
	else if (size == 2) {
	  return ((const unsigned short *)dst)[0] == ((const unsigned short *)src)[0];
	  }
	else if (size == 3) {
	  return ((const unsigned short *)dst)[0] == ((const unsigned short *)src)[0]
	      && ((const unsigned char *)dst)[2] == ((const unsigned char *)src)[2];
	  }
	else if (size == 4) {
	  return ((const unsigned *)dst)[0] == ((const unsigned *)src)[0];
	  }
	else if (size == 5) {
	  return ((const unsigned *)dst)[0] == ((const unsigned *)src)[0]
	      && ((const unsigned char *)dst)[4] == ((const unsigned char *)src)[4];
	  }
	else if (size == 6) {
	  return ((const unsigned *)dst)[0] == ((const unsigned *)src)[0]
	      && ((const unsigned short *)dst)[2] == ((const unsigned short *)src)[2];
	  }
	else if (size == 8) {
	  return ((const unsigned *)dst)[0] == ((const unsigned *)src)[0]
	      && ((const unsigned *)dst)[1] == ((const unsigned *)src)[1];
	  }
      }
  return _memeql (dst, src, size);
  }

extern inline void *memset (void *dst, int value, unsigned size)
  {
  if (__builtin_constant_p (size)) {
      if (size == 1) {
	  ((unsigned char *)dst)[0] = value;
	  return dst;
	  }
	else if (size == 2) {
	  unsigned short val = (value & 0xFF) | (value << 8);
	  ((unsigned short *)dst)[0] = val;
	  return dst;
	  }
	else if (size == 3) {
	  unsigned short val = (value & 0xFF) | (value << 8);
	  ((unsigned short *)dst)[0] = val;
	  ((unsigned char *)dst)[2] = val;
	  return dst;
	  }
	else if (size == 4) {
	  unsigned val = value & 0xFF;
	  val |= val << 8;
	  val |= val << 16;
	  ((unsigned *)dst)[0] = val;
	  return dst;
	  }
	else if (size == 5) {
	  unsigned val = value & 0xFF;
	  val |= val << 8;
	  val |= val << 16;
	  ((unsigned *)dst)[0] = val;
	  ((unsigned char *)dst)[4] = val;
	  return dst;
	  }
	else if (size == 6) {
	  unsigned val = value & 0xFF;
	  val |= val << 8;
	  val |= val << 16;
	  ((unsigned *)dst)[0] = val;
	  ((unsigned short *)dst)[2] = val;
	  return dst;
	  }
	else if (size == 8) {
	  unsigned val = value & 0xFF;
	  val |= val << 8;
	  val |= val << 16;
	  ((unsigned *)dst)[0] = val;
	  ((unsigned *)dst)[1] = val;
	  return dst;
	  }
      }
  _memset (dst, value, size);
  return dst;
  }

extern inline int memcmp (const void *s1, const void *s2, unsigned nb)
  {
  return _memcmp (s1, s2, nb);
  }

/*
 * It is easy to set a register to -1, "orl $-1,%reg"
 * which do not contain a 32 bit inline constant on
 * 80x86. Unfortunately, GCC-3.0 does not do it very
 * often...
 */
#define MAXSTRING (~0)

extern inline unsigned strlen (const char *str)
  {
  return _strnchr (str, '\0', MAXSTRING) - str;
  }

extern inline void _strncpy (char *dst, const char *src, unsigned nb)
  {
#if ASSEMBLY_TYPE == ASSEMBLY_NONE
  do {
      if (UNUSUAL(nb-- == 0)) {
	  *dst = '\0';
	  break;
	  }
      *dst++ = *src;
      } while (*src++);
#else
#if 0
  unsigned len = _strnchr (s1, '\0', nb) - s1;
#else
  unsigned len = strlen(src);
  if (len > nb)
      len = nb;
#endif
  memcpy (dst, src, len);
  dst[len] = '\0';
#endif
  }

extern inline int _strncmp (const char *s1, const char *s2, unsigned nb)
  {
#if ASSEMBLY_TYPE == ASSEMBLY_NONE
  while (*s1) {
      if (--nb == 0)
	  break;
      if (*s1 != *s2)
	  break;
      s1++; s2++;
      }
  return (*s1 > *s2)? 1 : ((*s1 < *s2)? -1 : 0);
#else
#if 0
  unsigned len = _strnchr (s1, '\0', nb) - s1 + 1;
#else
  unsigned len = strlen(s1) + 1;
  if (len > nb)
      len = nb;
#endif
  return _memcmp (s1, s2, len);
#endif
  }

extern inline const char *strchr (const char *str, char c)
  {
  const char *returned = _strnchr (str, c, strlen(str));

  if (*returned == c)
      return returned;
    else
      return 0;
  }

extern inline char *strncpy (char *dst, const char *src, unsigned nb)
  {
  _strncpy (dst, src, nb);
  return dst;
  }

extern inline char *strcpy (char *dst, const char *src)
  {
  return strncpy (dst, src, MAXSTRING);
  }

extern inline char *strcat (char *dst, const char *src)
  {
  strcpy (dst + strlen (dst), src);
  return dst;
  }

extern inline char *stradd (char *dst, const char *src)
  {
  unsigned len = strlen (src);
  memcpy (dst, src, len + 1); /* do not forget '\0' */
  return dst + len; /* points to the zero char */
  }

extern inline int strncmp (const char *s1, const char *s2, unsigned nb)
  {
  return _strncmp (s1, s2, nb);
  }

extern inline int strcmp (const char *s1, const char *s2)
  {
  return strncmp (s1, s2, MAXSTRING);
  }

extern inline int strnicmp (const char *s1, const char *s2, unsigned nb)
  {
  unsigned char tmp1 = *s1, tmp2 = *s2;

  if (!tmp1)
      return (tmp2 == 0)? 0 : 1;

  while (tmp1) {
      if (tmp1 >= 'a')
	  tmp1 -= 'a' - 'A';
      if (tmp2 >= 'a')
	  tmp2 -= 'a' - 'A';
      if (--nb == 0)
	  break;
      if (tmp1 != tmp2)
	  break;
      tmp1 = *++s1;
      tmp2 = *++s2;
      }
  return (tmp1 > tmp2)? 1 : ((tmp1 < tmp2)? -1 : 0);
  }

extern inline int stricmp (const char *s1, const char *s2)
  {
  unsigned char tmp1 = *s1, tmp2 = *s2;

  if (!tmp1)
      return (tmp2 == 0)? 0 : 1;

  while (tmp1) {
      if (tmp1 >= 'a')
	  tmp1 -= 'a' - 'A';
      if (tmp2 >= 'a')
	  tmp2 -= 'a' - 'A';
      if (tmp1 != tmp2)
	  break;
      tmp1 = *++s1;
      tmp2 = *++s2;
      }
  return (tmp1 > tmp2)? 1 : ((tmp1 < tmp2)? -1 : 0);
  }

extern inline void strlwr (char *str)
  {
  if (!str)
      return;
  while (*str) {
      if (*str >= 'A' && *str <= 'Z')
	  *str += 'a' - 'A';
      str++;
      }
  }

#endif /* TEST_LIBINCL_H */

#endif /* inline_all_stringops */

/*
 * Few standard functions:
 */
/* This one should _not_ be called with val == 0: */
__attribute__ ((const)) extern inline unsigned my_ffs (unsigned val)
  {
  unsigned result;
  asm (" bsf	%1,%0 " : "=g" (result) : "g" (val) : "cc");
  return result + 1; /* if bit 0 set, return 1 */
  }

/*
 * Self explaining name - keep it big enough for BIOS interrupts:
 * (used in malloc and DEBUG_ADDR)
 */
#define stack_safety 0x400

/*
 * these are 32bit aligned, defined by linker:
 */
extern unsigned _etext[], _srodata[], _end[], _sbss[], _edata[];
extern unsigned char deltaseg[]; /* 8 bits aligned, defined in linker file */

#if SETUP & EXTRA_SEGMENT
extern unsigned _eextra[];
extern unsigned char extraseg[];
#endif

/*
 * The linker will refuse to link this symbol, used to check
 * for compile/link time error:
 */
void __ERROR(void);

/*
 * How is your knowledge of i80x86 segments ? Would you like to
 * implement:
 * void function (int i) __attribute__ (( farcall (extraseg) )) {}
 *
 * It should produce a "fcallw" assembly instruction instead of a "calll"
 * when you call this function, and should finish the assembly body of
 * the function with "lretw" or "lretw n" when compiled with "-mrtd".
 * "extraseg" has to be a unsigned or unsigned short - it could be
 * optimised if it is constant but should be able to handle a variable
 * (or a C "constant" - known at execution time). Attribute "farcall"
 * alone could say "fcallw $cstseg,$cstoff" with the address of
 * cstseg (in code segment) stored in ".reloc" segment for loader
 * processing.
 *
 * In a perfect world, a warning could be generated if such function is
 * not a leaf function (or calls something else than a "farcall" function)
 * - because then my software would crash if the function called is a
 * hidden memcpy or memset ! But I would still like to be authorised to
 * call static functions in the right segment.
 * Are you interested ? I'll test the GCC patch !
 *
 * Also, how about to create an attribute to change the return code
 * of a function - so that the function can be called as a standart
 * one or as a far one: if the return address is < 0x10000, a "retl"
 * is done, else a "lretw" is done. Something like, after pops:
 *	  cmpw $0,2(%esp)
 *	  jne  1f
 *	  retl
 *	  1:
 *	  lretw
 * Would be usefull for malloc(), free(), and should work with "-mrtd",
 * and with or without -fomit-frame-pointer.
 *
 * If you enjoy patching GCC, how about an attribute for variables like:
 * "int variable_in_code_segment __attribute__ ((segment(cs)));" , valid
 * for cs, ds, es, fs, gs, ss ; default been ds (usual data segment).
 * While you are there, segment "io" could be defined to access I/O ports
 * transparently: "struct UART __attribute__ ((segment(io)));" - but this
 * one should be multiprocessor: a lot of processors have "Special
 * Function Registers", like the PPC (mtspr/mfspr).
 *
 * Then you can add on IA32 the "processor special" registers called
 * __segment_cs, __segment_ds, __segment_es, __segment_fs, __segment_gs,
 * (or which can be named with "explicit reg vars" extension 'asm ("gs");')
 * so that the user can do "__segment_fs = farptr >> 16;" - and that can
 * be optimised like any other variable: if the variable is set twice
 * with the same value, the second one is optimised away. Same for the
 * flags: "__flags.direction = 1;" to be optimised when set twice to 1,
 * or to be used like "if (__flags.overflow) {}".
 *
 *   Just for completeness, I would also like:
 * unsigned globalvar1 : 8;	// 255 + 1 = 0
 * unsigned globalvar2 : 16;	// 0 - 1 = 0xFFFF
 * signed   globalvar3 : 64;
 *   These reserved for later implementation of "saturating" integers:
 * unsigned globalvar1 : -8;	// 255 + 1 = 255
 * unsigned globalvar2 : -16;	// 0 - 1 = 0
 * unsigned globalvar3 : -128 ... 127;	// maybe
 *   So then:
 * typedef unsigned byte : 8;
 * typedef unsigned boolean : -1;
 *
 *   And what C++ wrongly calls "static const" in the C cleverness:
 * inline const int compile_flag = 1;
 * inline const struct { int fct1 : 1; } configuration = { };
 *
 *   And the logical:
 * if (!typeof(bcopy))
 *   memset (ptr, 0, nb);
 * if (!typeof(dirent.d_namelen))
 *   ...;
 *   And:
 * struct {...} a, b;
 * if (a != b)
 *   ...;
 *   And the CPP:
 * #if !include ("string.h")
 * #include "strings.h"
 * #endif
 *
 * BTW, how shall I write the following thing?
 *  void fct (unsigned, int);
 *  typeof (fct) stub (*) { asm ("code_for_stub"); }
 * or:
 *  typeof (fct) stub (*) = { asm ("code_for_stub"); }
 *
 * And how about reserving the word "alias" for variable declaration:
 *     alias the->complex->structure.field myfield;
 *     alias the->complex->structure.field *myptr;
 * instead of:
 *     typeof (the->complex->structure.field) myfield =
 *            the->complex->structure.field;
 *     typeof (the->complex->structure.field) *myptr =
 *            &the->complex->structure.field;
 *
 * And if someone wants to implement the "farptr" type, based on
 * the type "complex short" but with different method to add.
 *
 * And if someone would encode a way to describe to an asm ("" : :) that
 * the _content_ of a pointer is read or/and written, and not the pointer
 * itself being read or written, i.e. something like:
 * void memcpy (char *dest, const char *src, unsigned nb) {
 *    asm (" code_to_memcpy "
 *          : "+r" (nb)
 *          : "r" (dest), "r" (src)
 *          : "=" (1) / * written * /, "*" (2) / * read * /
 *          );
 *    }
 * And also a way to say for instance that %ah has to be set, not
 * necessarily the complete %eax register (code space saving).
 *
 * And when a C switch is written, when optimising for size,
 * use a (base+offset) jump (so 16 bits per 'case:') instead
 * of a full word, 32 bits per 'case:', the jump table size
 * is also important...
 *
 * And if someone would like to put another optimisation in GCC, so
 * that (at least when optimising with -Os) consecutive read/write/compare
 * to incremented pointers would be optimised:
 * struct { int a1, a2, a3; char b1, b2, b3, b4; } src;
 * struct { int a1, a2; char b1, b2, b3, b4; } dst;
 *   dst.b4 = src.b4; dst.b3 = src.b3; dst.b2 = src.b2; dst.b1 = src.b1;
 * result in:  movl src,var2
 * and:
 *   var2.a1 = var1.a1; var2.a2 = var1.a2
 * result in: rep movl ...
 * Then we would no more need memcpy(), memcmp() : just a for(;;) with
 * typed arguments. Moreover, there will be no need for functions to
 * search a char in a string (strchr/scasb), a short in a short array (scasw)
 * or a long in a long array (scasl)...
 * This scas[bwl] can be used to encode a switch with sparce cases.
 *
 * And if someone would remove the assumption that "rol $8,%%ax"
 * modify the upper bits of %%eax, so that we can do in C:
 *  extern inline unsigned swab32 (unsigned val) {
 *    union { unsigned l; unsigned short s; } tmp = { val };
 *    tmp.s = (tmp.s >> 8) | (tmp.s << 8);
 *    tmp.l = (tmp.l >> 16) | (tmp.l << 16);
 *    tmp.s = (tmp.s >> 8) | (tmp.s << 8);
 *    return tmp.l;
 *    }
 * and that is compiled to either a "bswap %reg" or three "rol $,%mem"
 *
 *
 * But the dream stopped there, I still cannot mark as clobbered in asm()
 * the segment registers nor %ebp; and the bugs I reported on GCC-2.95.1
 * are not corrected on GCC-3.0 (.align 32 of strings & structures) ...
 */
#if SETUP & EXTRA_SEGMENT
/*
 * Note that the "stub" function is NOT re-entrant, and it will be
 * very difficult to change it (modification of the ljmpw not stacked).
 * This works only if the stub function do not touch the stack,
 * so will not work without optimisation (or with too old compiler?),
 * or if the software is compiled _without_ -fomit-frame-pointer.
 *
 * This stub uses %cs so can be called from the two segments ".text"
 * and ".extra". Shall I add ",@progbits" to the ".section" ?
 *
 * Replacing the pushw/pushw/lretw by a lcallw produces self
 * modifying code, so on my i486DX2-66 the assembly "jmp 1f\n 1:"
 * has to be added before lcallw to not crash.
 */

#define EXTRASEG_STUB(fct)			\
    asm volatile (" \n "			\
	" movw	2+extraseg_adr,%%fs	\n"	\
	" popl	%%fs:1+1f		\n"	\
	" movw	%%cs,%%fs:3+1f		\n"	\
	" pushl	$1f			\n"	\
	" pushw	%%fs			\n"	\
	" pushw	%w0			\n"	\
	" lretw				\n"	\
	".section .Xextra2text,\"ax\"	\n"	\
	"1:				\n"	\
	" ljmpw $0,$0			\n"	\
	".previous # .Xtext2extra	\n"	\
	: : "p" (fct))

#define GENERATE_VOID_CODE2EXTRA_STUB(fctname, params...) \
	__attribute__ ((section (STRING(.Xtext2extra_ ## fctname))))	\
	void codeseg_ ## fctname (params) {				\
	EXTRASEG_STUB (fctname);					\
	}

#define GENERATE_CODE2EXTRA_STUB(fctname, ret, params...) \
	__attribute__ ((section (STRING(.Xtext2extra_ ## fctname))))	\
	ret codeseg_ ## fctname (params) {				\
	EXTRASEG_STUB (fctname);					\
	return 0; /* THAT IS NOT THE VALUE RETURNED */			\
	}

#define EXTRASEG_REVERSE_STUB(fct)		\
    asm volatile (" \n "			\
	" movw	2+codeseg_adr,%%fs	\n"	\
	" popl	%%fs:1+1f		\n"	\
	" movw	%%cs,%%fs:3+1f		\n"	\
	" pushl	$1f			\n"	\
	" pushw %%fs			\n"	\
	" pushw	%w0			\n"	\
	" lretw				\n"	\
	".section .Xtext2extra,\"ax\"	\n"	\
	"1:				\n"	\
	" ljmpw $0,$0			\n"	\
	".previous # .Xextra2text	\n"	\
	: : "p" (fct))

#define GENERATE_VOID_EXTRA2CODE_INDIRECT_STUB(fctname, fctadr, params...) \
	__attribute__ ((section (STRING(.Xextra2text_ ## fctname))))	\
	void extraseg_ ## fctname (params) {				\
	EXTRASEG_REVERSE_STUB (fctadr);					\
	}

#define GENERATE_EXTRA2CODE_INDIRECT_STUB(fctname, fctadr, ret, params...) \
	__attribute__ ((section (STRING(.Xextra2text_ ## fctname))))	\
	ret extraseg_ ## fctname (params) {				\
	EXTRASEG_REVERSE_STUB (fctadr);					\
	return 0; /* THAT IS NOT THE VALUE RETURNED */			\
	}

#define GENERATE_VOID_EXTRA2CODE_STUB(fctname, params...) \
	GENERATE_VOID_EXTRA2CODE_INDIRECT_STUB(fctname, fctname, params)

#define GENERATE_EXTRA2CODE_STUB(fctname, ret, params...) \
	GENERATE_EXTRA2CODE_INDIRECT_STUB(fctname, fctname, ret, params)

/*
 * Why not a dot as usual between the segment and the function name?
 * Why not a minus sign? Good question!
 * Because of the CPP warning:
 * pasting "." and "xxx" does not give a valid preprocessing token
 */
#define SEGMENT_EXTRA(fct) \
	__attribute__ ((section (STRING(.extra_ ## fct))))

#else /* EXTRA_SEGMENT */
#define GENERATE_VOID_CODE2EXTRA_STUB(fctname, params...) /**/
#define GENERATE_CODE2EXTRA_STUB(fctname, ret, params...) /**/
#define GENERATE_VOID_EXTRA2CODE_STUB(fctname, params...) /**/
#define GENERATE_EXTRA2CODE_STUB(fctname, ret, params...) /**/
#define GENERATE_VOID_EXTRA2CODE_INDIRECT_STUB(fctname, params...) /**/
#define GENERATE_EXTRA2CODE_INDIRECT_STUB(fctname, fctadr, ret, params...) /**/
#define SEGMENT_EXTRA(fct)	/**/
#endif /* EXTRA_SEGMENT */

#if (SETUP & EXTRA_SEGMENT) && defined (CALLING_FROM_EXTRASEG)
#define ASMNAME(x)	asm (STRING(extraseg_ ## x))
#else
#define ASMNAME(x)	/* call in same segment */
#endif

int printf (const char *format, ...) ASMNAME(printf);
void print (const char *msg) ASMNAME(print);
void puts (const char *msg) ASMNAME(puts);
#if !(SETUP & QUITE)
int verbose_printf (const char *format, ...) ASMNAME(printf);
void verbose_print (const char *msg) ASMNAME(print);
void verbose_puts (const char *msg) ASMNAME(puts);
#endif
int sprintf (char *dest, const char *format, ...) ASMNAME(sprintf);
#ifdef TREAT_EXCEPTION
void restore_vectors (void) ASMNAME(restore_vectors);
#endif
unsigned enable_A20 (unsigned timeout) ASMNAME(enable_A20);
unsigned get_A20 (unsigned timeout) ASMNAME(get_A20);

/*
 * The GS door opening:	 .....	I hate windows -:)
 * This stuff is not new, but this way to use it is GPL.
 */
#if BUILD_TESTER()
extern volatile unsigned segment_door_key;
#endif

unsigned segment_door (unsigned open) ASMNAME(segment_door);

void BIOS_killme (void) ASMNAME(BIOS_killme);

enum detect_memory_enum {
    mem_E801 = 0, mem_AMI, mem_Phoenix, mem_8800, mem_nvram
    };
enum detect_memory_enum detect_memory (unsigned *basemem, unsigned *extendmem)
	 ASMNAME(detect_memory);

#undef ASMNAME

/* This one is also used in DEBUG, in codeseg: */
char *prtf (char *buffer, char *endbuffer, const char *format, void **param);

#if !(SETUP & QUITE)
extern const char *const mem_name[];
#endif

/*
 * Segment playing:
 */

#define FCT_SET_SEGREG(seg) \
    extern inline void set##seg (unsigned short val) {		\
	asm volatile (" movw %w0,%%"#seg" " : : "r" (val));	\
	}

#define FCT_GET_SEGREG(seg) \
    extern inline unsigned short get##seg (void) {		\
	unsigned short returned;				\
								\
	asm ASMREAD (" movw %%"#seg",%w0 " : "=r" (returned));	\
	return returned;					\
	}

#define FCT_UPDATE_SEG(seg) \
    extern inline void update_##seg (void) {			\
	asm volatile (" movw %%"#seg":(0),%%ax " : : : "ax");	\
	}

/*
 * Segment door playing:
 */
#define FCT_TRYDOOR(seg) \
    extern inline void trydoor_peek_##seg (void) {	\
	unsigned dummy = 0x10000;			\
	asm volatile (" movl  %%"#seg":(%0),%0 ; nop "	\
			: "+a" (dummy));		\
	}

FCT_SET_SEGREG(ds)
FCT_SET_SEGREG(es)
FCT_SET_SEGREG(fs)
FCT_SET_SEGREG(gs)
FCT_SET_SEGREG(ss)

FCT_GET_SEGREG(cs)
FCT_GET_SEGREG(ds)
FCT_GET_SEGREG(es)
FCT_GET_SEGREG(fs)
FCT_GET_SEGREG(gs)
FCT_GET_SEGREG(ss)

FCT_UPDATE_SEG(ds)
FCT_UPDATE_SEG(es)
FCT_UPDATE_SEG(fs)
FCT_UPDATE_SEG(gs)
FCT_UPDATE_SEG(ss)

FCT_TRYDOOR(ds)
FCT_TRYDOOR(es)
FCT_TRYDOOR(fs)
FCT_TRYDOOR(gs)
FCT_TRYDOOR(cs)
FCT_TRYDOOR(ss)

/*
 * Get standart regs:
 */
extern inline unsigned geteax (void)
  {
  unsigned returned;
  asm ASMREAD (" # geteax " : "=a" (returned));
  return returned;
  }

extern inline unsigned getebx (void)
  {
  unsigned returned;
  asm ASMREAD (" # getebx " : "=b" (returned));
  return returned;
  }

extern inline unsigned getecx (void)
  {
  unsigned returned;
  asm ASMREAD (" # getecx " : "=c" (returned));
  return returned;
  }

extern inline unsigned getedx (void)
  {
  unsigned returned;
  asm ASMREAD (" # getedx " : "=d" (returned));
  return returned;
  }

extern inline unsigned getesi (void)
  {
  unsigned returned;
  asm ASMREAD (" # getesi " : "=S" (returned));
  return returned;
  }

extern inline unsigned getedi (void)
  {
  unsigned returned;
  asm ASMREAD (" # getedi " : "=D" (returned));
  return returned;
  }


#define FCT_GET_REG(reg) \
    extern inline unsigned get##reg (void) {			\
	unsigned returned;					\
								\
	asm ASMREAD (" movl %%"#reg",%0 " : "=g" (returned));	\
	return returned;					\
	}

FCT_GET_REG(ebp)
FCT_GET_REG(esp)

/*
 * Intel segment %%fs is considered as non preserved,
 * and segment %%gs as preserved across calls.
 * The complete software should keep %%es == %%ds
 *
 * Everywhere where a pointer is passed as an argument
 * to a function, it is assumed that %ss == %es if the
 * address is taken from a variable in stack, or
 * %ds == %es if it is taken from a global variable.
 *
 * So here, to simplify, %ss == %ds == %es.
 *
 * "farptr" type:
 * The segment is at upper address, offset at lower address,
 * "farptr" can do address calculus: +=4; ... within 64 Kbytes.
 * If this is not accepted, one can use these
 * assembly instructions to set up %es each times:
 * (after a push, you can no more access local variables)
 *
 *  " pushl	%%es
 *    rol	$16,%%edi	# changes carry
 *    mov	%%di,%%es
 *    rol	$16,%%edi
 *    ...
 *    popl	%%es " : : "D" (myfarptr)
 *
 * Then, to call it, the parameter is a "farptr" (no strong
 * type checking) and you get it using: "data_adr(&param)"
 */

typedef unsigned farptr;

__attribute__ ((const)) extern inline unsigned
farptr2linear (farptr addr)
  {
  return (addr & 0xFFFF) + ((addr >> 16) << 4);
  }

__attribute__ ((const)) extern inline farptr
linear2farptr (unsigned addr)
  {
  /* consider segment as counter of 64 Kb, used for DMA floppy overflow tests */
  return (addr & 0xFFFF) + ((addr >> 16) << 28);
  }

__attribute__ ((const)) extern inline farptr
smallestoffset (farptr addr)
  {
  return (((addr >> 16) + (addr >> 4)) << 16) + (addr & 0x0F);
  }

__attribute__ ((const)) extern inline farptr
data_adr (const void *data)
  {
  extern const unsigned dataseg_adr;

#if SETUP & BIG_MALLOC
  if ((unsigned) data > 0x10000) {
      unsigned linear = (dataseg_adr >> 12) + (unsigned) data;
      return (farptr)(((linear & 0xFFFF0) << 12) | (linear & 0x0F));
      }
    else
#endif
      return (farptr)(dataseg_adr | (unsigned) data);
  }

__attribute__ ((const)) extern inline farptr
stack_adr (const void *data)
  {
  extern const unsigned dataseg_adr;

  /* stack is max 64 K */
  return (farptr)(dataseg_adr | (unsigned) data);
  }

__attribute__ ((const)) extern inline farptr
code_adr (const void *data)
  {
  extern const unsigned codeseg_adr;

  return (farptr)(codeseg_adr | (unsigned) data);
  }

__attribute__ ((const)) extern inline farptr
extra_adr (const void *data)
  {
  extern const unsigned extraseg_adr;

  return (farptr)(extraseg_adr | (unsigned) data);
  }

__attribute__ ((const)) extern inline unsigned
linear2dsrelative (unsigned addr)
  {
  extern const unsigned dataseg_adr;

  return addr - (dataseg_adr >> 12);
  }

__attribute__ ((const)) extern inline unsigned
dsrelative2linear (unsigned addr)
  {
  extern const unsigned dataseg_adr;

  return addr + (dataseg_adr >> 12);
  }

__attribute__ ((const)) extern inline unsigned
farptr2dsrelative (farptr addr)
  {
  return linear2dsrelative (farptr2linear (addr));
  }

extern inline void fmemcpy (farptr dst, farptr src, unsigned short size)
  {
  asm (
"	pushl	%%es						\n"
"	pushl	%%esi						\n"
"	popw	%%si						\n"
"	popw	%%fs						\n"
"	pushl	%%edi						\n"
"	popw	%%di						\n"
"	popw	%%es						\n"
"	cld		# fmemcpy modify %3			\n"
"	rep movsb %%fs:(%%si),%%es:(%%di)	# no macro	\n"
"	nop							\n"
"	popl	%%es						\n"
	: "+S" (src), "+D" (dst), "+c" (size), "+m" (*memory)
	:
	: "cc"
	);
  }

extern inline void fmemset (farptr dst, char val, unsigned short size)
  {
  unsigned dummy;
  asm (
"	pushw	%%es						\n"
"	popw	%%fs						\n"
"	pushl	%3						\n"
"	popw	%%di						\n"
"	popw	%%es						\n"
"	cld		# fmemset modify %2			\n"
"	rep stosb %%al,%%es:(%%di)		# no macro	\n"
"	pushw	%%fs						\n"
"	popw	%%es						\n"
	: "=D" (dummy), "+c" (size), "=m" (*memory)
	: "g" (dst), "a" (val)
	: "cc"
	);
  }

/* Copy from the stack to a farptr: */
extern inline void smemcpy (farptr dst, const void *src, unsigned short size)
  {
#if 0	// do not compile in gzlib.c ...
  unsigned dummy;

  asm (
"	pushw	%%es						\n"
"	popw	%%fs						\n"
"	pushl	%4						\n"
"	popw	%%di						\n"
"	popw	%%es						\n"
"	cld		# smemcpy modify %3			\n"
"	rep movsb %%ss:(%%si),%%es:(%%di)	# no macro	\n"
"	nop							\n"
"	pushw	%%fs						\n"
"	popw	%%es						\n"
	: "+S" (src), "=D" (dummy), "+c" (size), "+m" (*memory)
	: "g" (dst)
	: "cc"
	);
#else
  asm (
"	pushw	%%es						\n"
"	popw	%%fs						\n"
"	pushl	%%edi						\n"
"	popw	%%di						\n"
"	popw	%%es						\n"
"	cld		# smemcpy modify %3			\n"
"	rep movsb %%ss:(%%si),%%es:(%%di)	# no macro	\n"
"	nop							\n"
"	pushw	%%fs						\n"
"	popw	%%es						\n"
	: "+S" (src), "+D" (dst), "+c" (size), "+m" (*memory)
	:
	: "cc"
	);
#endif
  }

/* Copy from a farptr to the current data area: */
extern inline void lmemcpy (void *dst, farptr src, unsigned short size)
  {
  unsigned dummy;
  asm (
"	pushl	%4						\n"
"	popw	%%si						\n"
"	popw	%%fs						\n"
"	cld		# lmemcpy modify %3			\n"
"	rep movsb %%fs:(%%si),%%es:(%%di)	# no macro	\n"
"	nop							\n"
	: "=S" (dummy), "+D" (dst), "+c" (size), "+m" (*memory)
	: "g" (src)
	: "cc"
	);
  }

/*
 * This function copies memory from either a farptr or
 * the current data segment to either a farptr or
 * the current data segment.
 * If the source is in the data segment, the upper part
 * of the farptr is null - same for destination.
 */
extern inline void
farmemcpy (unsigned char *dest, const unsigned char *src, unsigned short size)
  {
  asm (
"	pushw	%%ds		\n"
"	pushw	%%es		\n"
"	test	%4,%%esi	\n"
"	jz	1f		\n"
"	pushl	%%esi		\n"
"	popw	%%si		\n"
"	popw	%%ds		\n"
"	1:			\n"
"	test	%4,%%edi	\n"
"	jz	1f		\n"
"	pushl	%%edi		\n"
"	popw	%%di		\n"
"	popw	%%es		\n"
"	1:			\n"
"	cld	# farmemcpy modify %3	\n"
"	rep movsb %%ds:(%%si),%%es:(%%di) # no macro, ignore (e[sd]i >> 16)\n"
"	nop			\n"
"	popw	%%es		\n"
"	popw	%%ds		\n"
	: "+S" (src), "+D" (dest), "+c" (size), "+m" (*memory)
	: "ri" (0xFFFF0000)
	: "cc");
}


/*
 * more standart stuff:
 */
extern inline void short_delay (unsigned count)
  {
  asm volatile (" loopl . " : "+c" (count));
  }

extern inline unsigned char inb (unsigned short port)
  {
  unsigned char result;
  asm ASMREAD (" inb %1,%0 " : "=a" (result) : "dN" (port));
  return result;
  }

extern inline unsigned short inw (unsigned short port)
  {
  unsigned short result;
  asm ASMREAD (" inw %1,%0 " : "=a" (result) : "dN" (port));
  return result;
  }

extern inline unsigned inl (unsigned short port)
  {
  unsigned result;
  asm ASMREAD (" inl %1,%0 " : "=a" (result) : "dN" (port));
  return result;
  }

extern inline void outb (unsigned short port, unsigned char data)
  {
  asm volatile (" outb %0,%1 " : : "a" (data), "dN" (port));
  }

extern inline void outw (unsigned short port, unsigned short data)
  {
  asm volatile (" outw %0,%1 " : : "a" (data), "dN" (port));
  }

extern inline void outl (unsigned short port, unsigned data)
  {
  asm volatile (" outl %0,%1 " : : "a" (data), "dN" (port));
  }

extern inline unsigned char getnvram (unsigned char addr)
  {
  outb (0x70, addr & ~0x80); /* bit 7 cleared: NMI enabled */
  short_delay (1);
  return inb (0x71);
  }

extern inline void reset_387 (void)
  {
  outb (0xF0, 0);	/* reset copro */
  outb (0xF1, 0);	/* reset copro part 2 */
  }

extern inline void disable_nmi (void)
  {
  outb (0x70, 0x80);
  short_delay (1);
  inb (0x71);
  }

extern inline void enable_nmi (void)
  {
  outb (0x70, 0x00);
  short_delay (1);
  inb (0x71);
  }

extern inline unsigned getflags (void)
  {
  unsigned returned;

  // volatile here: see has_cpuid() !
  asm volatile (" pushfl ; popl %0 " : "=r" (returned) );
  return returned;
  }

extern inline void setflags (unsigned value)
  {
  asm volatile (" pushl %0 ; popfl " : : "r" (value) );
  }

extern inline void setTFflag (void)
  {
  setflags (getflags () | 0x100);
  }

extern inline void clearTFflag (void)
  {
  setflags (getflags () & ~0x100);
  }

extern inline unsigned short getsw (void)
  {
  unsigned short returned;

  asm ASMREAD (" smsw %0 " : "=r" (returned) );
  return returned;
  }

extern inline void setsw (unsigned short SW)
  {
  asm volatile (" lmsw %0 " : : "r" (SW) );
  }

extern inline void disable_interrupt (void)
  {
  asm volatile (" cli ");
  }

extern inline void enable_interrupt (void)
  {
  asm volatile (" sti ");
  }

extern inline unsigned interrupt_enabled (void)
  {
  return (getflags() & 0x200);
  }

/*
 * Few inline functions:
 */

extern inline unsigned long long rdtsc (void)
  {
  unsigned long long returned;

  asm volatile (" rdtsc " : "=A" (returned) ); /* real volatile! */
  return returned;
  }

__attribute__ ((const)) extern inline unsigned char
us_div_uc (unsigned short us, unsigned char uc, unsigned char *remainder)
  {
  unsigned short result;

  asm ASMCONST (" divb %1 " : "=a" (result) : "qm" (uc), "a" (us) : "cc" );

  *remainder = result >> 8;
  return result & 0xFF;
  }

__attribute__ ((const)) extern inline unsigned short
ul_div_us (unsigned ul, unsigned short us, unsigned short *remainder)
  {
  unsigned short result;

  asm ASMCONST (" divw %2 "
	: "=a" (result), "=d" (*remainder)
	: "qm" (us), "a" (ul & 0xFFFF), "d" (ul >> 16) : "cc" );

  return result;
  }

__attribute__ ((const)) extern inline unsigned
ull_div_ul (unsigned long long ull, unsigned ul, unsigned *remainder)
  {
  unsigned result;

#if 0 // one day (before GCC-4.0) someone should fix the compiler
  asm ASMCONST (" divl %3 "
	: "=a" (result), "=d" (*remainder)
	: "A" (ull), "rm" (ul) : "cc" );
#else
  asm ASMCONST (" divl %2 "
	: "=A" (ull)
	: "0" (ull), "rm" (ul) : "cc" );
  result = ull & 0xFFFFFFFFU;
  *remainder = ull >> 32;
#endif

  return result;
  }

__attribute__ ((const)) extern inline unsigned long long
ul_mul_ul (unsigned ul1, unsigned ul2)
  {
  unsigned long long ull;

  asm ASMCONST (" mul %2 "
	: "=A" (ull)
	: "a" (ul1), "rm" (ul2) : "cc" );

  return ull;
  }

__attribute__ ((const)) extern inline unsigned
divrem8 (unsigned short oper, unsigned char *rem)
  {
  if (__builtin_constant_p (oper)) {
      *rem = oper % 8;
      return oper / 8;
      }
    else {
      unsigned divider;

      asm ASMCONST (
"	aam	$8			# operand in ax		\n"
"	movb	%%al,%%cl		# remainder		\n"
"	movzbl	%%ah,%%eax		# divider		\n"
	: "=a" (divider), "=c" (*rem) : "a" (oper) );
      return divider;
      }
  }

/*
 * Processor stuff:
 */
__attribute__ ((const)) extern inline unsigned is_80486 (void)
  {
  unsigned returned, flags = getflags(),
	   mask = 1 << 18;	/* AC, bit 18 */

  setflags (flags ^ mask);
  returned = ((getflags() & mask) != (flags & mask));
  setflags (flags & ~mask);
  return returned;
  }

extern inline unsigned has_cpuid (void)
  {
  unsigned returned, flags = getflags(), mask = 1 << 21; /* ID, bit 21 */

  setflags (flags ^ mask);
  returned = ((getflags() & mask) != (flags & mask));
  setflags (flags & ~mask);
  return returned;
  }

union cpuid_string {
    struct {
	unsigned str1, str2, str3;
	} number;
    char string[12];
    };

extern inline int cpuid0 (union cpuid_string *result)
  {
  int returned; /* the max CPUID input value, 2 for Pentium+ */

  asm ( " cpuid "
	: "=a" (returned),
	  "=b" (result->number.str1),
	  "=d" (result->number.str2),
	  "=c" (result->number.str3)
	: "a" (0));
  return returned;
  }

struct cpuid_str {
    unsigned	setpping: 4;
    unsigned	model	: 4;
    unsigned	family	: 4;
    unsigned	type	: 2;
    unsigned	unused4	: 18;

    unsigned	fpu	: 1;
    unsigned	vme	: 1;
    unsigned	de	: 1;
    unsigned	pse	: 1;
    unsigned	tsc	: 1;
    unsigned	msr	: 1;
    unsigned	pae	: 1;
    unsigned	mce	: 1;
    unsigned	cx8	: 1;
    unsigned	apic	: 1;
    unsigned	unused3	: 1;
    unsigned	sep	: 1;
    unsigned	mtrr	: 1;
    unsigned	pge	: 1;
    unsigned	mca	: 1;
    unsigned	cmov	: 1;
    unsigned	pat	: 1;
    unsigned	pse36	: 1;
    unsigned	psn	: 1;
    unsigned	clfl	: 1;
    unsigned	unused2	: 1;
    unsigned	dtes	: 1;
    unsigned	acpi	: 1;
    unsigned	mmx	: 1;
    unsigned	fxsr	: 1;
    unsigned	sse	: 1;
    unsigned	sse2	: 1;
    unsigned	ss	: 1;
    unsigned	htt	: 1;
    unsigned	tm	: 1;
    unsigned	ia64	: 1;
    unsigned	unused1	: 1;
    };

extern inline struct cpuid_str cpuid1 (void)
  {
  struct cpuid_str returned;
  unsigned reserved1, reserved2;
  asm (" cpuid "
	: "=A" (returned), "=b" (reserved1), "=c" (reserved2)
	: "a" (1)
	);
  return returned;
  }

/*
 * cpuid_string: AuthenticAMD or TransmetaCPU,
 * reserved for other processors.
 */
extern inline int cpuid8000 (union cpuid_string *result)
  {
  int returned;

  asm ( " cpuid "
	: "=a" (returned),
	  "=b" (result->number.str1),
	  "=d" (result->number.str2),
	  "=c" (result->number.str3)
	: "a" (0x80000000));
  return returned;
  }

struct cpuid_extended_str {
    unsigned	setpping: 4;
    unsigned	model	: 4;
    unsigned	family	: 4;
    unsigned	unused4	: 20;

    unsigned	fpu	: 1;
    unsigned	vme	: 1;
    unsigned	de	: 1;
    unsigned	pse	: 1;
    unsigned	tsc	: 1;
    unsigned	msr	: 1;
    unsigned	pae	: 1;
    unsigned	mce	: 1;
    unsigned	cx8	: 1;
    unsigned	apic	: 1;
    unsigned	unused3	: 1;
    unsigned	sep	: 1;
    unsigned	mtrr	: 1;
    unsigned	pge	: 1;
    unsigned	mca	: 1;
    unsigned	cmov	: 1;
    unsigned	pat	: 1; /* or FCMOV */
    unsigned	pse36	: 1;
    unsigned	unused2	: 4;
    unsigned	mmxplus	: 1;
    unsigned	mmx	: 1;
    unsigned	fxsr	: 1; /* or emmx, Cyrix */
    unsigned	unused1	: 4;
    unsigned	lm	: 1;
    unsigned	E3Dnow	: 1; /* Extended 3DNow!+ */
    unsigned	S3Dnow	: 1; /* Standard 3DNow! */
    };

extern inline struct cpuid_extended_str cpuid8001 (void)
  {
  struct cpuid_extended_str returned;
  unsigned reserved1, reserved2;
  asm (" cpuid "
	: "=A" (returned), "=b" (reserved1), "=c" (reserved2)
	: "a" (0x80000001)
	);
  return returned;
  }

union cpuid_extended_string {
    struct {
	unsigned str1, str2, str3, str4;
	unsigned str5, str6, str7, str8;
	unsigned str9, str10, str11, str12;
	} number;
    char string[3 * 4 * 4];
    };

extern inline char *cpuid_estring (union cpuid_extended_string *result)
  {
  char *ptr;

  *result = (union cpuid_extended_string) {};

  asm ( " cpuid "
	: "=a" (result->number.str1),
	  "=b" (result->number.str2),
	  "=c" (result->number.str3),
	  "=d" (result->number.str4)
	: "a" (0x80000002));
  asm ( " cpuid "
	: "=a" (result->number.str5),
	  "=b" (result->number.str6),
	  "=c" (result->number.str7),
	  "=d" (result->number.str8)
	: "a" (0x80000003));
  asm ( " cpuid "
	: "=a" (result->number.str9),
	  "=b" (result->number.str10),
	  "=c" (result->number.str11),
	  "=d" (result->number.str12)
	: "a" (0x80000004));
  ptr = result->string;
  while (*ptr && *ptr == ' ')
      ptr++;
  return ptr;
  }

/**
 ** Peek and Poke: Are you old enough to remember BASIC?
 **/

/*
 * For A20 opening, before peekw can be used because
 * the segment doors are still not opened.
 * (needed at least when ASSEMBLY_DSES is defined).
 */

/* This produces bigger code:
extern inline unsigned char peekb (farptr adr)
  {
  unsigned char returned;
  if (__builtin_constant_p (adr)) {
      setfs (adr >> 16);

      asm ASMREAD (" movb %%fs:%a1,%0 "
	: "=q" (returned) : "nbSDB" (adr & 0xFFFF));
      }
    else {
#if 0
      asm ASMREAD (
"	lfs	%1,%0		\n"
"	movb	%%fs:(%0),%0	\n"
	: "=BbSD" (returned) : "m" (adr));
#else
      asm ASMREAD (
"	roll	$16,%1		\n"
"	movw	%w1,%%fs	\n"
"	roll	$16,%1		\n"
"	movb	%%fs:(%w1),%0	\n"
	: "=q" (returned) : "bSDB" (adr) : "cc");
#endif
      }
  return returned;
  }
*/

extern inline unsigned peekb_doorclosed (farptr adr)
  {
  unsigned returned;
  unsigned short scratch;

  asm (
"	pushl	%2		\n"
"	popw	%w0		\n"
"	popw	%%fs		\n"
"	movzbl	%%fs:(%w0),%1	\n"
	: "=bSDB" (scratch), "=r" (returned)
	: "g" (adr), "m" (*memory)
	);
  return returned;
  }

extern inline unsigned peekw_doorclosed (farptr adr)
  {
  unsigned returned;
  unsigned short scratch;

  asm (
"	pushl	%2		\n"
"	popw	%w0		\n"
"	popw	%%fs		\n"
"	movzwl	%%fs:(%w0),%1	\n"
	: "=bSDB" (scratch), "=r" (returned)
	: "g" (adr), "m" (*memory)
	);
  return returned;
  }

extern inline unsigned peekl_doorclosed (farptr adr)
  {
  unsigned returned;
  unsigned short scratch;

  asm (
"	pushl	%2		\n"
"	popw	%w0		\n"
"	popw	%%fs		\n"
"	movl	%%fs:(%w0),%1	\n"
	: "=bSDB" (scratch), "=r" (returned)
	: "g" (adr), "m" (*memory)
	);
  return returned;
  }

extern inline void pokeb_doorclosed (farptr adr, unsigned char value)
  {
  unsigned short scratch;
  asm (
"	pushl	%3		\n"
"	popw	%w0		\n"
"	popw	%%fs		\n"
"	movb	%2,%%fs:(%w0)	\n"
	: "=&bSDB" (scratch), "=m" (*memory)
	: "qi" (value), "g" (adr)
	);
  }

extern inline void pokew_doorclosed (farptr adr, unsigned short value)
  {
  unsigned short scratch;
  asm (
"	pushl	%3		\n"
"	popw	%w0		\n"
"	popw	%%fs		\n"
"	movw	%2,%%fs:(%w0)	\n"
	: "=&bSDB" (scratch), "=m" (*memory)
	: "qi" (value), "g" (adr)
	);
  }

extern inline void pokel_doorclosed (farptr adr, unsigned value)
  {
  unsigned short scratch;
  asm (
"	pushl	%3		\n"
"	popw	%w0		\n"
"	popw	%%fs		\n"
"	movl	%2,%%fs:(%w0)	\n"
	: "=&bSDB" (scratch), "=m" (*memory)
	: "ri" (value), "g" (adr)
	);
  }

extern inline unsigned xchl_doorclosed (farptr adr, unsigned value)
  {
  unsigned short scratch;

  asm (
"	pushl	%4		\n"
"	popw	%w0		\n"
"	popw	%%fs		\n"
"	xchgl	%%fs:(%w0),%1	\n"
	: "=bSDB" (scratch), "=r" (value), "=m" (*memory)
	: "1" (value), "g" (adr)
	);
  return value;
  }

extern inline unsigned codeseg_peekb (void *offset)
  {
#ifdef CALLING_FROM_EXTRASEG
  extern const unsigned codeseg_adr;

  return peekb_doorclosed (codeseg_adr | (unsigned)offset);
#else
  unsigned returned;

  asm (" movzbl	%%cs:%a1,%0 "
	: "=r" (returned)
	: "g" (offset), "m" (*memory)
	);
  return returned;
#endif
  }

extern inline unsigned codeseg_peekw (void *offset)
  {
#ifdef CALLING_FROM_EXTRASEG
  extern const unsigned codeseg_adr;

  return peekw_doorclosed (codeseg_adr | (unsigned)offset);
#else
  unsigned returned;

  asm (" movzwl	%%cs:%a1,%0 "
	: "=r" (returned)
	: "g" (offset), "m" (*memory)
	);
  return returned;
#endif
  }

extern inline unsigned codeseg_peekl (void *offset)
  {
#ifdef CALLING_FROM_EXTRASEG
  extern const unsigned codeseg_adr;

  return peekl_doorclosed (codeseg_adr | (unsigned)offset);
#else
  unsigned returned;

  asm (" movl	%%cs:%a1,%0 "
	: "=r" (returned)
	: "g" (offset), "m" (*memory)
	);
  return returned;
#endif
  }

extern inline void codeseg_pokeb (void *offset, unsigned char value)
  {
#ifdef CALLING_FROM_EXTRASEG
  extern const unsigned codeseg_adr;

  pokeb_doorclosed (codeseg_adr | (unsigned)offset, value);
#else
  asm (" movb	%1,%%cs:%a2 "
	: "=m" (*memory)
	: "qi" (value), "g" (offset)
	);
#endif
  }

extern inline void codeseg_pokew (void *offset, unsigned short value)
  {
#ifdef CALLING_FROM_EXTRASEG
  extern const unsigned codeseg_adr;

  pokew_doorclosed (codeseg_adr | (unsigned)offset, value);
#else
  asm (" movw	%1,%%cs:%a2 "
	: "=m" (*memory)
	: "qi" (value), "g" (offset)
	);
#endif
  }

extern inline void codeseg_pokel (void *offset, unsigned value)
  {
#ifdef CALLING_FROM_EXTRASEG
  extern const unsigned codeseg_adr;

  pokel_doorclosed (codeseg_adr | (unsigned)offset, value);
#else
  asm (" movl	%1,%%cs:%a2 "
	: "=m" (*memory)
	: "ri" (value), "g" (offset)
	);
#endif
  }

/*
 * The generic ones:
 */

extern inline unsigned char peekb (farptr adr)
  {
#if ASSEMBLY_TYPE != ASSEMBLY_DSES
  return peekb_doorclosed (adr);
#else
  return *(unsigned char *)(farptr2dsrelative (adr));
#endif /* ASSEMBLY_DSES */
  }

extern inline unsigned short peekw (farptr adr)
  {
#if ASSEMBLY_TYPE != ASSEMBLY_DSES
  return peekw_doorclosed (adr);
#else
  return *(unsigned short *)(farptr2dsrelative (adr));
#endif
  }

extern inline unsigned peekl (farptr adr)
  {
#if ASSEMBLY_TYPE != ASSEMBLY_DSES
  return peekl_doorclosed (adr);
#else
  return *(unsigned *)(farptr2dsrelative (adr));
#endif
  }

extern inline void pokeb (farptr adr, unsigned char value)
  {
#if ASSEMBLY_TYPE != ASSEMBLY_DSES
  pokeb_doorclosed (adr, value);
#else
  *(unsigned char *)(farptr2dsrelative (adr)) = value;
#endif
  }

extern inline void pokew (farptr adr, unsigned short value)
  {
#if ASSEMBLY_TYPE != ASSEMBLY_DSES
  pokew_doorclosed (adr, value);
#else
  *(unsigned short *)(farptr2dsrelative (adr)) = value;
#endif
  }

extern inline void pokel (farptr adr, unsigned value)
  {
#if ASSEMBLY_TYPE != ASSEMBLY_DSES
  pokel_doorclosed (adr, value);
#else
  *(unsigned *)(farptr2dsrelative (adr)) = value;
#endif
  }

extern inline unsigned xchgl (farptr adr, unsigned value)
  {
#if ASSEMBLY_TYPE != ASSEMBLY_DSES
  return xchl_doorclosed (adr, value);
#else
  unsigned tmp = *(unsigned *)(farptr2dsrelative (adr));

  *(unsigned *)(farptr2dsrelative (adr)) = value;
  return tmp;
#endif
  }

#if ASSEMBLY_TYPE == ASSEMBLY_DSES
extern inline void pokell (farptr adr, unsigned long long value)
  {
  *(unsigned long long *)(farptr2dsrelative (adr)) = value;
  }

extern inline unsigned long long peekll (farptr adr)
  {
  return *(unsigned long long *)(farptr2dsrelative (adr));
  }
#endif

/*
 * setjmp/ljmp:
 */
typedef struct {
    unsigned esi, edi, ebp, ebx;
    unsigned short gs, fs, es, ds;
    unsigned short flags, ip, cs, sp, ss;
    } __attribute__ ((packed)) setjmp_buf;

extern inline void longjmp (setjmp_buf *ptr)
  {
  asm volatile (
"	cli						\n"
"	movw	%%cs,%%dx				\n"
"	addw	$deltaseg,%%dx				\n"
"	movw	%%dx,%%ss				\n"
"	movzwl	%%ax,%%esp				\n"
"	popl	%%esi					\n"
"	popl	%%edi					\n"
"	popl	%%ebp					\n"
"	popl	%%ebx					\n"
"	popw	%%gs					\n"
"	popw	%%fs					\n"
"	popw	%%es					\n"
"	popw	%%ds					\n"
"	popw	%%dx					\n"
"	popl	%%eax					\n"
"	lss	(%%esp),%%sp				\n"
"	pushw	%%dx					\n"
"	pushl	%%eax		# %%eax not null	\n"
"	iretw						\n"
	: "+a" (ptr) : : "ecx" , "edx", "memory", "cc");
  }

extern inline unsigned setjmp (setjmp_buf *ptr)
  {
  asm volatile (
"	movl	%%esi,(%%eax)				\n"
"	movl	%%edi,4(%%eax)				\n"
"	movl	%%ebp,8(%%eax)				\n"
"	movl	%%ebx,12(%%eax)				\n"
"	movw	%%gs,16(%%eax)				\n"
"	movw	%%fs,18(%%eax)				\n"
"	movw	%%es,20(%%eax)				\n"
"	movw	%%ds,22(%%eax)				\n"
"	pushfw						\n"
"	popw	%%dx					\n"
"	movw	%%dx,24(%%eax)				\n"
"	movw	$setjump_%=,26(%%eax)			\n"
"	movw	%%cs,28(%%eax)				\n"
"	movw	%%sp,30(%%eax)				\n"
"	movw	%%ss,32(%%eax)				\n"
"	xor	%%eax,%%eax	# %%eax null		\n"
"	setjump_%=:					\n"
	: "+a" (ptr)
	: : /* if eax != 0 */ "ecx", "edx", "memory");
  return (unsigned) ptr;
  }

/*
 * Basic protected mode segment description:
 */

typedef struct {
    unsigned	limit : 16;
    unsigned	base : 24;
    unsigned	accessed : 1;			/* if 1 */
    unsigned	writable : 1;			/* if 1 */
    unsigned	expansion_direction : 1;	/* down if 1 */
    unsigned	cste_10 : 2;
    unsigned	dpl : 2;
    unsigned	present : 1;
    unsigned	limit_msb : 4;
    unsigned	available : 1;
    unsigned	cste_0 : 1;
    unsigned	big : 1;
    unsigned	granularity : 1;
    unsigned	base_msb : 8;
    } __attribute__ ((packed)) dataseg_type;

typedef struct {
    unsigned	limit : 16;
    unsigned	base  : 24;
    unsigned	accessed : 1;
    unsigned	readable : 1;
    unsigned	conforming : 1;
    unsigned	cste_11 : 2;
    unsigned	dpl : 2;
    unsigned	present : 1;
    unsigned	limit_msb : 4;
    unsigned	available : 1;
    unsigned	cste_0 : 1;
    unsigned	deflt : 1;
    unsigned	granularity : 1;
    unsigned	base_msb : 8;
    } __attribute__ ((packed)) codeseg_type;

typedef struct {
    unsigned	limit : 16;
    unsigned	base  : 24;
    enum { reserved0 = 0,
	TSS_16bit_avail, LDT, TSS_16bit_busy, callgate_16bit,
	taskgate, interruptgate_16bit, trapgate_16bit,
	reserved8,
	TSS_32bit_avail, reserved10, TSS_32bit_busy, callgate_32bit,
	reserved13, interruptgate_32bit, trapgate_32bit
	} type : 4;
    unsigned	cste_0 : 1;
    unsigned	dpl : 2;
    unsigned	present : 1;
    unsigned	limit_msb : 4;
    unsigned	reserved : 3;
    unsigned	granularity : 1;
    unsigned	base_msb : 8;
    } __attribute__ ((packed)) systemseg_type;

typedef union {
    dataseg_type	data;
    codeseg_type	code;
    systemseg_type	system;
    } seg_type;

typedef struct {
    unsigned	protected_mode : 1;
    unsigned	wait_fwait_trapped : 1;
    unsigned	fpu_trapped : 1;
    unsigned	task_switch : 1;
    unsigned	cste_1 : 1;
    unsigned	internal_fpu_report : 1;	/* when set */
    unsigned	reserved1 : 10;
    unsigned	write_protect : 1;	/* supervisor to user read only */
    unsigned	reserved2 : 1;
    unsigned	alignment_check : 1;	/* enabled if set, CPL=3, protected */
    unsigned	reserved3 : 10;
    unsigned	no_writethrough : 1;	/* disabled if set */
    unsigned	no_caching : 1;		/* disabled if set */
    unsigned	paging : 1;		/* enabled if set */
    } __attribute__ ((packed)) CR0_type;

typedef struct {
    unsigned short limit;
    unsigned       base;
    } __attribute__ ((packed)) gdt_idt_param_type;

/* lgdt produces an exception with DOSEMU... anyway you cannot
   open the A20 gate with DOSEMU - at least I think so. */
/* 32 bit (and not 286 24 bits) forced by the use of addr32 : */
extern inline void set_gdt (gdt_idt_param_type *gdtptr)
  {
  asm volatile (" lgdtl %a0 " : /* no output */ : "p" (gdtptr));
  }

extern inline void set_gdt_24 (gdt_idt_param_type *gdtptr)
  {
  asm volatile (" lgdtw %a0 " : /* no output */ : "p" (gdtptr));
  }

extern inline void get_gdt (gdt_idt_param_type *gdtptr)
  {
  asm volatile (" sgdtl %a0 " : /* mem output */ : "p" (gdtptr) : "memory");
  }

extern inline void set_idt (gdt_idt_param_type *idtptr)
  {
  asm volatile (" lidtl %a0 " : /* no output */ : "p" (idtptr));
  }

extern inline void set_idt_24 (gdt_idt_param_type *idtptr)
  {
  asm volatile (" lidtw %a0 " : /* no output */ : "p" (idtptr));
  }

extern inline void get_idt (gdt_idt_param_type *idtptr)
  {
  asm volatile (" sidtl %a0 " : /* mem output */ : "p" (idtptr) : "memory");
  }

/* Not sure the following is working/is usable, but it is here: */
extern inline unsigned loadseglimit (unsigned descriptor)
  {
  unsigned returned;

  asm volatile (" lsll %1,%0 " : "=rm" (returned) : "r" (descriptor * 8));
  return returned;
  }

extern inline CR0_type get_CR0 (void)
  {
  CR0_type returned;

  asm ASMREAD (" mov %%cr0,%0 " : "=q" (returned));
  return returned;
  }

extern inline void set_CR0 (CR0_type CR0)
  {
  asm volatile (" mov %0,%%cr0 " : /* no output */ : "q" (CR0));
  }

#define GETREG(regname) \
    extern inline unsigned get_ ## regname (void) {		\
	unsigned returned;					\
	asm (" mov %%" #regname ",%0 " : "=r" (returned));	\
	return returned;					\
	}

#define SETREG(regname) \
    extern inline void set_ ## regname (unsigned val) {		\
	asm volatile (" mov %0,%%" #regname : : "r" (val));	\
	}

GETREG(CR2)
GETREG(CR3) SETREG(CR3)
GETREG(CR4) SETREG(CR4)

SETREG(DR0) GETREG(DR0)
SETREG(DR1) GETREG(DR1)
SETREG(DR2) GETREG(DR2)
SETREG(DR3) GETREG(DR3)
SETREG(DR4) GETREG(DR4)
SETREG(DR5) GETREG(DR5)
SETREG(DR6) GETREG(DR6)
SETREG(DR7) GETREG(DR7)

/*
 * I do the following my way, is that how real programmers
 * do it ? I wonder why everybody want to use assembly here...
 */

#if 0
extern inline unsigned protected_mode (void)
  {
  CR0_type CR0 = get_CR0();

  CR0.protected_mode = 1;
  set_CR0 (CR0);

  if (get_CR0().protected_mode != 0)
      return 0;
    else
      return 1; /* error */
  }
#else
extern inline unsigned protected_mode (void)
  {
  setsw (getsw() | 1);

  if ((getsw() & 1) != 0)
      return 0;
    else
      return 1; /* error */
  }
#endif

extern inline unsigned real_mode (void)
  {
  CR0_type CR0 = get_CR0();

  CR0.protected_mode = 0;
  set_CR0 (CR0);

  if (get_CR0().protected_mode == 0)
      return 0;
    else
      return 1; /* error */
  }

/*
 * The cache playing:
 */
extern inline CR0_type disable_cache (void)
  {
  CR0_type CR0 = get_CR0(), save_CR0;

  save_CR0 = CR0;

  if (CR0.no_caching == 0) { /* else 80386 or cache already disabled */
      CR0.no_writethrough = 1;
      CR0.no_caching = 1;
      set_CR0 (CR0);
      }
  return save_CR0;
  }

extern inline void enable_cache (CR0_type save_CR0)
  {
  if (save_CR0.no_caching == 0) { /* else 80386 or cache already disabled */
      set_CR0 (save_CR0);
      }
  }

extern inline void force_cache (void *adr, unsigned len)
  {
  unsigned unused, modified, change;

  len /= 4;
  asm volatile (
	" rep lodsl  %%ds:(%1),%0"
	: "=a" (unused), "=S" (modified), "=c" (change)
	: "S" (adr), "c" (len)
	);
  asm volatile (
	" rep lodsl  %%ds:(%1),%0"
	: "=a" (unused), "=S" (modified), "=c" (change)
	: "S" (adr), "c" (len)
	);
  asm volatile (
	" rep lodsl  %%ds:(%1),%0"
	: "=a" (unused), "=S" (modified), "=c" (change)
	: "S" (adr), "c" (len)
	);
  asm volatile (
	" rep lodsl  %%ds:(%1),%0"
	: "=a" (unused), "=S" (modified), "=c" (change)
	: "S" (adr), "c" (len)
	);
  }

#endif  /* SETUP, instboot.c */
#endif /* LIBRARY_H */

