/* malloc.h */
#ifndef MALLOC_H
#define MALLOC_H

/* Gujin is a bootloader, it loads a Linux kernel from cold boot or DOS.
 * Copyright (C) 1999-2001 Etienne Lorrain, fingerprint (2D3AF3EA):
 *   2471 DF64 9DEE 41D4 C8DB 9667 E448 FF8C 2D3A F3EA
 * E-Mail: etienne.lorrain@masroudeau.com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

/*
 * Dynamic memory management - over-simplified !
 *
 * The fact that it is "over-simplified" does not mean that
 * it is not covered by GPL: if you ever include [a part of]
 * this code, or even [a part of] its design into a private software,
 * the resulting code immediately goes to GPL and you have to
 * distribute the complete source code of your product.
 *
 * Using "allocmem *" instead of "void *" to show
 * their alignment. No "FASTCALL" possible because
 * they are used in libe2fs...
 */

/*
 * This configures the memory manager:
 */
#define MALLOC_ARRAY_SIZE	50
#define MALLOC_QUANTUM		32	/* bytes */
#define MALLOC_WASTE		2	/* in quantum */

#if SETUP & BIG_MALLOC
#define MALLOC_SIZE	0x60000U /* 4*64 Kbytes for the pool */
#define MALLOC_BASE	0x10000U /* pool just after the top of the stack */
#endif

/*
 * This is the type of memory managed by malloc:
 */
typedef void allocmem __attribute__((aligned(MALLOC_QUANTUM)));

/*
 * This is the trick to have those functions in one segment
 * or the other, transparently - and at compile time.
 */
#if (SETUP & EXTRA_SEGMENT) && \
    (defined (MALLOC_IN_EXTRASEG) ^ defined (CALLING_FROM_EXTRASEG))
#ifdef CALLING_FROM_EXTRASEG
#define ASMNAME(x)	asm (STRING(extraseg_ ## x))
#else
#define ASMNAME(x)	asm (STRING(codeseg_ ## x))
#endif /* CALLING_FROM_EXTRASEG */
#else
#define ASMNAME(x)	/* call in same segment */
#endif

allocmem *malloc (unsigned size) ASMNAME(malloc) ;
void free (allocmem *ptr) ASMNAME(free) ;
allocmem *realloc (allocmem *ptr, unsigned size) ASMNAME(realloc);

#if SETUP & MINIDEBUG
void report_malloc (void) ASMNAME(report_malloc) ;
#endif

#undef ASMNAME

/*
 * Some inline functions, in the caller segment:
 */
extern inline void init_malloc (void)
  {
  extern unsigned end_malloc;

#if SETUP & BIG_MALLOC
  end_malloc = MALLOC_BASE;
#else
  end_malloc = MALLOC_QUANTUM
		* DIVROUNDUP((unsigned)_edata, MALLOC_QUANTUM);
#endif
  }

extern inline unsigned usedstack (void)
  {
#if SETUP & BIG_MALLOC
  unsigned *stkptr = _edata;
#else
  extern unsigned end_malloc;

  unsigned *stkptr = (unsigned *)end_malloc;
#endif

  while (*stkptr == 0)
      stkptr++;

  return 0x10000 - ((unsigned)stkptr);
  }

extern inline unsigned totalstack (void)
  {
#if SETUP & BIG_MALLOC
  unsigned *stkptr = _edata;
#else
  extern unsigned end_malloc;

  unsigned *stkptr = (unsigned *)end_malloc;
#endif

  return 0x10000 - ((unsigned)stkptr);
  }

extern inline unsigned freemem (void)
  {
  extern unsigned end_malloc;

#if SETUP & BIG_MALLOC
  return MALLOC_SIZE - ((unsigned)end_malloc - MALLOC_BASE);
#else
  return getesp() - stack_safety - (unsigned)end_malloc;
#endif
  }

#endif /* MALLOC_H */

