/* messages.h */

/* Gujin is a bootloader, it loads a Linux kernel from cold boot or DOS.
 * Copyright (C) 1999-2001 Etienne Lorrain, fingerprint (2D3AF3EA):
 *   2471 DF64 9DEE 41D4 C8DB 9667 E448 FF8C 2D3A F3EA
 * E-Mail: etienne.lorrain@masroudeau.com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "make.h"

/**
 ** The Linux command line parameter:
 **/
#if LANGUAGE == ENGLISH
#define LANG_CMDLINE	"LANG=en_US"
#elif LANGUAGE == FRENCH
#define LANG_CMDLINE	"LANG=fr"
#else
    /*
     * Nothing added to cmdline, but
     * still defaults to english
     */
#endif

/**
 ** Completely untested / maybe useless / not a message:
 ** A string to send at initialisation of a serial line
 ** interface, to initialise a modem and make it dial a
 ** number - a kind of ATDT string.
 ** It could also initialise the terminal in VT compatible mode?
 ** Undefine it to disable the feature.
 **/
#if USER_SUPPORT & SERIAL_SUPPORT
#define SERIAL_INIT_MODEM_STRING "<String to initialise modem>\r\n"
#endif

/**
 ** The strings without language:
 **/
#define CRLF            "\r\n"
#define DOT_CRLF        "."CRLF
#define MSG_BEEP	"\007"

/**
 ** Use this pattern to add language:
 **/
#if LANGUAGE == FRENCH
#elif LANGUAGE == DUTCH
#elif LANGUAGE == SPANISH
#elif LANGUAGE == ITALIAN
#elif LANGUAGE == PORTUGESE
#elif LANGUAGE == SWEDISH
#elif LANGUAGE == FINNISH
#else /* LANGUAGE == ENGLISH */
#endif

/**
 ** The real short messages:
 **/
#if LANGUAGE == FRENCH
#define MSG_PANIC		"PANIQUE: "
#define MSG_ERROR               "ERREUR: "
#define MSG_WARNING             "Attention: "
#define MSG_INFO                "Info: "
#define MSG_FULL_STOP           "fin."
#define MSG_WITHOUT             "sans"
#define MSG_ON                  "sous"
#define MSG_NONE                "aucun"
#define MSG_TRYING_COLON        "essai: "
#define MSG_COLON_OK            ": OK."
#define MSG_COLON_FAILED        ": rat."
#define MSG_EXITING		"Sortie..."
#else /* LANGUAGE == ENGLISH */
#define MSG_PANIC		"PANIC: "
#define MSG_ERROR               "ERROR: "
#define MSG_WARNING             "Warning: "
#define MSG_INFO                "Info: "
#define MSG_FULL_STOP           "stop."
#define MSG_WITHOUT             "without"
#define MSG_ON                  "on"
#define MSG_NONE                "none"
#define MSG_TRYING_COLON        "trying: "
#define MSG_COLON_OK            ": OK."
#define MSG_COLON_FAILED        ": failed."
#define MSG_EXITING		"Exiting..."
#endif

/**
 ** The real long messages:
 **/

/*
 * General messages:
 */
#if LANGUAGE == FRENCH
#define PRESS_KEY_TO_CONTINUE   CRLF "Pressez une touche pour continuer"
#else /* LANGUAGE == ENGLISH */
#define PRESS_KEY_TO_CONTINUE   CRLF "Press a key to continue"
#endif

#if LANGUAGE == FRENCH
#define PRESS_PAUSE_TO_STOP_DISPLAY \
CRLF"Appuyer sur \"Pause\" pour figer temporairement l'cran...\r"
#else /* LANGUAGE == ENGLISH */
#define PRESS_PAUSE_TO_STOP_DISPLAY \
CRLF"Press \"Pause\" if you want to temporary stop display...\r"
#endif

#if LANGUAGE == FRENCH
#define PANIC_VGA_SUPPORT_NOT_EQUAL \
MSG_PANIC "attributs vga_support non identiques!" CRLF
#else /* LANGUAGE == ENGLISH */
#define PANIC_VGA_SUPPORT_NOT_EQUAL \
MSG_PANIC "vga_support attributes not in sync!" CRLF
#endif

/*
 * Disk messages:
 */
#if LANGUAGE == FRENCH
#define BIOS_DISK_X     "BIOS 0x%X"
#define EBIOS_DISK_X    "EBIOS 0x%X"
#define DOS_DISK_C      "DOS %c:"
#define IDECHS_DISK_X_s "IDE/CHS %s@0x%X"
#define IDELBA_DISK_X_s "IDE/LBA %s@0x%X"
#define IDE_MASTER      "matre"
#define IDE_SLAVE       "esclave"
#define DISKNAME_ERROR  "erreur diskname!"
#else /* LANGUAGE == ENGLISH */
#define BIOS_DISK_X     "BIOS 0x%X"
#define EBIOS_DISK_X    "EBIOS 0x%X"
#define DOS_DISK_C      "DOS %c:"
#define IDECHS_DISK_X_s "IDE/CHS %s@0x%X"
#define IDELBA_DISK_X_s "IDE/LBA %s@0x%X"
#define IDE_MASTER      "master"
#define IDE_SLAVE       "slave"
#define DISKNAME_ERROR  "diskname error!"
#endif

#if LANGUAGE == FRENCH
#define DISK_NAME_SIZE_NBPARTITION	"    disque %u: %s, taille %u %s, %u partitions." CRLF
#else /* LANGUAGE == ENGLISH */
#define DISK_NAME_SIZE_NBPARTITION	"    disk %u: %s, size %u %s, %u partitions." CRLF
#endif

#if LANGUAGE == FRENCH
#define INFO_DISK_X_BIOS_EBIOS_INTERFACE_BUGGY_EXCLUDED \
    MSG_INFO "Disque 0x%X, interface BIOS & EBIOS bogues, exclus." CRLF
#define INFO_DISK_X_EXTENDED_BIOS_WITHOUT_EXTENDED_FCT \
    MSG_INFO "Disque 0x%X avec EBIOS mais sans fonctions tendues,"
#define USING_MULT_U_BIOS \
    " utilise mult(%u) BIOS." CRLF
#define USING_BIOS \
    " utilise BIOS." CRLF
#else /* LANGUAGE == ENGLISH */
#define INFO_DISK_X_BIOS_EBIOS_INTERFACE_BUGGY_EXCLUDED \
    MSG_INFO "Disk 0x%X BIOS & EBIOS interface buggy, excluded." CRLF
#define INFO_DISK_X_EXTENDED_BIOS_WITHOUT_EXTENDED_FCT \
    MSG_INFO "Disk 0x%X with EBIOS but without extended function,"
#define USING_BIOS \
    " using BIOS." CRLF
#define USING_MULT_U_BIOS \
    " using mult(%u) BIOS." CRLF
#endif

#if LANGUAGE == FRENCH

#define WARNING_LINUX_HOSTILE_PARTITION_SCHEME_ON_S \
    MSG_WARNING "Systme de partition hostile  Linux sur %s," MSG_FULL_STOP
#define INFO_ADJUSTING_PARAMETERS_FROM_BOOTSECTOR_S \
    MSG_INFO "adjustement des paramtres suite au bootsecteur %s." CRLF
#define WARNING_EXTENDED_PARTITION_BLOCK_NOT_FOUND_AT_U \
    MSG_WARNING "bloc de partition tendue non trouv  LBA = %u, "
#define S_PARTITION_STARTING_AT_U_OVERLAP_DETECTED \
    MSG_ERROR "%s chevauchement detect sur la partition dbutant  %u !" CRLF
#define S_S_LAST_PARTITION_OVER_DISK_LIMIT_DETECTED_U_KB \
    "%s%s la dernire partition dborde du disque (%u Ko)!" CRLF

#else /* LANGUAGE == ENGLISH */

#define WARNING_LINUX_HOSTILE_PARTITION_SCHEME_ON_S \
    MSG_WARNING "Linux hostile partition scheme on %s," MSG_FULL_STOP
#define INFO_ADJUSTING_PARAMETERS_FROM_BOOTSECTOR_S \
    MSG_INFO "adjusting parameters from bootsector %s." CRLF
#define WARNING_EXTENDED_PARTITION_BLOCK_NOT_FOUND_AT_U \
    MSG_WARNING "Extended partition block not found at LBA = %u, "
#define S_PARTITION_STARTING_AT_U_OVERLAP_DETECTED \
    MSG_ERROR "%s partition starting at %u, overlap detected!" CRLF
#define S_S_LAST_PARTITION_OVER_DISK_LIMIT_DETECTED_U_KB \
    "%s%s last partition over disk limit detected (%u Kb)!" CRLF

#endif


#if LANGUAGE == FRENCH

#define DETECTING_DISKS \
CRLF"Dtection des disques:"    CRLF

#define HAVE_FOUND_U_DISKS \
"%u disques trouvs:"   CRLF

#define FOUND_D_WAY_TO_BOOT_U_INITRD \
"%d faons de booter le systme trouves (et %u disques RAM initiaux)." CRLF

#else /* LANGUAGE == ENGLISH */

#define DETECTING_DISKS \
CRLF"Detecting disks:"  CRLF

#define HAVE_FOUND_U_DISKS \
"Have found %u disks:"  CRLF

#define FOUND_D_WAY_TO_BOOT_U_INITRD \
"Have found %d ways to boot the system (and %u initial RAM disks)." CRLF

#endif

/*
 * User Interface messages:
 */
#if LANGUAGE == FRENCH

#define NON_EDID_SCREEN CRLF \
"Votre moniteur/carte vido ne dclarent pas leur frquence maximum de" CRLF \
"fonctionnement. C'est  _vous_ de vrifier, avec la documentation de"  CRLF \
"votre cran, si vous pouvez augmenter la rsolution au dessus du mode" CRLF \
"par dfaut. Il est possible de dtruire certain (habituellement trs"  CRLF \
"anciens) moniteurs en demandant au BIOS un mode vido avec une trop"   CRLF \
"grande dfinition."

#else /* LANGUAGE == ENGLISH */

#define NON_EDID_SCREEN CRLF \
"Your video card/monitor do not report their maximum frequencies."      CRLF \
"It is up to _you_ to check - with your monitor documentation -"        CRLF \
"if you can increase the screen definition over the default,"           CRLF \
"it can physically damage (usually very old) monitors to ask the"       CRLF \
"video BIOS to switch to a very high definition video mode."

#endif

#if LANGUAGE == FRENCH
#define VIDEO_CARD_CHANGED	"Nouvelle carte vido, paramtres VGA r-initialiss."
#define SCREEN_INTERFACE	"Interface cran: "
#define UNKNOWN_SCREEN          "inconnu." CRLF
#define BLACKWHITE_SCREEN       "noir et blanc." CRLF
#define MONOCHROME_SCREEN       "monochrome." CRLF
#define COLOR_SCREEN            "couleur." CRLF
#define COLOR_SCREEN_TYPE_X     "couleur, type 0x%X." CRLF
#define EDID_TYPE		"        " \
	"EDID: %ux%u @ %u Hz sur un moniteur de %u pouces visibles." CRLF
#else /* LANGUAGE == ENGLISH */
#define VIDEO_CARD_CHANGED	"New video card, VGA parameters reseted."
#define SCREEN_INTERFACE	"Screen interface: "
#define UNKNOWN_SCREEN		"unknown screen." CRLF
#define BLACKWHITE_SCREEN	"B&W screen." CRLF
#define MONOCHROME_SCREEN	"monochrome screen." CRLF
#define COLOR_SCREEN		"color screen." CRLF
#define COLOR_SCREEN_TYPE_X	"color screen type 0x%X." CRLF
#define EDID_TYPE		"        " \
	"EDID: %ux%u @ %u Hz on a %u inches viewable monitor." CRLF
#endif

/*
 * Menu messages:
 */
#if LANGUAGE == FRENCH
#define TOPLINE_XSSSU		"mode 0x%X: %s %s%s; %u modes."
#define MSG_GRAPHIC		"graphique"
#define MSG_TEXT		"texte"
#define BOOT_DEVICE_READ_ONLY   "Disque(tte) de boot en lecture seul(e)."
#else /* LANGUAGE == ENGLISH */
#define TOPLINE_XSSSU		"mode 0x%X: %s %s%s; %u modes."
#define MSG_GRAPHIC		"graphic"
#define MSG_TEXT		"text"
#define BOOT_DEVICE_READ_ONLY   "Boot device read only."
#endif

#if LANGUAGE == FRENCH
#define MENU_RESOLUTION         "Rsolution: "
#define MENU_DISPLAY_RESOLUTION "%ux%u "
#define MENU_COLORS             "Couleurs: "
#define MENU_DISPLAY_COLORS     "%u bpp "
#define MENU_ALTERNATE          "Autre <Tab>"
#define MENU_TO_TEXT            "va en texte <.>"
#define MENU_TO_GRAPHIC         "va en graphique <.>"
#define MENU_SETUP		"Setup < >"
#define MENU_APPLICATION	"Appli < >"

#define HOME_KEYNAME            "<Dbut>"
#define END_KEYNAME             "<Fin>"
#define PAGEUP_KEYNAME          "<Page Prcdente>"
#define PAGEDOWN_KEYNAME        "<Page Suivante>"
#define NOT_EVEN_ONE_SELECTABLE_ITEM	"Pas une seule slection disponible !"

				       /*may be displayed on a 40 columns screen:*/
				       /*1234567890123456789012345678901234567890*/
#define SMENU_verbose			"bavard"
#define SMENU_lock_bpp			"BPP vrouill quand +/-"
#define SMENU_lock_text_graphic		"texte/graphique vrouill quand +/-"
#define SMENU_force_textmode		"force dernier mode texte au dmarrage"
#define SMENU_menu_with_disk		"affiche les disques dans le menu"
#define SMENU_menu_with_partition	"affiche les partitions dans le menu"
#define SMENU_VGA_interface		"VGA interface"
#define SMENU_refuse_8x14font		"refuse la police 8x14"
#define SMENU_VESA_interface		"VESA interface"
#define SMENU_enable_VESA_hardwin	"VESA dtection I/O fentre"
#define SMENU_VESA2_interface		"VESA linaire interface"
#define SMENU_enable_joystick		"enable Joystick"
#define SMENU_search_disk_mbr		"cherche les MBR des disques"
#define SMENU_prefer_enddisk_mbr	"prfre les MBR en fin de disque"
#define SMENU_search_part_mbr		"cherche les MBR des partitions"
#define SMENU_search_topdir_kernel	"cherche les noyaux dans la racine"
#define SMENU_search_subdir_kernel	"cherche les noyaux dans le repertoire"
#define SMENU_probe_bios_floppy_disk	"dtecte les disquettes BIOS"
#define SMENU_probe_bios_hard_disk	"dtecte les disques durs BIOS"
#define SMENU_probe_ide_disk		"dtecte les disques IDE"
#define SMENU_probe_atapi_cdrom		"dtecte les CDROM/ATAPI (non implemente)"
#define SMENU_probe_dos_disk		"dtecte les disques DOS"

#define SMENU_PROBE_VESA		"analyse les modes VESA 0..0x7F"
#define SMENU_PROBE_VGA_0_13		"analyse les modes VGA  0..0x13"
#define SMENU_PROBE_VGA_14_127		"analyse les modes VGA 0x14..0x7F"
#define SMENU_RESET_VGA_MODES		"rinitialise les modes VGA/VESA 0..0x7F"
#define SMENU_RESET_ALL_PARAMS		"rinitialise tous les paramtres"
#define SMENU_POWER_DOWN		"teind l'ordinateur"
#define SMENU_FAIL_BOOT			"dclare dmarrage rat"
#define SMENU_SET_TIMEOUT_AUTOLOAD	"configure timeout auto-dmarrage"

#define MENU_UNKNOWN_KEY	"Commande inconnue (code touche 0x%X)."
#else /* LANGUAGE == ENGLISH */
#define MENU_RESOLUTION         "Resolution: "
#define MENU_DISPLAY_RESOLUTION "%ux%u "
#define MENU_COLORS             "Colors: "
#define MENU_DISPLAY_COLORS     "%u bpp "
#define MENU_ALTERNATE          "Alternate <Tab>"
#define MENU_TO_TEXT            "go to text <.>"
#define MENU_TO_GRAPHIC         "go to graphic <.>"
#define MENU_SETUP		"Setup < >"
#define MENU_APPLICATION	"Appli < >"

#define HOME_KEYNAME            "<Home>"
#define END_KEYNAME             "<end>"
#define PAGEUP_KEYNAME          "<PageUp>"
#define PAGEDOWN_KEYNAME        "<PageDown>"
#define NOT_EVEN_ONE_SELECTABLE_ITEM	"Not even one selectable item!"

				       /*may be displayed on a 40 columns screen:*/
				       /*1234567890123456789012345678901234567890*/
#define SMENU_verbose			"verbose"
#define SMENU_lock_bpp			"BPP locked when +/-"
#define SMENU_lock_text_graphic		"text/graphic locked when +/-"
#define SMENU_force_textmode		"force last text mode at startup"
#define SMENU_menu_with_disk		"display disk in menu"
#define SMENU_menu_with_partition	"display partition in menu"
#define SMENU_VGA_interface		"VGA interface"
#define SMENU_refuse_8x14font		"refuse 8x14 font"
#define SMENU_VESA_interface		"VESA interface"
#define SMENU_enable_VESA_hardwin	"VESA window I/O detection"
#define SMENU_VESA2_interface		"VESA linear interface"
#define SMENU_enable_joystick		"enable Joystick"
#define SMENU_search_disk_mbr		"search disk's mbr"
#define SMENU_prefer_enddisk_mbr	"prefer end of disk MBR"
#define SMENU_search_part_mbr		"search partition's mbr"
#define SMENU_search_topdir_kernel	"search kernels in root"
#define SMENU_search_subdir_kernel	"search kernels in directory"
#define SMENU_probe_bios_floppy_disk	"probe BIOS floppy disks"
#define SMENU_probe_bios_hard_disk	"probe BIOS hard disks"
#define SMENU_probe_ide_disk		"probe IDE disks"
#define SMENU_probe_atapi_cdrom		"probe ATAPI/CDROM (not implemented)"
#define SMENU_probe_dos_disk		"probe DOS disks"

#define SMENU_PROBE_VESA		"probe VESA modes 0..0x7F"
#define SMENU_PROBE_VGA_0_13		"probe VGA modes 0..0x13"
#define SMENU_PROBE_VGA_14_127		"probe VGA modes 0x14..0x7F"
#define SMENU_RESET_VGA_MODES		"reset VGA/VESA modes 0..0x7F"
#define SMENU_RESET_ALL_PARAMS		"reset all saved parameters"
#define SMENU_POWER_DOWN		"turn power off"
#define SMENU_FAIL_BOOT			"declare boot failed"
#define SMENU_SET_TIMEOUT_AUTOLOAD	"set auto-booting timeout"

#define MENU_UNKNOWN_KEY	"Unknown command (keycode 0x%X)."
#endif

#if LANGUAGE == FRENCH
#define NO_MODE_FOUND		"Pas de mode trouv."
#define MODE_X_U_U_OVER_MONITOR_SPEC "Mode 0x%X, %ux%u est au dessus des spc. du moniteur."
#define ERROR_VESA_SWITCH_FAILED MSG_ERROR "passage en VESA rat !"
#define ERROR_VGA_SWITCH_FAILED	 MSG_ERROR "passage en VGA rat !"
#define SETTING_MODE_X		 "   Passage au mode 0x%X"
#define ERROR_SETTING_X		 ": erreur 0x%X"
#else /* LANGUAGE == ENGLISH */
#define NO_MODE_FOUND		"No mode found."
#define MODE_X_U_U_OVER_MONITOR_SPEC "Mode 0x%X, %ux%u is over monitor specs."
#define ERROR_VESA_SWITCH_FAILED MSG_ERROR "VESA switch failed!"
#define ERROR_VGA_SWITCH_FAILED	 MSG_ERROR "VGA switch failed!"
#define SETTING_MODE_X		 "   Setting mode 0x%X" /* shifted for probe */
#define ERROR_SETTING_X		 ": error 0x%X"
#endif

/*
 * Loader messages:
 */
#if LANGUAGE == FRENCH

#define SWITCHED_TO_TEXT_MODE \
    "Passage en mode texte (compilation du framebuffer dans vmlinuz?)" CRLF
#define LOADING_KERNEL_S        "Chargement noyau \"%s\","
#define LOADING_INITRD_S        "chargement disque RAM initial \"%s\","
#define ROOT_X                  " racine: 0x%X;"
#define COMMAND_LINE_S          " Ligne de commande: %s" CRLF
#define STARTING_KERNEL         "Lancement noyau."

#define ERROR_XMS_ENABLE_A20    " erreur activation A20 par HIMEM.SYS."
#define ERROR_NOT_ENOUGH_RAM    " erreur pas assez de mmoire."
#define ERROR_SEGMENT_DOOR      " erreur ouverture des portes de segment."
#define ERROR_XMS_ALLOCATE      " erreur 0x%X allocation mmoire XMS."
#define ERROR_LOADING_FILE      " erreur 0x%X au chargement du fichier."
#define ERROR_XMS_LOCK          " erreur 0x%X vrouillage mmoire."

#define ERROR_STARTUP_ENABLE_A20        " erreur activation A20 au dmarrage !"
#define ERROR_SWITCH_PROTECTED_MODE     " erreur passage en mode protg."

#else /* LANGUAGE == ENGLISH */

#define SWITCHED_TO_TEXT_MODE \
    "Switched to text mode (framebuffer compiled in vmlinuz?)" CRLF
#define LOADING_KERNEL_S        "Loading kernel \"%s\","
#define LOADING_INITRD_S        "loading initial RAMdisk \"%s\","
#define ROOT_X                  " root: 0x%X;"
#define COMMAND_LINE_S          " Command line: %s" CRLF
#define STARTING_KERNEL         "Starting kernel."

#define ERROR_XMS_ENABLE_A20    " error HIMEM.SYS enabling A20."
#define ERROR_NOT_ENOUGH_RAM    " error not enough RAM."
#define ERROR_SEGMENT_DOOR      " error opening segment doors."
#define ERROR_XMS_ALLOCATE      " error 0x%X allocating XMS memory."
#define ERROR_LOADING_FILE      " error 0x%X loading file."
#define ERROR_XMS_LOCK          " error 0x%X locking memory."

#define ERROR_STARTUP_ENABLE_A20        " error enabling A20 at startup."
#define ERROR_SWITCH_PROTECTED_MODE     " error switching to protected mode."

#endif

/*
 * Other messages:
 */
#if LANGUAGE == FRENCH
#define POWERING_OFF	        "Arrt"
#else /* LANGUAGE == ENGLISH */
#define POWERING_OFF            "Power OFF"
#endif


#if LANGUAGE == FRENCH
#define PROCESSOR_MSG           "processeur %s famille %d"
#define AT_FREQUENCY		"  %U MHz"
#define WITHOUT_FPU             " sans FPU"
#else /* LANGUAGE == ENGLISH */
#define PROCESSOR_MSG           "%s processor family %d"
#define AT_FREQUENCY		" at %U MHz"
#define WITHOUT_FPU             " w/o FPU"
#endif

#if LANGUAGE == FRENCH
#define KEYBOARD_122            ", clavier 122 touches"
#define KEYBOARD_EXT            ", clavier tendu"
#define KEYBOARD_STD            ", clavier standard"
#else /* LANGUAGE == ENGLISH */
#define KEYBOARD_122            ", 122 key keyboard"
#define KEYBOARD_EXT            ", ext keyboard"
#define KEYBOARD_STD            ", std keyboard"
#endif

#if LANGUAGE == FRENCH
#define U_SERIAL_PORT			 "%u port%s srie, "
#define BIOS_PS2_MOUSE			 "sourie BIOS (PS2)"
#define SERIAL_ON_COMX_MOUSE		 "sourie srie sur COM%u"
#define SERIAL_3_BUTTON_ON_COMX_MOUSE	 "sourie srie 3 boutons sur COM%u"
#define SERIAL_MOUSESYSTEM_ON_COMX_MOUSE "sourie srie MouseSystems sur COM%u"
#define NO_MOUSE			 "pas de sourie"
#define ERROR_MOUSE			 "erreur sourie"
#define JOYSTICK_PRESENT		 ", joystick"
#else /* LANGUAGE == ENGLISH */
#define U_SERIAL_PORT			 "%u serial port%s, "
#define BIOS_PS2_MOUSE			 "BIOS (PS2) mouse"
#define SERIAL_ON_COMX_MOUSE		 "serial on COM%u mouse"
#define SERIAL_3_BUTTON_ON_COMX_MOUSE	 "serial 3 buttons on COM%u mouse"
#define SERIAL_MOUSESYSTEM_ON_COMX_MOUSE "serial MouseSystems on COM%u mouse"
#define	NO_MOUSE			 "no mouse"
#define ERROR_MOUSE			 "error mouse"
#define JOYSTICK_PRESENT		 ", joystick"
#endif

#if LANGUAGE == FRENCH
#define SERIAL_TERMINAL		"Terminal srie, "
#define TERMINAL_MODES		", modes: 0..%u" DOT_CRLF
#define SERIAL_INIT_FAILED	"Erreur initialisation srie !"
#define VGA_CARD_PRESENT_MODES	"Carte VGA prsente, modes: "
#define S_CARD_VESA_VERSION_U_U	"Carte vido %s, version VESA %u.%u." CRLF
#define VESA_USING_IOPORTS	MSG_INFO "carte VESA utilisant les I/O ports: "
#else /* LANGUAGE == ENGLISH */
#define SERIAL_TERMINAL		"Serial terminal, "
#define TERMINAL_MODES		", modes: 0..%u" DOT_CRLF
#define SERIAL_INIT_FAILED	"Serial initialisation failed!"
#define VGA_CARD_PRESENT_MODES	"VGA card present, modes: "
#define S_CARD_VESA_VERSION_U_U	"%s video card, VESA version %u.%u." CRLF
#define VESA_USING_IOPORTS	MSG_INFO "VESA card is using I/O ports: "
#endif

#if LANGUAGE == FRENCH
#define FORCE_VIDEO_MODE_3 \
    MSG_INFO "Touche Control presse, mode vido forc en texte simple."
#define IMPORTANT_PARAMETERS_RESETED \
    MSG_INFO "Maj-Lock actif, paramtres inportants r-initialiss."
#else /* LANGUAGE == ENGLISH */
#define FORCE_VIDEO_MODE_3 \
    MSG_INFO "Control key pressed, video mode set to simple text."
#define IMPORTANT_PARAMETERS_RESETED \
    MSG_INFO "Caps-Lock active, most important parameters reseted."
#endif

#if LANGUAGE == FRENCH
#define HIMEM_MAXFREE		"Himem libre max: %u Ko, "
#define BASE_MEMORY_U_EXTENDED_MEMORY_U_S \
    "mmoire de base: %u Ko, mmoire tendue: %u Ko [%s]"   CRLF
#else /* LANGUAGE == ENGLISH */

#define HIMEM_MAXFREE		"Himem maxfree: %u Kb, "
#define BASE_MEMORY_U_EXTENDED_MEMORY_U_S \
    "base memory: %u Kb, extended memory: %u Kb [%s]"       CRLF
#endif

/*
 * Following messages can have "\n" only, no need of "\r\n",
 * because they are used in centered_print()
 */
#if LANGUAGE == FRENCH
#define TURN_OFF_MONITOR \
    "Il est prfrable d'teindre votre moniteur pour ces tests,\n"	\
    "jusqu' ce que vous entendiez le bip de fin.\n"			\
    "Cela peut prendre quelques minutes sur un ancien PC.\n"
#define IN_U_SECOND_AUTOLOAD_S \
   "Dans %u secondes, chargement de:\n" "%s\n" "Pressez une touche pour annuler"
#define ADVERTISEMENT_BANNER \
    "Bienvenue dans le monde du logiciel libre.\n"			  \
    "SVP, consultez au moins une fois la licence des programes que\n"	  \
    "vous utilisez. Gujin est disponible sous GPL v2, lisez GPL.TXT ou\n" \
    "visitez http://www.fsf.org/licenses (traductions)"
#else /* LANGUAGE == ENGLISH */
#define TURN_OFF_MONITOR \
    "It is better to turn OFF your monitor for these tests,\n"	\
    "until you hear the end BIP.\n"				\
    "This can take few minutes on slow PC.\n"
#define IN_U_SECOND_AUTOLOAD_S \
    "In %u seconds, loading of:\n" "%s\n" "Press any key to cancel"
#define ADVERTISEMENT_BANNER \
    "Welcome to the world of free (as in speech) software.\n"	\
    "Please read once the license of software you are using.\n"	\
    "Gujin is available under GPL v2, read file GPL.TXT or\n"	\
    "visit http://www.fsf.org/licenses (translations)."
#endif

#define ADVERTISEMENT_BANNER_NBLINES 4
#define ADVERTISEMENT_BANNER_MINWIDTH	70
