/* mouse.h */
#ifndef MOUSE_H
#define MOUSE_H

/* Gujin is a bootloader, it loads a Linux kernel from cold boot or DOS.
 * Copyright (C) 1999-2001 Etienne Lorrain, fingerprint (2D3AF3EA):
 *   2471 DF64 9DEE 41D4 C8DB 9667 E448 FF8C 2D3A F3EA
 * E-Mail: etienne.lorrain@masroudeau.com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "make.h"

#if USER_SUPPORT & (BIOS_MOUSE_SUPPORT | SERIAL_MOUSE_SUPPORT)

/*
 * This is the heigh of the mouse graphic cursor.
 * It has to be changed depending on:
 * static const union mouse_cursor cursor_mask[]
 */
#define SIZEOF_CURSOR_MASK	14

union mouse_cursor {
      struct {
	unsigned char b0 :1, b1 :1, b2 :1, b3 :1, b4 :1, b5 :1, b6 :1, b7 :1;
	} __attribute__ ((packed)) bit;
      unsigned char group;
      };
/*
 * Colors of the mouse itself when outside a field,
 * inside a field, or (prioritary) when one button is pressed:
 */
#define MOUSE_DEFAULT_COLOR			UI.stdcolor[brown]
//#define MOUSE_DEFAULT_COLOR UI.fgcolor

#define MOUSE_INFIELD_COLOR			UI.stdcolor[magenta]
#define MOUSE_OUTFIELD_COLOR_RIGHTBUTTON	UI.stdcolor[green]
#define MOUSE_OUTFIELD_COLOR_LEFTBUTTON		UI.stdcolor[red]
#define MOUSE_OUTFIELD_COLOR_TWOBUTTON		UI.stdcolor[magenta]
#define MOUSE_OUTFIELD_COLOR_MIDDLEBUTTON	UI.stdcolor[lightcyan]
#define MOUSE_OUTFIELD_COLOR_TOPBUTTON		UI.stdcolor[cyan]

#if USER_SUPPORT & SERIAL_MOUSE_SUPPORT
/*
 * To use the extended serial BIOS instead of direct I/O access
 * to read the "modem" lines.
 */
//#define EXTENDED_SERIAL_BIOS

#ifdef EXTENDED_SERIAL_BIOS
#define SETMODEM(cpt, ioaddr, modem) \
	_BIOS_extended_set_modem_control (cpt, modem)
#define GETMODEM(cpt, ioaddr, modem) \
	_BIOS_extended_get_modem_control (cpt, &modem)
#else
#define SETMODEM(cpt, ioaddr, modem) \
	outb (ioaddr+4, *(unsigned char *)&modem)
#define GETMODEM(cpt, ioaddr, modem) \
	modem = inb (ioaddr+4)
#endif

#if 0 /* too slow on a 386: */
#define SERIAL_DATA_RECEIVED(Mouse_data_serial) \
    (_BIOS_getserialstatus(Mouse_data_serial.comport).receive_data_ready != 0)
#else
#define SERIAL_DATA_RECEIVED(Mouse_data_serial) \
    ((inb (Mouse_data_serial.comIOaddr+5) & 0x01) != 0)
#endif

/* The only way which do not wait modem lines: */
#define SERIAL_DATA_GET(Mouse_data_serial) \
    (inb (Mouse_data_serial.comIOaddr))

#endif /* SERIAL_MOUSE_SUPPORT */

/*
 * The main structure:
 */
struct mouse_interface_str {
    coord		position, maximum;
    struct mouse_button_str {
	unsigned char left	: 1;
	unsigned char right	: 1;
	unsigned char middle	: 1;
	unsigned char top	: 1;
	unsigned char reserved	: 3;
	unsigned char hidden	: 1;
	} button;

    enum MOUSE_type {
	MOUSE_NONE= 0,
	MOUSE_PS2 = 0xAA,
	MOUSE_SERIAL_COM1 = 0x10, MOUSE_SERIAL_COM2,   /* 2 buttons */
	MOUSE_SERIAL_COM3, MOUSE_SERIAL_COM4,
	MOUSE_SERIAL3_COM1 = 0x30, MOUSE_SERIAL3_COM2, /* 3 buttons */
	MOUSE_SERIAL3_COM3, MOUSE_SERIAL3_COM4,         /* Logitech */
	MOUSE_SERIALM_COM1 = 0x40, MOUSE_SERIALM_COM2, /* 3 buttons */
	MOUSE_SERIALM_COM3, MOUSE_SERIALM_COM4         /* MouseSystems */
#define MOUSE_SERIAL_MASK 0xF0
	} type;

    unsigned short		nb_field;
    struct mousefield_str	*fields;

#ifdef MOUSE_ON_VT
    unsigned char curs_row, curs_col;
#endif
#if USER_SUPPORT & JOYSTICK_SUPPORT
    struct {
	unsigned short Xmin, Xmid, Xmax, Ymin, Ymid, Ymax, accel;
	unsigned last_sample_tick;
	struct JoystickButton_str lastbutton;
	} joystick;
#endif

    unsigned savearray[8 * SIZEOF_CURSOR_MASK];

    union {
	struct PS2_struct {
	    unsigned short reservedword	: 16;
	      signed char y_data	: 8;
	      signed char y_reserved	: 8;
	      signed char x_data	: 8;
	      signed char x_reserved	: 8;
	    unsigned char left_button	: 1;
	    unsigned char right_button	: 1;
	    unsigned char middle_button	: 1;
	    unsigned char reservedbit	: 1;
	    unsigned char x_negative	: 1;
	    unsigned char y_negative	: 1;
	    unsigned char x_overflow	: 1;
	    unsigned char y_overflow	: 1;
	    unsigned char reservedbyte	: 8;
	    } __attribute__ ((packed)) PS2;
	struct {
	    unsigned comIOaddr			: 16;
	    unsigned comport			: 8;
	    unsigned cpt			: 8;
	    union {
		unsigned char bytes[5]; /* 3, 4 or 5 */
		struct {
		    unsigned xmsb		: 2;
		    unsigned ymsb		: 2;
		    unsigned right_button	: 1;
		    unsigned left_button	: 1;
		    unsigned cste1		: 1;
		    unsigned sevenbit1		: 1;
		    unsigned x			: 6;
		    unsigned cste0_1		: 1;
		    unsigned sevenbit2		: 1;
		    unsigned y			: 6;
		    unsigned cste0_2		: 1;
		    unsigned sevenbit3		: 1;
		    /* 32 = 0x20 when pressed, 0 when released: */
		    unsigned logitech		: 8;
		    } __attribute__ ((packed)) stddata;
		struct {
		    unsigned char right_button	: 1;
		    unsigned char middle_button	: 1;
		    unsigned char left_button	: 1;
		    unsigned char synchro_0x10	: 5;
		      signed char x		: 8;
		      signed char y		: 8;
		      signed char deltax	: 8;
		      signed char deltay	: 8;
		    } __attribute__ ((packed)) mmdata;
		} received;
	    } serial;
	} data;
    };

extern struct mouse_interface_str Mouse;

/*
 * The functions:
 */
struct mousefield_str {
    const char *str;
    unsigned short keypress;
    coord upperleft, lowerright	;
    struct mousefieldattr_str {
	unsigned short upperfield	: 1;
	unsigned short mainfield	: 1;
	unsigned short active		: 1;
	unsigned short pressed		: 1;
	unsigned short checkbox		: 1;
	unsigned short greyed		: 1;
	} attr;
    };

void MOUSE_reset_field (struct mousefieldattr_str attr);
void MOUSE_print_field (const char *str,
			struct mousefieldattr_str attr,
			unsigned short key,
			unsigned short row,
			unsigned short col);

unsigned MOUSE_init (unsigned nb_serial_port);

extern inline enum MOUSE_type MOUSE_gettype (void)
  {
  return Mouse.type;
  }

extern inline int MOUSE_has_moved (void)
  {
  /* Called after a MOUSE_stop() */
  /* Mouse is no more in its default position, so cancel the autoload timer */
  return (Mouse.position.x != 0xFFFFU / 8 || Mouse.position.y != 0xFFFFU / 16);
  }

#if USER_SUPPORT & JOYSTICK_SUPPORT
extern inline void JoyStick_init_released (unsigned short Ax, unsigned short Ay)
  {
  Mouse.joystick.Xmid = Ax;
  Mouse.joystick.Ymid = Ay;
  }
#endif

#endif /* BIOS_MOUSE_SUPPORT | SERIAL_MOUSE_SUPPORT */

#endif /* MOUSE_H */

