/* user.h */
#ifndef USER_H
#define USER_H

/* Gujin is a bootloader, it loads a Linux kernel from cold boot or DOS.
 * Copyright (C) 1999-2001 Etienne Lorrain, fingerprint (2D3AF3EA):
 *   2471 DF64 9DEE 41D4 C8DB 9667 E448 FF8C 2D3A F3EA
 * E-Mail: etienne.lorrain@masroudeau.com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "make.h"

#if USER_SUPPORT != 0

#if USER_SUPPORT & VGA_SUPPORT
/*
 * Define this if you think extended VGA modes (> 0x37)
 * with less than or equal to 2 BPP are valid modes:
 * There is probably some NDA to sign somewhere to get
 * the real description of version 2 of the structure
 * "functionality_table_type" in "vga.h", but there is not
 * enought time in live to sign all the NDA of PC BIOSes.
 */
#undef KEEP_ALL_VGA_MODE
#endif

#if USER_SUPPORT & VESA_SUPPORT
/*
 * This value has been chosen (for winsize) because:
 *  if (offset1 / UI.parameter.winsize
 *	 != offset2 / UI.parameter.winsize)
 *      is always false if VESA2 because 0 == 0
 * And also:
 *  if (offset - UI.parameter.winadr >= UI.parameter.winsize)
 *      is always false if VESA2
 */
#define VESA2_MARKER UINT_MAX

#if (USER_SUPPORT & VESA_4BPP_EGA)
/*
 * Define that to the fastest read_pixel (0 or 1) in 16 color mode:
 */
//#define EGA_DEFAULT_READ_MODE 0
#define EGA_DEFAULT_READ_MODE 1

#endif /* VESA_4BPP_EGA */

#if USER_SUPPORT & VESA_HARDWINDOW
/*
 * Define it if it works, report if it does not work,
 * that saves quite a lot of malloc'ed memory.
 * It is normal to have some mode without compression,
 * but I would like to know if I failed to compress
 * some obvious pattern when it would save few Kbytes:
 */
#define COMPRESS_HARDWIN

#ifdef COMPRESS_HARDWIN
/*
 * ET4000 needs a special treatment because it outputs
 * the read and write page in the same outb() - This
 * define fixes it but is not very clean an algorithm.
 */
#define COMPRESS_EXTENDED
#endif /* COMPRESS_HARDWIN */
#endif /* VESA_HARDWINDOW */
#endif /* VESA_SUPPORT */

/*
 * We cannot send "null" character with our interface (to slow down
 * the commands because even at 9600 bauds a true VT420 cannot
 * process everything) so we use _BIOS_wait () with this parameter
 * for "slow" operations (parameter in microseconds):
 */
#define SERIAL_DELAY 30000  /* 3/100 of a second - or should be... */

/*
 * This is just for testing purposes, it is only
 * using get/setpixel to scroll and display horizontal lines.
 */
//#define VERY_SLOW

/***************************************************************
 * The graphic interface:
 */

enum stdcolor_enum {
    black,    blue,         green,      cyan,
    red,      magenta,      brown,      lightgray,
    darkgray, lightblue,    lightgreen, lightcyan,
    lightred, lightmagenta, yellow,     white
    };

struct DAC_str { /* same as VESA_RGB */
    unsigned char blue, green, red, align;
    };

typedef struct {
    unsigned short x, y;	/* x should be the easyest to read */
    } __attribute__ ((packed)) coord;

struct user_interface_str {
    unsigned (*getkey) (unsigned timeout); /* timeout in ticks, 18/s, 0 = none */

    unsigned fgcolor, bgcolor;
    unsigned stdcolor[16];	/* content: VESA_RGB, VGA_RGB or 1..16 */
    struct {
	struct attribute_str {
	    unsigned char underline	:1;
	    unsigned char reverse	:1;
	    unsigned char blink		:1;
	    unsigned char brighter	:1;
	    unsigned char darker	:1;
	    unsigned char transparent	:1; /* no background drawn in VESA */
	    unsigned char reserved	:2;
	    } __attribute__ ((packed)) current, valid;
	} attributes;
    unsigned short monitor;    /* enum unknown, B&W, mono, color or EDID type */

    unsigned char	 nbmode;
    struct videomode_str {
	unsigned short	number;
	unsigned short	width;
	unsigned short	heigh;
	unsigned char	bpp;
	unsigned char	text     : 1;
	unsigned char	attr     : 1;
	unsigned char	vesa     : 1;
	unsigned char	font8x14 : 1;
	} *mode;

#if USER_SUPPORT & VGA_FREQUENCY
    unsigned short frequency;
#endif

    union {
#if USER_SUPPORT & VESA_SUPPORT
	struct {
	    unsigned short	version;
	    unsigned short	memory_64Kb;
	    unsigned char	cardname[40];
	    } vesa;
#endif
#if USER_SUPPORT & SERIAL_SUPPORT
	struct {
	    unsigned		max_mode;
	    unsigned char	terminalname[40];
	    } serial;
#endif
	} info;

#if USER_SUPPORT & VESA_EDID
    struct edid_str {
	unsigned short	horizontal_resolution;
	unsigned short	vertical_resolution;
	unsigned short	frequency;
	unsigned char	Hsize_cm, Vsize_cm;
	} edid;
#endif
#if USER_SUPPORT & VESA_PMINFO
    unsigned short ProtectedModeIO[32];
    struct {
	unsigned address;
	unsigned short length;
	} ProtectedModeMem[4];
#endif

    struct video_parameter_str {
	unsigned        identification; /* do not have linear or clear bit */
	unsigned        nbcolor;
	unsigned short  width, heigh;
	unsigned	base_address;
	unsigned char   row, col;
	unsigned char   nbrow, nbcol;
	unsigned char   charwidth, charheigh;
	struct {
	    unsigned char isgraphic	 : 1;
	    /*
	     * Some VGA cards, in VGA mode, can accept 256 colors background,
	     * color given by the "page" parameter, VESA card have always
	     * this flag set when graphic (at least >= 16 colors):
	     */
	    unsigned char graphicbgcolor : 1;
	    } attr;

	/* just to report to vmlinuz: */
	unsigned char	nbimage;

	/* used also for VGA 256 colors modes: */
	unsigned char	red_mask_size;
	unsigned char	red_field_pos;
	unsigned char	green_mask_size;
	unsigned char	green_field_pos;
	unsigned char	blue_mask_size;
	unsigned char	blue_field_pos;

#if USER_SUPPORT & VESA_SUPPORT
	farptr		font;	  /* 0 if text mode, hardware font used */
	unsigned	winadr;
	unsigned	winsize;
	unsigned short  linelength;
	unsigned char   bitperpixel;
	unsigned char   nbplane;
	unsigned char   memtype;
	unsigned char	EGA_compatible;
	unsigned char	BIOS_compatible;

#if USER_SUPPORT & VESA_WINDOW
	unsigned char	NbWin;	/* Number of VESA windows */
	unsigned char	winNo;	/* R/W window,
				   or WO window, then RO is (winNo ^ 1) */
#if USER_SUPPORT & VESA_HARDWINDOW
	unsigned short	*offsetarray;
	struct hardwindow_str {
	    unsigned short instruction;
	    unsigned short port;
	    unsigned	   data;
	    } *hardwindow;
#endif
#if USER_SUPPORT & VESA_STRANGE_WINDOW
	unsigned	wingranul;
#else
	unsigned char	wingranul;
#endif
#if USER_SUPPORT & VESA_2WINDOWS
	unsigned	winreadadr;
	unsigned	readdelta; /* HAS TO BE null if NbWin < 2 */
#if USER_SUPPORT & VESA_HARDWINDOW
	unsigned short	*readoffsetarray;
	struct hardwindow_str *readhardwindow;
#endif
#endif
#if USER_SUPPORT & VESA_WINFUNCTION
	unsigned        addrpositionfarfct;
#endif
#endif /* VESA_WINDOW */
#endif /* VESA_SUPPORT */
	} parameter;

    struct user_interface_fct_str {
	unsigned (*getsize) (unsigned mode, struct video_parameter_str *param);
	unsigned (*setmode) (unsigned mode);
	unsigned (*getmode) (void);
	void     (*clearscreen) (void);

	/* Cursor: home at (0,0) */
	unsigned (*setcursor) (unsigned char row, unsigned char col);
	unsigned (*getcursor) (unsigned char *row, unsigned char *col);

	/* Colors are set by functions, and get directly from structure: */
	unsigned (*setfgcolor) (unsigned color);
	unsigned (*setbgcolor) (unsigned color);

	void     (*setattribute) (struct attribute_str attr);

	void     (*putstr) (const char *str);

	void     (*setpixel) (coord xy, unsigned color) FASTCALL;
	unsigned (*getpixel) (coord xy) FASTCALL;
	void	 (*plotHline) (coord xy, unsigned short xend, unsigned color);

	unsigned (*setDACblock) (unsigned short startindex, unsigned short nb,
				 const struct DAC_str *palette);
	} function;

    struct {
	void (*start) (void);
	void (*stop) (void);
	unsigned (*poll) (void);
	} mouse;
    };

/* static consts removal (by GCC) does not seems to work, so do that: */
#define all_attribute		((struct attribute_str) {1, 1, 1, 1, 1, 1})
#define no_attribute		((struct attribute_str) {0, 0, 0, 0, 0, 0})
#define egatxt_attribute	((struct attribute_str) {1, 1, 1, 1, 0, 0})
#define egagfx_attribute	((struct attribute_str) {0, 0, 1, 1, 0, 0})
#define graphic_attribute	((struct attribute_str) {1, 1, 0, 0, 0, 1})

#define attr_underline		((struct attribute_str) {1, 0, 0, 0, 0, 0})
#define attr_reverse		((struct attribute_str) {0, 1, 0, 0, 0, 0})
#define attr_blink		((struct attribute_str) {0, 0, 1, 0, 0, 0})
#define attr_brighter		((struct attribute_str) {0, 0, 0, 1, 0, 0})
#define attr_darker		((struct attribute_str) {0, 0, 0, 0, 1, 0})
#define attr_transparent	((struct attribute_str) {0, 0, 0, 0, 0, 1})

#define FCT_KEYCODE(i)	(((i <= 10)? (0x3AU + i) : (0x85U + i - 11)) << 8)
#define SFCT_KEYCODE(i)	(((i <= 10)? (0x53U + i) : (0x87U + i - 11)) << 8)
#define CFCT_KEYCODE(i)	(((i <= 10)? (0x5DU + i) : (0x89U + i - 11)) << 8)
#define AFCT_KEYCODE(i)	(((i <= 10)? (0x67U + i) : (0x8BU + i - 11)) << 8)

#define ESC_KEYCODE()		(0x0100 | '\033')
#define SPACE_KEYCODE()		(0x3900 | ' ')
#define BACKSPACE_KEYCODE()	(0x0E00 | '\010')
#define RETURN_KEYCODE()	(0x1C00 | '\r')
#define KP_RETURN_KEYCODE()	(0xE000 | '\r') /* only extkbd */
#define MINUS_KEYCODE()		(0x4A00 | '-')
#define PLUS_KEYCODE()		(0x4E00 | '+')
#define DIVIDE_KEYCODE()	(0x3500 | '/')
#define KP_DIVIDE_KEYCODE()	(0xE000 | '/') /* only extkbd */
#define MULTIPLY_KEYCODE()	(0x3700 | '*')
#define DOT_KEYCODE()		(0x3400 | '.')
#define NUMPAD_DOT_KEYCODE()	(0x5300 | '.')
#define DIGIT_KEYCODE(i)	(((i+1)<<8) | ('0' + i))
#define NUMPAD_DIGIT_KEYCODE(i) \
    ((((i == 0)? 0x52 : (0x4E - ((i - 1)/3 * 7) + i)) << 8) | ('0' + i))

#define PAGEUP_KEYCODE()	(0x4900)
#define PAGEDOWN_KEYCODE()	(0x5100)
#define HOME_KEYCODE()		(0x4700)
#define END_KEYCODE()		(0x4F00)
#define UP_KEYCODE()		(0x4800)
#define DOWN_KEYCODE()		(0x5000)
#define RIGHT_KEYCODE()		(0x4D00)
#define LEFT_KEYCODE()		(0x4B00)
#define TAB_KEYCODE()		(0x0F09)
#define ON_OFF_KEYCODE()	(0x5F80)
#define CRTLBREAK_KEYCODE()	(0x0000)	/* Hopes it work everywhere */
// Not sure of those:
#define POWER_KEYCODE()		(0x5EE0)
#define SLEEP_KEYCODE()		(0x5FE0)
#define WAKE_KEYCODE()		(0x63E0)

// Fake key for timeouts:
#define TIMEOUT_KEYCODE()	0xFFFE

/*
 * Colors of the fields themselves:
 */
#if USER_SUPPORT & (VGA_SUPPORT | VESA_SUPPORT)
/* Backgrounds in graphic modes >= 256 colors: */
#define FIRST_LINES_BACKGROUND	VESA_color (35, 30, 30)
#define TOP_MENU_BACKGROUND	VESA_color (35, 45, 35)
#define TOP_MENU_BUTTON_BG	VESA_color (35, 40, 35)
#define BOTTOM_MENU_BACKGROUND	VESA_color (35, 35, 35)
#define BOTTOM_MENU_BUTTON_BG	VESA_color (30, 30, 35)
#define ADVERT_BACKGROUND	VESA_color (30, 30, 30)
#define WORK_ADVERT_BACKGROUND	VESA_color (35, 35, 30)
#define COPYRIGHT_BACKGROUND	VESA_color (20, 40, 40)
#endif /* VGA_SUPPORT | VESA_SUPPORT */

#define CHECKBOX_BG		UI.stdcolor[white]
#define CHECKBOX_INACTVE_BG	UI.stdcolor[lightgray]

/*
 * Only if no interresting attribute available (even if no mouse):
 * (but also on VT/serial interface if no attributes)
 */
#define MOUSE_SELECTABLE_CHARS	UI.stdcolor[lightblue]


#define USER_ACTIVE()	(UI.parameter.nbcol != 0)

#if USER_SUPPORT & VGA_SUPPORT
#if USER_SUPPORT & VESA_SUPPORT
#define VGA_ACTIVE()	(USER_ACTIVE() && (UI.parameter.linelength == 0))
#define VESA_ACTIVE()	(USER_ACTIVE() && (UI.parameter.linelength != 0))
#else
#define VGA_ACTIVE()	USER_ACTIVE()
#define VESA_ACTIVE()	0
#endif
#else /* VGA_SUPPORT */
#define VGA_ACTIVE()	0
#define VESA_ACTIVE()	USER_ACTIVE()
#endif /* VGA_SUPPORT */

extern inline unsigned char
get_bpp (struct video_parameter_str *ptr)
  {
#if USER_SUPPORT & VESA_SUPPORT
  if (   ptr->bitperpixel != 16	/* 15 & 16 bpp are same */
      && ptr->bitperpixel != 1	/* 4bpp EGA */
      && ptr->bitperpixel != 0) /* VGA modes */
      return ptr->bitperpixel;
    else
#endif
      {
      if (ptr->nbcolor != 0)
	  return my_ffs (ptr->nbcolor) - 1;
	else
	  return 0;
      }
  }

extern struct user_interface_str UI;

#if !defined (CALLING_FROM_EXTRASEG)
/* The following functions cannot be called from extraseg,
   because UI.function.* are intra-segment calls: */
extern inline unsigned
UI_getkey (unsigned timeout) {
    return UI.getkey (timeout);
    }


extern inline unsigned
UI_getsize (unsigned mode, struct video_parameter_str *param) {
    return UI.function.getsize (mode, param);
    }

extern inline unsigned
UI_setmode (unsigned mode) {
    return UI.function.setmode (mode);
    }

extern inline unsigned
UI_getmode (void) {
    return UI.function.getmode ();
    }

extern inline void
UI_clearscreen (void) {
    UI.function.clearscreen ();
    }

extern inline unsigned
UI_setcursor (unsigned char row, unsigned char col) {
    return UI.function.setcursor (row, col);
    }

extern inline unsigned
UI_getcursor (unsigned char *row, unsigned char *col) {
    return UI.function.getcursor (row, col);
    }

extern inline unsigned
UI_setfgcolor (unsigned color) {
    return UI.function.setfgcolor (color);
    }

extern inline unsigned
UI_setbgcolor (unsigned color) {
    return UI.function.setbgcolor (color);
    }

extern inline void
UI_setattribute (struct attribute_str attr) {
    UI.function.setattribute (attr);
    }

extern inline void
UI_putstr (const char *str) {
    UI.function.putstr (str);
    }

extern inline void
UI_setpixel (coord xy, unsigned color) {
    UI.function.setpixel (xy, color);
    }

extern inline unsigned
UI_getpixel (coord xy) {
    return UI.function.getpixel (xy);
    }

extern inline void
UI_plotHline (coord xy, unsigned short xend, unsigned color) {
    UI.function.plotHline (xy, xend, color);
    }

extern inline unsigned
UI_setDACblock (unsigned short startindex,
		unsigned short nb,
		const struct DAC_str *palette) {
    return UI.function.setDACblock (startindex, nb, palette);
    }

#if USER_SUPPORT & VESA_SUPPORT
unsigned VESA_init (void);
#endif

#if USER_SUPPORT & VGA_SUPPORT
unsigned VGA_init (void);
#endif

/* Following are never called from extra segment: */
#if USER_SUPPORT & (VGA_SUPPORT | VESA_SUPPORT)
unsigned BIOS_probemonitor (void);
unsigned __attribute__ ((const))
VESA_color (unsigned char _red, unsigned char _green, unsigned char _blue);
#endif

#if (USER_SUPPORT & VGA_SUPPORT) && (SETUP & UPDATE_BOOTPARAM)
void setup_gujin_param_vga_mode (void);
void restore_gujin_param_vga_mode (void);
#endif

#if USER_SUPPORT & SERIAL_SUPPORT
unsigned SERIAL_init (void);
#endif

#ifdef MOUSE_ON_VT /* Called from mouse.c: */
unsigned SERIAL_setcursor (unsigned char row, unsigned char col);
unsigned SERIAL_getcursor (unsigned char *row, unsigned char *col);
#endif

void UI_init (unsigned serial);
void reset_low_valid_mode (unsigned up_to);

#elif SETUP & EXTRA_SEGMENT
/* What can be called from extra segment: */
#define ASMNAME(x)	asm (STRING(extraseg_ ## x))

unsigned UI_getkey (unsigned timeout) ASMNAME(UI_getkey);

#if USER_SUPPORT & VGA_SUPPORT
unsigned VGA_init (void) ASMNAME(VGA_init);
#endif

#if USER_SUPPORT & VESA_SUPPORT
unsigned VESA_init (void) ASMNAME(VESA_init);
#endif

unsigned UI_setmode (unsigned mode) ASMNAME(UI_setmode);
unsigned UI_getmode (void) ASMNAME(UI_getmode);
void UI_clearscreen (void) ASMNAME(UI_clearscreen);
unsigned UI_setcursor (unsigned char row, unsigned char col) ASMNAME(UI_setcursor);
unsigned UI_getcursor (unsigned char *row, unsigned char *col) ASMNAME(UI_getcursor);
unsigned UI_setfgcolor (unsigned color) ASMNAME(UI_setfgcolor);
unsigned UI_setbgcolor (unsigned color) ASMNAME(UI_setbgcolor);

#undef ASMNAME
#endif

#endif /* USER_SUPPORT == 0 */

#endif /* USER_H */
