/* util.c */

/* Gujin is a bootloader, it loads a Linux kernel from cold boot or DOS.
 * Copyright (C) 1999-2001 Etienne Lorrain, fingerprint (2D3AF3EA):
 *   2471 DF64 9DEE 41D4 C8DB 9667 E448 FF8C 2D3A F3EA
 * E-Mail: etienne.lorrain@masroudeau.com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "make.h"
#if SETUP & EXTRA_SEGMENT	/* before other includes */
#define CALLING_FROM_EXTRASEG
#endif

#include "library.h"
#include "debug.h"
#include "boot.h"	/* bootsector_type */
#include "malloc.h"
#include "disk.h"
#include "util.h"
#include "vmlinuz.h"
#include "gzlib.h"


#if DISK_SUPPORT & DOS_SUPPORT
#include "dos.h"
#endif

#if DISK_SUPPORT & E2FS_PROBE

#define FILE    void *		/* to not have compilation errors */
#define EXT2_FLAT_INCLUDES 0

#include "ext2_fs.h"
#include "ext2fs.h"

#undef FILE

#endif /* E2FS_PROBE */

/*
 * Stuff used all arround:
 */

struct treat_directory_str {
#if DISK_SUPPORT & E2FS_PROBE
    ext2_filsys		fs;
#endif
    unsigned char	disk, part;
    unsigned		slashboot_inode; /* or cluster */
    unsigned		nb; /* already present in array */
    struct desc_str	*array;
    unsigned char	in_slashboot;    /* 0 or 1 */
    struct gujin_param_attrib gujin_attr;
    };

const char *const filename_array[] = {
#ifdef INITRDNAME
    INITRDNAME,		/* initrd filename at first index */
#endif
#ifdef LINUZNAME1
    LINUZNAME1,
#endif
#ifdef LINUZNAME2
    LINUZNAME2,
#endif
#ifdef LINUZNAME3
    LINUZNAME3,
#endif
#ifdef KERNELNAME
    KERNELNAME,	/* kernel filename at last index */
#endif
    };

/*
 * GZLIB stuff & drawing:
 */
SEGMENT_EXTRA(gzlib_extract) static int
gzlib_extract (z_stream *gzlib, unsigned char *buffer, unsigned buflen)
  {bound_stack();{
  static unsigned char cpt;
  char array[3] = { '\010', '#', '\0' };
  int err;

#if defined (LINUZNAME1) || defined (LINUZNAME2) || defined (LINUZNAME3)
  if (gzlib->mode == INIT) {
      cpt = 0;
      print (array);
      err = vmlinuz_header_treat (gzlib->extra, buffer, buflen);
      if (err != 0)
	  return 0xB0 | err;
      if (gzlib->extra->gzip_start == 0)
	  return 0;	/* read more bytes */
      buflen = buffer + buflen - (unsigned char *)gzlib->extra->gzip_start;
      buffer = (unsigned char *)gzlib->extra->gzip_start;
      }
#endif

  gzlib->next_in = buffer;

#ifdef TEST_BYTE_PER_BYTE_INFLATE
  gzlib->avail_in = 1;

  for (;;) {
      array[1] = "|/-\\"[cpt++ % (sizeof ("|/-\\") - 1)];
      print (array);

      err = inflate (gzlib);
      if (err == Z_STREAM_END) {
	  UDBG ((" [inflate returns Z_STREAM_END] "));
	  print ("\010, ");
	  return 0;
	  }
	else if (err != Z_OK) {
	  UDBG (("inflate returns %d!\r\n", err));
	  array[1] = '!';
	  print (array);
	  return 0x90 | err;
	  }

      if (gzlib->avail_in == 0) {
	  if (--buflen == 0)
	      return 0;
	  gzlib->avail_in = 1;
	  }
      }
#else /* TEST_BYTE_PER_BYTE_INFLATE */
  gzlib->avail_in = buflen;

  array[1] = "|/-\\"[cpt++ % (sizeof ("|/-\\") - 1)];
  print (array);

  err = inflate (gzlib);
  if (err == Z_STREAM_END) {
      UDBG ((" [inflate returns Z_STREAM_END] "));
      print ("\010, ");
      return 0;
      }
    else if (err != Z_OK) {
      UDBG (("inflate returns %d!\r\n", err));
      array[1] = '!';
      print (array);
      return 0x90 | err;
      }
    else if (gzlib->avail_in != 0) {
      UDBG (("inflate returns with non empty input!\r\n"));
      array[1] = '?';
      print (array);
      return 0x9F;
      }
  return 0;
#endif /* TEST_BYTE_PER_BYTE_INFLATE */
  }}

/**
 ** E2FS filesystem:
 **/

#if DISK_SUPPORT & E2FS_PROBE
/*
 * All the logic of the E2FS cache subsystem:
 */

#if defined (E2FSCACHE) && (E2FSCACHE != 0)
static struct e2fscache_str {
    unsigned last_read[3 * E2FSCACHE];
    unsigned char disk_letter, part_letter;
    unsigned short nb_saved_read;
    struct e2fsdata_str {
	unsigned block;
	unsigned short size;
	unsigned short lastaccess;
	unsigned *buffer;
	} data[E2FSCACHE];
    } e2fscache;	/* in BSS, zero initialised: */

#if SETUP & BIG_MALLOC
#define E2FSCACHE_MEMCPY_FROMCACHE(data, index, size) \
	memcpy (data, e2fscache.data[index].buffer, size)
#define E2FSCACHE_MEMCPY_TOCACHE(index, data, size) \
	memcpy (e2fscache.data[index].buffer, data, size)
#define E2FSCACHE_FREE(index) \
	free (e2fscache.data[index].buffer)
#define E2FSCACHE_MALLOC(size, i) \
	malloc (size)
#else /* BIG_MALLOC */
#define E2FSCACHE_MEMCPY_FROMCACHE(data, index, size) \
	fmemcpy (data_adr(data), (farptr)e2fscache.data[index].buffer, size)
#define E2FSCACHE_MEMCPY_TOCACHE(index, data, size) \
	fmemcpy ((farptr)e2fscache.data[index].buffer, data_adr(data), size)
#define E2FSCACHE_FREE(index)	/* */

#ifdef GZLIB_BIG_MALLOC
/* here the 32 Kb area after the stack is (will be) taken by the gzlib buffer,
 * so allocate after it: */
#define E2FSCACHE_BASEMALLOC 0x1800U
#else
/* Allocate just after the stack.  */
#define E2FSCACHE_BASEMALLOC 0x1000U
#endif

/* WARNING: ASSUMES that all blocks have the same constant size */
#define E2FSCACHE_MALLOC(size, i) \
	(unsigned *) (((E2FSCACHE_BASEMALLOC + getss()) << 16) + (i * size))
#endif /* BIG_MALLOC */

SEGMENT_EXTRA(E2FSCACHE_init) static inline void
E2FSCACHE_init (unsigned char disk_letter, unsigned char part_letter)
  {
  if (e2fscache.disk_letter == 0 && e2fscache.part_letter == 0) {
      CDBG (("E2FSCACHE start caching disk %c part %c.\r\n",
		disk_letter, part_letter));
      e2fscache = (struct e2fscache_str) {};
      e2fscache.disk_letter = disk_letter;
      e2fscache.part_letter = part_letter;
      }
  }

SEGMENT_EXTRA(E2FSCACHE_end) static inline void
E2FSCACHE_end (unsigned char disk_letter, unsigned char part_letter)
  {
  if (   e2fscache.disk_letter == disk_letter
      && e2fscache.part_letter == part_letter)  {
      /* done also here to save malloced memory */
      unsigned short i;

      for (i = 0; i < nbof(e2fscache.data); i++)
	  if (e2fscache.data[i].buffer)
	      E2FSCACHE_FREE (i);
      CDBG (("E2FSCACHE end caching, saved %u physical reads.\r\n",
		e2fscache.nb_saved_read));
      e2fscache = (struct e2fscache_str) {};
      }
  }

SEGMENT_EXTRA(E2FSCACHE_get_line) static inline unsigned
E2FSCACHE_get_line (unsigned char disk_letter, unsigned char part_letter,
		    unsigned block, unsigned size, void *data)
  {
  if (   e2fscache.disk_letter == disk_letter
      && e2fscache.part_letter == part_letter) {
      unsigned short i;
      unsigned char one_line_free = 0;

      for (i = 0; i < nbof(e2fscache.data); i++) {
	  if (e2fscache.data[i].buffer == 0)
	      one_line_free = 1;
	    else if (   block == e2fscache.data[i].block
		     && size == e2fscache.data[i].size) {
	      e2fscache.data[i].lastaccess = e2fscache.nb_saved_read;
	      e2fscache.nb_saved_read++;
	      E2FSCACHE_MEMCPY_FROMCACHE (data, i, size);
	      CDBG (("E2FSCACHE using cached data block %u size 0x%X "
			"for block %u.\r\n", i, size, block));
	      return 1;
	      }
	  }
      /* not cached: */
      if (!one_line_free) {
	  for (i = 0; i < nbof(e2fscache.last_read); i++)
	      if (e2fscache.last_read[i] == block)
		 break;
	  if (i < nbof(e2fscache.last_read)) {
	      unsigned short minindex = 0;
	      CDBG (("E2FSCACHE block %u physically read not long ago, "
			"need to cache.\r\n", block));
	      for (i = 1; i < nbof(e2fscache.data); i++) {
		  if (e2fscache.data[i].lastaccess
			< e2fscache.data[minindex].lastaccess)
		      minindex = i;
		  }
	      /* e2fscache.data[minindex].buffer != 0 because !one_line_free */
	      CDBG (("E2FSCACHE freing cache line %u block %u.\r\n",
			minindex, e2fscache.data[minindex].block));
	      E2FSCACHE_FREE (minindex);
	      e2fscache.data[minindex] = (struct e2fsdata_str) {};
	      }
	    else
	      CDBG (("E2FSCACHE reading block %u, no need to cache.\r\n",
			block));
	  }

      for (i = 0; i < nbof(e2fscache.last_read) - 1; i++)
	  e2fscache.last_read[i] = e2fscache.last_read[i+1];
      e2fscache.last_read[i] = block;
      }
  return 0;
  }

SEGMENT_EXTRA(E2FSCACHE_put_line) static inline void
E2FSCACHE_put_line (unsigned char disk_letter, unsigned char part_letter,
	    unsigned block, unsigned size, void *data)
  {
  if (   e2fscache.disk_letter == disk_letter
      && e2fscache.part_letter == part_letter) {
      unsigned short i;
      for (i = 0; i < nbof(e2fscache.data); i++) {
	  if (e2fscache.data[i].buffer == 0) {
	      CDBG (("E2FSCACHE filling cache line %u block %u.\r\n",
			 i, block));
	      e2fscache.data[i].buffer = E2FSCACHE_MALLOC(size, i);
	      if (e2fscache.data[i].buffer) {
		  e2fscache.data[i].lastaccess = e2fscache.nb_saved_read;
		  e2fscache.data[i].block = block;
		  e2fscache.data[i].size = size;
		  E2FSCACHE_MEMCPY_TOCACHE(i, data, size);
		  }
	      break;
	      }
	  }
      }
  }

SEGMENT_EXTRA(E2FSCACHE_disable_line) static inline void
E2FSCACHE_disable_line (unsigned char disk_letter, unsigned char part_letter,
			unsigned block, unsigned size_in_block)
  {
  if (   e2fscache.disk_letter == disk_letter
      && e2fscache.part_letter == part_letter) {
      unsigned short i;
      for (i = 0; i < nbof(e2fscache.data); i++) {
	  if (   e2fscache.data[i].buffer != 0
	      && block >= e2fscache.data[i].block
	      && block < e2fscache.data[i].block + size_in_block) {
	      CDBG (("E2FSCACHE invalidate cache line %u block %u.\r\n",
			i, block));
	      E2FSCACHE_FREE (i);
	      e2fscache.data[i] = (struct e2fsdata_str) {};
	      }
	  }
      }
  }

#else  /* E2FSCACHE */

/* We can not declare them as inline functions because their
   parameters do not even exist: */
#define E2FSCACHE_init(disk_letter, part_letter) /* */
#define E2FSCACHE_end(disk_letter, part_letter) /* */
#define E2FSCACHE_get_line(disk_letter, part_letter, block, size, data) 0
#define E2FSCACHE_put_line(disk_letter, part_letter, block, size, data) /* */
#define E2FSCACHE_disable_line(disk_letter, part_letter, block, size_in_block) /* */

#endif /* E2FSCACHE */

/*
 * The driver part of E2FS
 */

extern const struct struct_io_manager struct_gujin_manager;

SEGMENT_EXTRA(time) time_t
time (time_t *timer)
  {
  const time_t null_time = {0};
  if (timer)
      *timer = null_time;
  return null_time;
  }

SEGMENT_EXTRA(gujin_open) static errcode_t
gujin_open (const char *name, int flags, io_channel *channel)
  {bound_stack();{
  io_channel ioch;
  /* inline */ const union {
      unsigned char string[4];
      unsigned      number;
      } disk_signature = {{ 'd', 'i', 's', 'k' }},
	part_signature = {{ 'p', 'a', 'r', 't' }};

  ODBG (("gujin_open name: %s, flags: 0x%X, io_channel: 0x%X.\r\n",
	name, flags, channel));

  if (   *(unsigned *)name != disk_signature.number
      || *(unsigned *)(&name[5]) != part_signature.number
      || name[4] < '0' || name[4] - '0' >= (signed)DI.nbdisk
      || name[9] < '0'
      || name[9] - '0' >= (signed)DI.param[name[4] - '0'].nbpartition
      || name[10] != '\0') {
      ODBG (("name error (disk %c part %c / maxdisk %u maxpart %u)!\r\n",
		name[4], name[9], DI.nbdisk,
		DI.param[name[4] - '0'].nbpartition));
      return 0x10;
      }

  ioch = malloc(sizeof(struct struct_io_channel));
  if (!ioch)
      return 0x20;
  memset (ioch, 0, sizeof (struct struct_io_channel));
  ioch->magic = EXT2_ET_MAGIC_IO_CHANNEL;
  ioch->block_size = EXT2_MIN_BLOCK_SIZE;
	/* remove the const attribute by the cast: */
  ioch->manager = (struct struct_io_manager *)&struct_gujin_manager;
//  ioch->read_error = 0;
//  ioch->write_error = 0;
  ioch->name = malloc (strlen(name) + 1);

  if (!ioch->name) {
      free (channel);
      return 0x30;
      }
  strcpy (ioch->name, name);
  *channel = ioch;
  E2FSCACHE_init (name[4], name[9]);
  return 0;
  }}

SEGMENT_EXTRA(gujin_close) static errcode_t
gujin_close (io_channel channel)
  {bound_stack();{
  ODBG (("gujin_close io_channel: 0x%X.\r\n", channel));

#if DEBUG & DEBUG_IOCHANNEL
  if (channel->magic != EXT2_ET_MAGIC_IO_CHANNEL)
      return 0x10;
#endif

  E2FSCACHE_end (channel->name[4], channel->name[9]);
  free (channel->name);
  free (channel);
  return 0;
  }}

SEGMENT_EXTRA(gujin_set_blksize) static errcode_t
gujin_set_blksize (io_channel channel, int blksize)
  {
  ODBG (("gujin_set_blksize io_channel: 0x%X, blksize: 0x%X.\r\n",
	channel, blksize));

#if DEBUG & DEBUG_IOCHANNEL
  if (channel->magic != EXT2_ET_MAGIC_IO_CHANNEL)
      return 0x10;
#endif

  channel->block_size = blksize;

  return 0;
  }

SEGMENT_EXTRA(gujin_read_blk) static errcode_t
gujin_read_blk (io_channel channel, unsigned long block,
		int count, void *data)
  {bound_stack();{
  unsigned size;
  unsigned short disk, part;
  unsigned char returned;

  ODBG (("%s io_channel: 0x%X, block: 0x%X, count: %d, data: 0x%X: ",
	 __FUNCTION__, channel, block, count, data));

#if DEBUG & DEBUG_IOCHANNEL
  if (channel->magic != EXT2_ET_MAGIC_IO_CHANNEL)
      return 0x10;
#endif

  size = (count < 0) ? -count : count * channel->block_size;

  if (   count > 0
      && E2FSCACHE_get_line (channel->name[4], channel->name[9],
				block, size, data))
      return 0;

  disk = channel->name[4] - '0';
  part = channel->name[9] - '0';

  if (block * (channel->block_size / DI.param[disk].bytepersector)
	+ size / DI.param[disk].bytepersector
	>= DI.param[disk].partition[part].length) {
      ODBG (("Partition start: %u len %u\r\n",
		DI.param[disk].partition[part].start,
		DI.param[disk].partition[part].length));
      if (block * (channel->block_size / DI.param[disk].bytepersector)
		>= DI.param[disk].partition[part].length)
	  ODBG (("ERROR READ sector outside partition: %u, aborted.\r\n",
		block * (channel->block_size
				/ DI.param[disk].bytepersector) ));
	else
	  ODBG (("ERROR READ sector across partition: %u+%u = %u,"
		 " aborted.\r\n",
		 block * (channel->block_size
				/ DI.param[disk].bytepersector),
		 size / DI.param[disk].bytepersector,
		 block * (channel->block_size
				/ DI.param[disk].bytepersector)
		 + size / DI.param[disk].bytepersector));
      return 0x20;
      }

  returned = DISK_readsector (&DI.param[disk],
	block * (channel->block_size / DI.param[disk].bytepersector)
			+ DI.param[disk].partition[part].start,
	size / DI.param[disk].bytepersector,
	data_adr (data));

  if (returned) {
      ODBG (("DISK_readsector error 0x%X.\r\n", returned));
      return 0x30 | returned;
      }

  ODBG (("OK.\r\n"));
  if (count > 0)
      E2FSCACHE_put_line (channel->name[4], channel->name[9],
				 block, size, data);
  return 0;
  }}

SEGMENT_EXTRA(gujin_write_blk) static errcode_t
gujin_write_blk (io_channel channel, unsigned long block,
		 int count, const void *data)
  {bound_stack();{
#ifdef WRITE_ENABLE
  unsigned size;
  unsigned short disk, part;
  unsigned char returned;

  ODBG (("gujin_write_blk io_channel: 0x%X, block: 0x%X,"
	 " count: %d, data: 0x%X.\r\n", channel, block, count, data));

#if DEBUG & DEBUG_IOCHANNEL
  if (channel->magic != EXT2_ET_MAGIC_IO_CHANNEL)
      return 0x10;
#endif

  size = (count < 0) ? -count : count * channel->block_size;

  E2FSCACHE_disable_line (channel->name[4], channel->name[9],
			  block, size / channel->block_size);

  disk = channel->name[4] - '0';
  part = channel->name[9] - '0';

  if (block * (channel->block_size / DI.param[disk].bytepersector)
	+ size / DI.param[disk].bytepersector
	>= DI.param[disk].partition[part].length) {
      ODBG (("Partition start: %u len %u\r\n",
		DI.param[disk].partition[part].start,
		DI.param[disk].partition[part].length));
      if (block * (channel->block_size / DI.param[disk].bytepersector)
		>= DI.param[disk].partition[part].length)
	  ODBG (("ERROR WRITE sector outside partition: %u, aborted.\r\n",
		block * (channel->block_size
				/ DI.param[disk].bytepersector) ));
	else
	  ODBG (("ERROR WRITE sector across partition: %u+%u=%u,"
		 " aborted.\r\n",
		 block * (channel->block_size
				/ DI.param[disk].bytepersector),
		 size / DI.param[disk].bytepersector,
		 block * (channel->block_size
				/ DI.param[disk].bytepersector)
		 + size / DI.param[disk].bytepersector));
      return 0x20;
      }

  returned = DISK_writesector (&DI.param[disk],
	 block * (channel->block_size / DI.param[disk].bytepersector)
		+ DI.param[disk].partition[part].start,
	 size / DI.param[disk].bytepersector,
	 data_adr ((void *)data));
  if (returned != 0) {
      ODBG (("DISK_writesector error 0x%X.\r\n", returned));
      return 0x30 | returned;
      }
  ODBG (("OK.\r\n"));

  return 0;
#else
  ODBG (("gujin_write_blk: write disabled.\r\n"));
  return 0x30;
#endif	/* WRITE_ENABLE */
  }}

SEGMENT_EXTRA(gujin_flush) static errcode_t
gujin_flush (io_channel channel)
  {
  ODBG (("gujin_flush io_channel: 0x%X.\r\n", channel));

#if DEBUG & DEBUG_IOCHANNEL
  if (channel->magic != EXT2_ET_MAGIC_IO_CHANNEL)
      return 0x10;
#endif

  return 0;
  }

//#define NEW_IO_MANAGER
#ifdef NEW_IO_MANAGER
SEGMENT_EXTRA(gujin_write_byte) static errcode_t
gujin_write_byte (io_channel channel, unsigned long offset,
		  int size, const void *data)
  {
  ODBG (("gujin_write_byte: "));
  return gujin_write_blk (channel,
		offset / channel->block_size,
		-size,
		data);
  }
#endif

static const struct struct_io_manager struct_gujin_manager = {
    magic:	EXT2_ET_MAGIC_IO_MANAGER,
    name:	"GUJIN I/O Manager",
    open:	gujin_open,
    close:	gujin_close,
    set_blksize: gujin_set_blksize,
    read_blk:	gujin_read_blk,
    write_blk:	gujin_write_blk,
    flush:	gujin_flush,
#ifdef NEW_IO_MANAGER
    write_byte:	gujin_write_byte
#endif
    };

/*
 * Intermediate part:
 */
SEGMENT_EXTRA(e2fs_analyse_directory_or_file) static int
e2fs_analyse_directory_or_file (struct ext2_dir_entry	*dirent,
				int	offset,
				int	blocksize,
				char	*buf,
				void	*_opaque)
  {bound_stack();{
  struct treat_directory_str *opaque = _opaque;
  struct desc_str *elem;
  unsigned char name_len = ((struct ext2_dir_entry_2 *)dirent)->name_len;
  unsigned short i, len;
  struct ext2_inode the_inode;

  /*
   * ext2fs_check_directory() returns non-zero if it
   * is not a directory _or_ if there was an error (like not
   * enough memory) - and it uses quite a lot of memory...
   * Anyway there can only be _one_ file or directory
   * having the name PATHNAME - we will check that it is
   * really a directory later.
   * Shall we use for ((struct ext2_dir_entry_2 *)dirent)->file_type:
#define EXT2_FT_UNKNOWN		0
#define EXT2_FT_REG_FILE	1
#define EXT2_FT_DIR		2
#define EXT2_FT_CHRDEV		3
#define EXT2_FT_BLKDEV		4
#define EXT2_FT_FIFO		5
#define EXT2_FT_SOCK		6
#define EXT2_FT_SYMLINK		7

#define EXT2_FT_MAX		8
   */
  if (   name_len == sizeof (PATHNAME) - 1
      && !strncmp (dirent->name, PATHNAME, name_len)) {
      opaque->slashboot_inode = dirent->inode;
      return 0;
      }

  if (   !opaque->in_slashboot
      && !opaque->gujin_attr.search_topdir_kernel)
      return 0;

  for (i = 0; i < nbof (filename_array); i++) {
      len = strlen(filename_array[i]);
      if (name_len >= len && !strnicmp (filename_array[i], dirent->name, len))
	  break;
      }

  if (i < nbof (filename_array) && opaque->nb >= NB_DESC_ARRAY) {
      UDBG ((" [%s: no more space for file %s] ", __FUNCTION__, dirent->name));
      return 0xFFFFFFFFU; /* break caller loop */
      }

  elem = &opaque->array[opaque->nb];

  if (i >= nbof (filename_array))
      return 0;
#ifdef INITRDNAME
    else if (i == 0) {
      elem->is_kernel = 0;
      elem->has_vmlinuz_header = 0;
      }
#endif
#ifdef KERNELNAME
    else if (i == lastof(filename_array)) {
      elem->has_vmlinuz_header = 0;
      elem->is_kernel = 1;
      }
#endif
    else {
      elem->has_vmlinuz_header = 1;
      elem->is_kernel = 1;
      }

  /* fill in obvious fields before calling foreign functions: */
  elem->name_offset = len;
  elem->disk = opaque->disk;
  elem->partition = opaque->part;
  elem->in_slashboot = opaque->in_slashboot;
  elem->inode = dirent->inode;

  {
  char *ptr = elem->filename;

  if (opaque->in_slashboot) {
      ptr = STRINIT (ptr, "/" PATHNAME "/");
      elem->name_offset += sizeof ("/" PATHNAME "/") - 1;
      }
    else {
      *ptr++ = '/';
      elem->name_offset++;
      }
  strncpy (ptr, dirent->name, MIN(sizeof (elem->filename) - 1
					- (ptr - elem->filename),
					name_len));
  }

  if (ext2fs_read_inode (opaque->fs, dirent->inode, &the_inode)) {
      /* FIXME: dirent->name is not zero terminated */
      UDBG ((" [%s: error reading inode for file %s] ",
			__FUNCTION__, dirent->name));
      return 0;
      }
//  if (((struct ext2_dir_entry_2 *)dirent)->file_type != EXT2_FT_REG_FILE) {
//  if (LINUX_S_ISDIR(the_inode.i_mode)) {
  if (!LINUX_S_ISREG(the_inode.i_mode)) {
      /* for instance "/initrd" directory */
      UDBG ((" [%s: file %s is not a regular file] ",
		__FUNCTION__, dirent->name));
      return 0;
      }
  if (!ext2fs_inode_has_valid_blocks (&the_inode)) {
      UDBG ((" [%s: file %s has no valid block] ",
		__FUNCTION__, dirent->name));
      return 0;
      }

  UDBG ((" [%s: adding file %s] ", __FUNCTION__, dirent->name));
  elem->filesize = the_inode.i_size;
  opaque->nb++;

  return 0;
  }}

SEGMENT_EXTRA(e2fs_analyse) static inline void
e2fs_analyse (struct treat_directory_str *opaque)
  {
  errcode_t	retval;
  char device_name[] = "disk0part0";

  device_name[4] += opaque->disk;
  device_name[9] += opaque->part;

  retval = ext2fs_open (device_name, 0 /*|EXT2_FLAG_RW*/ /*|EXT2_FLAG_FORCE*/,
			0, 0,
			(struct struct_io_manager *)&struct_gujin_manager,
			&opaque->fs);

  if (retval != 0)
      UDBG (("ext2fs_open(%s) error 0x%X\r\n", device_name, retval));
    else {
      UDBG (("ext2fs_open(%s) OK\r\n", device_name));

#if defined (MAX_DISKNAMESIZE) && (MAX_DISKNAMESIZE != 0)
      strncpy (DI.param[opaque->disk].partition[opaque->part].name,
		opaque->fs->super->s_volume_name,
		sizeof (DI.param[0].partition[0].name)-1);
      /* fs->super->s_last_mounted not usually initialised */
#endif

      opaque->slashboot_inode = 0;
      opaque->in_slashboot = 0;
      retval = ext2fs_dir_iterate (opaque->fs, EXT2_ROOT_INO, 0, 0,
				e2fs_analyse_directory_or_file, opaque);
      if (retval)
	  UDBG (("ext2fs_dir_iterate in root: 0x%X\r\n", retval));

      if (   opaque->slashboot_inode != 0
	  && opaque->gujin_attr.search_subdir_kernel) {
	  unsigned tmp = ext2fs_check_directory (opaque->fs,
						 opaque->slashboot_inode);

	  if (tmp != 0 && tmp != EXT2_ET_NO_DIRECTORY)
	      UDBG (("ext2fs_check_directory returns %u, look at "
			"e2fsprogs-1.19/build/lib/ext2fs/ext2_err.h "
			"what it means...\r\n", tmp));
	  if (tmp == 0) {
	      UDBG ((" [%s: entering directory "PATHNAME"] ", __FUNCTION__));
	      DI.param[opaque->disk].partition[opaque->part].misc.maybe_root = 1;
	      opaque->in_slashboot = 1;
	      retval = ext2fs_dir_iterate (opaque->fs, opaque->slashboot_inode,
			0, 0, e2fs_analyse_directory_or_file, opaque);
	      if (retval)
		  UDBG ((" [%s: ext2fs_dir_iterate in subdir /" PATHNAME
			 ": 0x%X]\r\n", __FUNCTION__, retval));

	      UDBG ((" [%s: exiting directory "PATHNAME"] ", __FUNCTION__));
	      }
	    else
	      UDBG ((" [%s: "PATHNAME": not a directory] ", __FUNCTION__));
	  }

      UDBG (("ext2fs_close(%s)\r\n", device_name));

      retval = ext2fs_close (opaque->fs);
      if (retval != 0)
	  UDBG (("ext2fs_close 0x%X\r\n", retval));
      }
  }

#ifdef NEW_E2FS_READ
struct ctx_str {
    unsigned *filesize;
    z_stream *gzlibptr;
    unsigned char *buffer;
    };

SEGMENT_EXTRA(e2fs_system_file_load) static int
process_file_block (ext2_filsys	fs,
			blk_t		*blocknr,
			e2_blkcnt_t	blockcnt,
			blk_t		ref_block,
			int		ref_offset,
			void		*priv_data)
  {
  struct ctx_str *ctx = (struct ctx_str *)priv_data;
  int returned;

  if (fs->io->block_size > DEFAULT_BUFFER_SIZE) {
      UDBG ((" [%s: buffer not big enought] ", __FUNCTION__));
      return BLOCK_ERROR;
      }

  if (fs->io->manager->read_blk (fs->io, *blocknr, 1, ctx->buffer))
      return BLOCK_ERROR;
  *ctx->filesize += fs->io->block_size;
  /* do something */
  returned = gzlib_extract (ctx->gzlibptr, ctx->buffer, fs->io->block_size);

  if (returned) {
      UDBG ((" [gzlib_extract error 0x%X] ", returned));
      return BLOCK_ABORT;
      }
  return 0;
  }
#endif /* NEW_E2FS_READ */

SEGMENT_EXTRA(e2fs_system_file_load) static inline unsigned
e2fs_system_file_load (unsigned disk, unsigned part,
		       unsigned inode, unsigned *filesize,
		       z_stream *gzlibptr, unsigned char *buffer)
  {
  errcode_t	retval;
  ext2_filsys	fs;
  char device_name[] = "disk0part0";

  device_name[4] += disk;
  device_name[9] += part;

  retval = ext2fs_open (device_name, 0, 0, 0,
		(struct struct_io_manager *)&struct_gujin_manager, &fs);
  if (retval) {
      UDBG ((" [ext2fs_open error: 0x%X] ", retval));
      return 0x100 | retval;
      }

#ifdef NEW_E2FS_READ
  {
  struct ctx_str ctx = {
      filesize: filesize,
      gzlibptr: gzlibptr,
      buffer: buffer
      };
  /* DOME: change inode from unsigned to ext2_ino_t */
  retval = ext2fs_block_iterate2 (fs, inode, BLOCK_FLAG_DATA_ONLY,
					0, process_file_block, &ctx);
  if (retval) {
      UDBG ((" [ext2fs_block_iterate2 error: 0x%X] ", retval));
      if ((retval = ext2fs_close (fs)) != 0)
	  UDBG (("ext2fs_close error: 0x%X\r\n", retval));
      return 0x200 | retval;
      }
  }
#else /* NEW_E2FS_READ */
  {
  ext2_file_t	file;
  unsigned got;

  retval = ext2fs_file_open (fs, inode, 0, &file);
  if (retval) {
      UDBG ((" [ext2fs_file_open error: 0x%X] ", retval));
      if ((retval = ext2fs_close (fs)) != 0)
	  UDBG (("ext2fs_close error: 0x%X\r\n", retval));
      return 0x200 | retval;
      }

  while ((retval = ext2fs_file_read (file, buffer, DEFAULT_BUFFER_SIZE, &got))
		== 0) {
      int returned;

      UDBG ((" [read %u bytes] ", got));
      *filesize += got;
      /* do something */
      returned = gzlib_extract (gzlibptr, buffer, got);

      if (returned) {
	  UDBG ((" [E2FS_ReadFile treatfct error 0x%X] ", returned));
	  return 0x300 | returned;
	  }

      if (got != DEFAULT_BUFFER_SIZE) {
	 UDBG ((" [normal exit, read %u bytes when %u requested,"
		" total read: %u] ",
		got, DEFAULT_BUFFER_SIZE, *filesize));
	 break;
	 }
      }

  if (retval) {
      UDBG ((" [ext2fs_file_read error 0x%X] ", retval));
      if ((retval = ext2fs_file_close (file)) != 0)
	  UDBG ((" [ext2fs_file_close error: 0x%X] ", retval));
      if ((retval = ext2fs_close (fs)) != 0)
	  UDBG (("ext2fs_close error: 0x%X\r\n", retval));
      return 0x400 | retval;
      }

  retval = ext2fs_file_close (file);
  if (retval) {
      UDBG ((" [ext2fs_file_close error: 0x%X] ", retval));
      if ((retval = ext2fs_close (fs)) != 0)
	  UDBG (("ext2fs_close error: 0x%X\r\n", retval));
      return 0x500 | retval;
      }
  }
#endif /* NEW_E2FS_READ */

  retval = ext2fs_close (fs);
  if (retval) {
      UDBG (("ext2fs_close error: 0x%X\r\n", retval));
      return 0x600 | retval;
      }

  return 0;
  }
#endif /* E2FS_PROBE */

/**
 ** FAT* filesystem:
 **/
#if DISK_SUPPORT & (FAT12_PROBE|FAT16_PROBE|FAT32_PROBE)

struct sector_chain_str {
    unsigned lba;
    unsigned nb;	/* end marker = 0 */
    };

SEGMENT_EXTRA(FAT_get_filename) static inline const directory_type *
FAT_get_filename (char *filename, const directory_type *direlem)
  {
  unsigned short i;
  const char *ptr;

  /* Has probably to be modified for FAT32/long filenames: */
  ptr = direlem->name;
  for (i = sizeof (direlem->name); --i  && ptr[i] == ' ';)
      /* nothing */;
  for (i += 2; --i;)
      *filename++ = *ptr++;

  ptr = direlem->extension;
  for (i = sizeof (direlem->extension); --i && ptr[i] == ' ';)
      /* nothing */;
  if (i != 0 || ptr[i] != ' ') {
      if (!direlem->attributes.volumelabel)
	  *filename++ = '.';
      for (i += 2; --i;)
	  *filename++ = *ptr++;
      }
  *filename = '\0';
  return ++direlem;
  }

SEGMENT_EXTRA(FAT_get_parameter) static inline unsigned
FAT_get_parameter (unsigned disk, unsigned part,
		    unsigned short  *Bytepersector,
		    unsigned        *start_fat,
		    unsigned        *end_fat,
		    unsigned        *start_dir,
		    unsigned        *start_data,
		    unsigned short  *fatbit,
		    unsigned short  *Sectorpercluster,
		    bootsector_type *bootsector)
  {
  unsigned totalsector;

  if (DISK_readsector (&DI.param[disk], DI.param[disk].partition[part].start,
			1, data_adr(bootsector)) != 0) {
      UDBG ((" [%s: error read MBR]\r\n", __FUNCTION__));
      return 1;
      }

  if (bootsector->after.Signature0xAA55 != 0xAA55) {
      UDBG ((" [%s: did not find 0xAA55 signature]\r\n", __FUNCTION__));
      return 2;
      }

  /* Is that true for FAT32: */
  if (bootsector->before.Signaturebyte0x29 != 0x29) {
      UDBG ((" [%s: did not find 0x29 signature]\r\n", __FUNCTION__));
      return 3;
      }

  totalsector = bootsector->before.NbTotalsector?
	bootsector->before.NbTotalsector : bootsector->before.NbTotalsector2;
  /*
   * If the compiler is good, and there is only _one_ of FAT12_PROBE
   * and FAT16_PROBE, the constant "*fatbit" will be propagated
   * _very_ deeply...
   */
  if (totalsector / bootsector->before.Sectorpercluster < 0xFF1)
#if DISK_SUPPORT & FAT12_PROBE
      *fatbit = 12;
#else
      return 4;
#endif
    else if (totalsector / bootsector->before.Sectorpercluster < 0xFFF1)
#if DISK_SUPPORT & FAT16_PROBE
      *fatbit = 16;
#else
      return 5;
#endif
    else
#if DISK_SUPPORT & FAT32_PROBE
      *fatbit = 28;
#else
      return 6;
#endif

  /* Use FAT 1 */
  *start_fat = DI.param[disk].partition[part].start
		+ bootsector->before.Reservedsector;
  *end_fat = *start_fat + bootsector->before.NbSectorperFAT;
  *Bytepersector = bootsector->before.Bytepersector;
  *Sectorpercluster = bootsector->before.Sectorpercluster;
  *start_dir = *start_fat + bootsector->before.NbFAT
				* bootsector->before.NbSectorperFAT;
  *start_data = *start_dir
		+ DIVROUNDUP (bootsector->before.NbRootdirentry
					 * sizeof (directory_type),
				*Bytepersector);

  UDBG (("Disk %u part %u: start_fat %u, end_fat %u, Bytepersector %u,"
	 " Sectorpercluster %u, start_dir %u, start_data %u\r\n",
	 disk, part, *start_fat, *end_fat, *Bytepersector,
	 *Sectorpercluster, *start_dir, *start_data));

  return 0;
  }

SEGMENT_EXTRA(FAT_get_next_inode) static inline unsigned
FAT_get_next_inode (unsigned char disk, unsigned Bytepersector,
		     unsigned inode, unsigned char fatbit_div_4,
		     unsigned mask,
		     unsigned start_fat, unsigned end_fat,
		     unsigned *base_sector_cached, unsigned nb_sector_cached,
		     unsigned char *sector_cache)
  {
  unsigned quartet_offset, next_inode, lba;
  unsigned start_quartet_cached = 2 * *base_sector_cached * Bytepersector;
  unsigned end_quartet_cached = start_quartet_cached
				+ 2 * nb_sector_cached * Bytepersector;

  quartet_offset = inode * fatbit_div_4;

  if (   quartet_offset < start_quartet_cached
      || quartet_offset + fatbit_div_4 >= end_quartet_cached) {
      *base_sector_cached = quartet_offset/2 / Bytepersector;
      start_quartet_cached = 2 * *base_sector_cached * Bytepersector;
      lba = start_fat + *base_sector_cached;

      if (lba >= end_fat) {
	  UDBG ((" [%s: error calculating FAT: %u > %u]\r\n", __FUNCTION__,
		lba, end_fat));
	  return 0xFFFFFFFEU;
	  }

      UDBG ((" [%s: reading FAT: %u sectors at %u] ", __FUNCTION__,
		nb_sector_cached, lba));

      if (DISK_readsector (&DI.param[disk], lba, nb_sector_cached,
				data_adr(sector_cache)) != 0) {
	  UDBG ((" [%s: error reading FAT at sector %u]\r\n",
		__FUNCTION__, lba));
	  return 0xFFFFFFFFU;
	  }
      }

  next_inode = *(unsigned *)(sector_cache +
			((quartet_offset - start_quartet_cached)/2));

  if ((quartet_offset - start_quartet_cached) & 1)
      next_inode >>= 4;
  return next_inode & mask;
  }

SEGMENT_EXTRA(FAT_get_sector_chain) static inline struct sector_chain_str *
FAT_get_sector_chain (unsigned disk, unsigned part, unsigned inode,
		       unsigned short fatbit, unsigned short Bytepersector,
		       unsigned start_data, unsigned start_dir,
		       unsigned start_fat, unsigned end_fat,
		       unsigned short Sectorpercluster,
		       unsigned short fat_buffer_len,
		       unsigned char *fat_buffer
#if defined (FAT_MAX_FRAGMENT) && (FAT_MAX_FRAGMENT != 0)
		       , struct sector_chain_str *returned
#endif
		       )
  {
#if !defined (FAT_MAX_FRAGMENT) || (FAT_MAX_FRAGMENT == 0)
  struct sector_chain_str *returned;
#endif
  unsigned short current;
  unsigned mask = (1U << fatbit) - 1;
  unsigned char fatelem_size_quartet = fatbit/4;
  unsigned lba;
  /* used because FAT is too big to fit completely in memory: */
  unsigned fat_buffer_sector = 0xFFFFFU; /* guess a wrong value */

#if DISK_SUPPORT & FAT32_PROBE
  if (fatelem_size_quartet == 7) /* FAT32 unit is 32 bits */
      fatelem_size_quartet = 8;
#endif /* FAT32_PROBE */

  if (inode == 0) {
      UDBG (("%s: request chain of inode 0!\r\n", __FUNCTION__));
      return 0;
      }

#if !defined (FAT_MAX_FRAGMENT) || (FAT_MAX_FRAGMENT == 0)
  /* alloc also for the end marker: */
  returned = malloc (2 * sizeof (struct sector_chain_str));
  if (returned == 0) {
      UDBG (("%s: malloc failed!\r\n", __FUNCTION__));
      return 0;
      }
#endif

  current = 0;
  if (inode >= 2) {
      returned[current].lba = start_data + (inode - 2) * Sectorpercluster;
      returned[current].nb = Sectorpercluster;
      }
    else { /* == 1, ask for chain of root directory */
      returned[0].lba = start_dir;
      returned[0].nb = start_data - start_dir;
      returned[1].lba = 0; /* safety */
      returned[1].nb = 0;
      UDBG (("%s returns [[%u, +%u][0,0]]\r\n",
		__FUNCTION__, returned[0].lba, returned[0].nb));

      return returned;
      }

  for (;;) {
      inode = FAT_get_next_inode (disk, Bytepersector,
				inode, fatelem_size_quartet, mask,
				start_fat, end_fat,
				&fat_buffer_sector,
				fat_buffer_len/Bytepersector,
				fat_buffer);

      if (inode & ~mask) {
	  UDBG ((" [%s: error in FAT cache!] ", __FUNCTION__));
#if !defined (FAT_MAX_FRAGMENT) || (FAT_MAX_FRAGMENT == 0)
	  free (returned);
#endif
	  return 0;
	  }
      if (inode == 0) {
	  UDBG ((" [%s: FAT chain contains free blocks!] ", __FUNCTION__));
#if !defined (FAT_MAX_FRAGMENT) || (FAT_MAX_FRAGMENT == 0)
	  free (returned);
#endif
	  return 0;
	  }
      if (inode > (mask & ~0x000F)) {
	  if (inode < (mask & ~0x0007)) {
	      UDBG ((" [%s: FAT chain contains bad blocks!] ", __FUNCTION__));
#if !defined (FAT_MAX_FRAGMENT) || (FAT_MAX_FRAGMENT == 0)
	      free (returned);
#endif
	      return 0;
	      }
	    else {
	      struct sector_chain_str *ptr;
	      UDBG ((" [%s: ", __FUNCTION__));
	      current++;
	      for (ptr = returned; ptr < &returned[current]; ptr++)
		  UDBG (("%u sectors at %u, ", ptr->nb, ptr->lba));
	      UDBG (("end with 0x%X] ", inode));
	      returned[current].nb = 0;
	      returned[current].lba = 0; /* safety */
	      return returned;
	      }
	  }

      lba = start_data + (inode - 2) * Sectorpercluster;

      if (lba == returned[current].lba + returned[current].nb) {
	  returned[current].nb += Sectorpercluster;
	  }
	else {
	  struct sector_chain_str *ptr;
#if !defined (FAT_MAX_FRAGMENT) || (FAT_MAX_FRAGMENT == 0)
	  ptr = realloc (returned,
			 (current+3) * sizeof (struct sector_chain_str));
	  if (ptr == 0) {
	      UDBG (("%s: realloc failed!\r\n", __FUNCTION__));
	      free (returned);
	      return 0;
	      }
	  returned = ptr;
#else
	  if (current >= FAT_MAX_FRAGMENT) {
	      UDBG (("%s: FAT file has too many fragments!\r\n", __FUNCTION__));
	      return 0;
	      }
#endif
	  for (ptr = returned; ptr <= &returned[current]; ptr++) {
	      if (ptr->lba == lba) {
		  UDBG (("%s: FAT loop detected!\r\n", __FUNCTION__));
#if !defined (FAT_MAX_FRAGMENT) || (FAT_MAX_FRAGMENT == 0)
		  free (returned);
#endif
		  return 0;
		  }
	      }
	  current ++;
	  returned[current].lba = lba;
	  returned[current].nb = Sectorpercluster;
	  }
      }
  }

SEGMENT_EXTRA(FAT_file_treat) static unsigned
FAT_file_treat (unsigned disk, unsigned part,
		 unsigned inode, unsigned *filesize,
		 int (*treat) (void *, unsigned char *, unsigned),
		 void *opaque,
		 unsigned char *buffer, unsigned buflen)
  {bound_stack();{
  struct sector_chain_str *current, *chain;
#if defined (FAT_MAX_FRAGMENT) && (FAT_MAX_FRAGMENT != 0)
  struct sector_chain_str bigarray[FAT_MAX_FRAGMENT];
#endif

  unsigned start_fat, end_fat, start_dir, start_data, tmp;
  unsigned short fatbit, Sectorpercluster, bytepersector;

  /* here buflen = sizeof(buffer) has to be >= sizeof (bootsector_type) */

  if ((tmp = FAT_get_parameter (disk, part, &bytepersector,
			&start_fat, &end_fat, &start_dir, &start_data,
			&fatbit, &Sectorpercluster,
			(bootsector_type *)buffer)) != 0) {
      UDBG ((" [%s: error 0x%X in FAT_get_parameter!] ",
		__FUNCTION__, tmp));
      return 0x100 | tmp;
      }

  buflen = (buflen / bytepersector) * bytepersector;

  chain = FAT_get_sector_chain (disk, part, inode, fatbit,
			bytepersector, start_data, start_dir,
			start_fat, end_fat, Sectorpercluster,
			buflen, buffer
#if defined (FAT_MAX_FRAGMENT) && (FAT_MAX_FRAGMENT != 0)
			, bigarray
#endif
			);

  if (chain == 0) {
      UDBG (("%s: cannot get LBA chain!\r\n", __FUNCTION__));
      return 0x200;
      }
  current = chain;
  for (;;) {
      unsigned short nbread = 0;
      int tmp;

      while (nbread < buflen && current->nb != 0) {
	  unsigned short nb = MIN ((buflen-nbread) / bytepersector,
					current->nb);

	  if ((tmp = DISK_readsector (&DI.param[disk], current->lba,
					nb, data_adr(buffer + nbread))) != 0) {
	      UDBG ((" [%s: error reading file at LBA=%u]\r\n",
			__FUNCTION__, current->lba));
#if !defined (FAT_MAX_FRAGMENT) || (FAT_MAX_FRAGMENT == 0)
	      free (chain);
#endif
	      return 0x300 | tmp;
	      }

	  nbread += nb * bytepersector;
	  current->lba += nb;
	  current->nb -= nb;
	  if (current->nb == 0)
	      current++;
	  }
      *filesize += nbread;
      tmp = (*treat) (opaque, buffer, nbread);

      if (tmp != 0) {
	  UDBG ((" [%s: treat function returns %d] ", __FUNCTION__, tmp));
#if !defined (FAT_MAX_FRAGMENT) || (FAT_MAX_FRAGMENT == 0)
	  free (chain);
#endif
	  return 0x400 | tmp;
	  }
      if (current->nb == 0) {
	  UDBG (("%s: done, read %u bytes\r\n", __FUNCTION__, *filesize));
#if !defined (FAT_MAX_FRAGMENT) || (FAT_MAX_FRAGMENT == 0)
	  free (chain);
#endif
	  return 0;
	  }
      }
  }}

SEGMENT_EXTRA(FAT_system_file_load) static inline unsigned
FAT_system_file_load (unsigned disk, unsigned part,
		       unsigned inode, unsigned *filesize,
		       z_stream *gzlibptr, unsigned char *buffer)
  {
  typedef int (*fctptr) (void *, unsigned char *, unsigned);
  return FAT_file_treat (disk, part, inode, filesize,
			  (fctptr) gzlib_extract,
			  gzlibptr,
			  buffer, DEFAULT_BUFFER_SIZE);
  }

SEGMENT_EXTRA(FAT_treat_directory) static int
FAT_treat_directory (struct treat_directory_str *opaque,
		      const directory_type *dir, unsigned len)
  {bound_stack();{
  unsigned short i, namelen;
  char name[NAME_LENGTH];
  const directory_type *nextdir = dir,
		       *lastdir = dir + (len / sizeof(directory_type));
  struct desc_str *elem;

  while ((dir = nextdir) < lastdir) {
      if (dir->attributes.unused) {
	  UDBG ((" [%s: not a directory sector] ", __FUNCTION__));
	  return -1;
	  }
      if (dir->name[0] == (char)0xE5) {
	  //UDBG ((" [erased file] "));
	  nextdir++; /* CHECKME: erase of a long filename? */
	  continue;
	  }
      if (dir->name[0] == 0) {
	  UDBG ((" [%s: normal end of directory] ", __FUNCTION__));
	  return 1; /* can stop processing of this directory */
	  }
      /* FIXME: if the filename spread over the "len" parameter?
	  i.e. DEFAULT_BUFFER_SIZE bytes (long filename only). */
      nextdir = FAT_get_filename (name, dir);
#if defined (MAX_DISKNAMESIZE) && (MAX_DISKNAMESIZE != 0)
      {
      /* inline */ const struct file_attr_str volumeattr = {
		readonly	: 0,	hidden		: 0,
		system		: 0,	volumelabel	: 1,
		subdirectory	: 0,	archive		: 1,
		unused		: 0
		};
      if (*(unsigned char *)&dir->attributes
		== *(unsigned char *)&volumeattr) {
	  UDBG ((" [volumename: %s] ", name));
	  strncpy (DI.param[opaque->disk].partition[opaque->part].name,
		   name, sizeof (DI.param[0].partition[0].name)-1);
	  continue;
	  }
      }
#endif
      if (   !stricmp (name, PATHNAME) /* only one in this directory */
	  && dir->attributes.subdirectory) {
	  UDBG ((" [boot directory %s at %u] ", name, dir->cluster));
	  opaque->slashboot_inode = dir->cluster;
	  continue;
	  }

      if (   !opaque->in_slashboot
	  && !opaque->gujin_attr.search_topdir_kernel)
	  continue;

      if (dir->attributes.volumelabel || dir->attributes.subdirectory) {
	  UDBG ((" [ignoring %s] ", name));
	  continue;
	  }
      for (i = 0; i < nbof (filename_array); i++) {
	  namelen = strlen (filename_array[i]);
	  if (!strnicmp (filename_array[i], name, namelen))
	      break;
	  }

      if (i < nbof (filename_array) && opaque->nb >= NB_DESC_ARRAY) {
	  UDBG ((" [%s: no more space for file %s] ", __FUNCTION__, name));
	  return -2;
	  }

      elem = &opaque->array[opaque->nb];

      if (i >= nbof (filename_array)) {
	  UDBG ((" [ignore: %s] ", name));
	  continue;
	  }
#ifdef INITRDNAME
	else if (i == 0) {
	  UDBG ((" [initrd: %s] ", name));
	  elem->has_vmlinuz_header = 0;
	  elem->is_kernel = 0;
	  }
#endif
#ifdef KERNELNAME
	else if (i == lastof(filename_array)) {
	  UDBG ((" [kernel: %s] ", name));
	  elem->has_vmlinuz_header = 0;
	  elem->is_kernel = 1;
	  }
#endif
	else {
	  UDBG ((" [vmlinuz: %s] ", name));
	  elem->has_vmlinuz_header = 1;
	  elem->is_kernel = 1;
	  }
      elem->in_slashboot = opaque->in_slashboot;
      elem->disk = opaque->disk;
      elem->partition = opaque->part;
      elem->filesize = dir->size;
      elem->inode = dir->cluster;
      elem->name_offset = namelen;

      {
      char *ptr = elem->filename;

      if (opaque->in_slashboot) {
	  ptr = STRINIT (ptr, "/" PATHNAME "/");
	  elem->name_offset += sizeof ("/" PATHNAME "/") - 1;
	  }
	else {
	  *ptr++ = '/';
	  elem->name_offset++;
	  }
      strncpy (ptr, name, sizeof (elem->filename) - 1
				- (ptr - elem->filename));
      strlwr (ptr);    /* just for display */
      }

      opaque->nb++;
      }
  UDBG ((" [%s: need to read more sectors] ", __FUNCTION__));
  return 0;
  }}

/*
 * Do not declare this inline because of the stack it uses:
 * (stack space is locked and e2fs_analyse needs it)
 */
SEGMENT_EXTRA(FAT_analyse) static void
FAT_analyse (struct treat_directory_str *opaque)
  {bound_stack();{
  typedef int (*fctptr) (void *, unsigned char *, unsigned);
  unsigned filesize;
  unsigned char buffer[DEFAULT_BUFFER_SIZE];

#if defined (MAX_DISKNAMESIZE) && (MAX_DISKNAMESIZE != 0)
  DI.param[opaque->disk].partition[opaque->part].name[0] = 0;
#endif

  UDBG (("%s: scan main directory ", __FUNCTION__));

  opaque->in_slashboot = 0;
  filesize = 0;
  FAT_file_treat (opaque->disk,
		  opaque->part,
		  1 /* root directory */,
		  &filesize,
		  (fctptr)FAT_treat_directory,
		  opaque,
		  buffer,
		  DEFAULT_BUFFER_SIZE);

  if (   opaque->slashboot_inode != 0
      && opaque->gujin_attr.search_subdir_kernel) {
      UDBG ((", scan /" PATHNAME " directory "));
      filesize = 0;
      opaque->in_slashboot = 1;
      FAT_file_treat (opaque->disk,
		      opaque->part,
		      opaque->slashboot_inode,
		      &filesize,
		      (fctptr)FAT_treat_directory,
		      opaque,
		      buffer,
		      DEFAULT_BUFFER_SIZE);
      }
  UDBG ((", end scan.\r\n"));
  }}
#endif /* FAT12_PROBE|FAT16_PROBE|FAT32_PROBE */

#if DISK_SUPPORT & DOS_SUPPORT
SEGMENT_EXTRA(DOS_analyse) static inline void
DOS_analyse (struct treat_directory_str *opaque)
  {
  unsigned short i, error;
  struct DOS_DTA_str DosDta;
  /* inline */ const struct DOS_attribute_str attribute = {
	read_only    : 1,  hidden    : 1, system  : 1,
	volume_label : 0,  directory : 0, archive : 1
	};

  if (sizeof (DosDta.name_ext) + sizeof (PATHNAME "/")
		>= sizeof (opaque->array->filename))
      __ERROR();

  UDBG ((" [%s: disk %u, name %c:] ", __FUNCTION__,
		 opaque->disk, DI.param[opaque->disk].disknb));
  DOS_SetDTA (&DosDta);

#if defined (MAX_DISKNAMESIZE) && (MAX_DISKNAMESIZE != 0)
  {
 /*
  * Dirty if "MAX_DISKNAMESIZE != 0" and access == dos_part,
  * the "partition" pointer is used as the DOS name pointer
  * because nbpartition == 0.
  */
  /* inline */ const struct DOS_attribute_str vol_label = {
	read_only    : 0,  hidden    : 0, system  : 0,
	volume_label : 1,  directory : 0, archive : 0
	};
#ifdef DISK_USE_MALLOC
  if (DOS_FindFirst ("*.*", vol_label, &error) == 0) {
      DI.param[opaque->disk].partition = malloc(MAX_DISKNAMESIZE+1);
      strncpy ((char *)DI.param[opaque->disk].partition, DosDta.name_ext,
		MAX_DISKNAMESIZE);
      }
    else
      DI.param[opaque->disk].partition = 0;
#else
  if (DOS_FindFirst ("*.*", vol_label, &error) == 0) {
      strncpy ((char *)DI.param[opaque->disk].partition, DosDta.name_ext,
		MIN (MAX_DISKNAMESIZE,
		     sizeof(DI.param[opaque->disk].partition)));
      }
    else
      DI.param[opaque->disk].partition[0].start = 0;
#endif
  }
#endif

  for (i = 0; i < 2 * nbof(filename_array); i++) {
      char pattern[NAME_LENGTH], *pat;

      if (i & 1) {
	  if (!opaque->gujin_attr.search_subdir_kernel)
	      continue;
	  }
	else {
	  if (!opaque->gujin_attr.search_topdir_kernel)
	      continue;
	  }
      pat = pattern;
      *pat++ = DI.param[opaque->disk].disknb; /* the letter 'C' */
      pat = STRINIT (pat, ":/");
      if (i & 1)
	  pat = STRINIT (pat, PATHNAME "/");
      strcpy (pat, filename_array[i >> 1]);
      strcat (pat, "*.*");

      //UDBG ((" [%s: searching \"%s\"] ", __FUNCTION__, pattern));
      if (DOS_FindFirst (pattern, attribute, &error) != 0)
	  UDBG ((" [%s: failed DOS_FindFirst, error 0x%X] ",
			__FUNCTION__, error));
	else {
	  for (;;) {
	      struct desc_str *elem = &opaque->array[opaque->nb];

	      if (opaque->nb >= NB_DESC_ARRAY) {
		  UDBG ((" [%s: search finnished early]\r\n", __FUNCTION__));
		  return;
		  }
	      elem->disk = opaque->disk;
	      elem->partition = 0;

	      elem->in_slashboot = i & 1;
	      elem->name_offset = strlen (filename_array[i >> 1]);
	      elem->name_offset += 1; /* "C:/", keep only "/" */
	      if (elem->in_slashboot)
		  elem->name_offset += sizeof (PATHNAME "/") - 1;
	      elem->filesize = DosDta.size;
#ifdef INITRDNAME
	      /* initrd for i = 0 and 1 */
	      elem->is_kernel = ((unsigned)(i >> 1) >= 1);
#else
	      elem->is_kernel = 1;
#endif
	      elem->has_vmlinuz_header = elem->is_kernel;
#ifdef KERNELNAME
	      elem->has_vmlinuz_header &= ((unsigned)(i >> 1) < lastof(filename_array));
#endif

	      {
	      char *ptr = elem->filename;
	      /* pattern is a kind of "C:/boot/vmlinuz*.*",
		 pat points into pattern at the v */
	      memcpy (ptr, pattern + 2, pat - pattern - 2); /* ignore "C:" */
	      ptr += pat - pattern - 2;
	      strncpy (ptr, DosDta.name_ext, MIN (sizeof (DosDta.name_ext),
						sizeof (elem->filename) - 1
						- (ptr - elem->filename)));
	      strlwr (ptr);
	      }

	      UDBG ((" [%s: adding \"%s\"]", __FUNCTION__, elem->filename));
	      opaque->nb++;
	      if (DOS_FindNext (&error)) {
		  UDBG ((" [next pattern] "));
		  break;
		  }
	      }
	  }
      }
  UDBG ((" [%s: search finnished]\r\n", __FUNCTION__));
  }

SEGMENT_EXTRA(DOS_system_file_load) static inline unsigned
DOS_system_file_load (unsigned disk, char *filename, unsigned *filesize,
		      z_stream *gzlibptr, unsigned char *buffer)
  {
  unsigned short handle, nbread_errorcode;
  int returned;
  /* declare that in %ss, i.e. in %ds < 64 K, remember SETUP & BIG_MALLOC ! */
  char fullfilename[NAME_LENGTH];

  fullfilename[0] = DI.param[disk].disknb; /* the letter, i.e. 'C' */
  fullfilename[1] = ':';
  strcpy (&fullfilename[2], filename);

  if (DOS_OpenFile (fullfilename, DOS_open_read_only, &handle) != 0) {
      UDBG ((" [DOS_OpenFile %s failed, error 0x%X] ",
		fullfilename, handle));
      return 0x100;
      }

  /* we cannot have buffer malloced if using BIG_MALLOC here: */
  while (DOS_ReadFile (handle, buffer, DEFAULT_BUFFER_SIZE, &nbread_errorcode)
		== 0) {
      UDBG ((" [read %u bytes] ", nbread_errorcode));
      *filesize += nbread_errorcode;
      /* do something */
      returned = gzlib_extract (gzlibptr, buffer, nbread_errorcode);

      if (returned) {
	  UDBG ((" [DOS_ReadFile treatfct error 0x%X] ", returned));
	  return 0x200 | returned;
	  }

      if (nbread_errorcode != DEFAULT_BUFFER_SIZE) {
	  UDBG ((" [DOS_ReadFile loop normal exit, read %u bytes "
		 "when %u requested, total read: %u] ",
		nbread_errorcode, DEFAULT_BUFFER_SIZE, *filesize));
	  nbread_errorcode = DEFAULT_BUFFER_SIZE; /* used as a marker */
	  break;
	  }
      }

  if (nbread_errorcode != DEFAULT_BUFFER_SIZE) {
      UDBG ((" [DOS_ReadFile failed, error 0x%X] ", nbread_errorcode));
      if (DOS_CloseFile (handle, &nbread_errorcode) != 0)
	  UDBG ((" [DOS_CloseFile failed after error, error 0x%X] ",
		 nbread_errorcode));
      return 0x300;
      }

  if (DOS_CloseFile (handle, &nbread_errorcode) != 0) {
      UDBG ((" [DOS_CloseFile failed, error 0x%X] ", nbread_errorcode));
      return 0x400;
      }
  return 0;
  }
#endif /* DOS_SUPPORT */

/*
Standart DOS MBR:
0x10F : "Invalid partition table"
0x127 : "Error loading operating system"
0x146 : "Missing operating system"
0x1DC..0x1FD : filled with zeroes
0x1FE : 0x55 0xAA

Standart DOS bootsector (not MBR):
0x003 : "MSDOS5.0"
0x02B : 'NO NAME    ' (not zero terminated)
0x036 : 'FAT12   '
0x1A0 : "Disque non-syst\x8Ame/erreur disque\r\nRemplacez et pressez touche\r\n"
0x1E0 : 'IO      SYS'
0x1EB : 'MSDOS   SYS'
0x1FE : 0x55 0xAA

Standart LILO MBR and bootsector:
0x006 : 'LILO'
0x1FE : 0x55 0xAA

Standart LILO in LBA MBR and bootsector:
0x003 : 'lbaLILO'
0x1FE : 0x55 0xAA

Another LILO MBR and bootsector:
0x005 : 'zLILO'
0x1FE : 0x55 0xAA

Extended partition by fdisk:
0x000..0x1BD : 0
0x1FE : 0x55 0xAA

Swap:
0x000 : 0xFE
0x001..0x200 : 0xFF

Standart DOSEMU MBR:
0x000 : "DOSEMU"
0x1FE : 0x00 0x00

Standart DOSEMU bootsector:
0x003 : "DOSEMU  "
0x02B : 'DOSEMU     '
0x036 : 'FAT12   '
0x043..0x1FD : filled with 0
0x1FE : 0x55 0xAA

Standart FREEDOS MBR:
0x000 : "DOSEMU"
0x1FE : 0x00 0x00

Standart FREEDOS bootsector:
0x003 : zeroes
0x02B : 'FREEDOS    '
0x036 : 'FAT12   '
0x1E4 : 'Boot error' (no terminator)
0x1EE : "IPL     SYS"
0x1FE : 0x55 0xAA

XOSLOAD:
0x181 : 'XOSLLOAD'

NT:
0x1E4 : "NTLDR is compressed.\r\n"
 */

#if DISK_SUPPORT & BOOTSECT_PROBE

SEGMENT_EXTRA(add_bootsector) static inline void
add_bootsector (struct treat_directory_str *opaque)
  {bound_stack();{
  bootsector_type bootsect;
  const unsigned char *ptr;
  struct desc_str *elem;
  unsigned cpt;
  lba_t bootsector_lba;

  if (opaque->slashboot_inode == 0 && opaque->gujin_attr.prefer_enddisk_mbr)
      bootsector_lba = DI.param[opaque->disk].nbtotalsector - 1;
    else
      bootsector_lba = opaque->slashboot_inode;

  for (;;) {
      if (   DISK_readsector (&DI.param[opaque->disk], bootsector_lba,
			1, stack_adr (&bootsect)) == 0
	  && bootsect.after.Signature0xAA55 == 0xAA55)
	  break;
      if (opaque->slashboot_inode != 0)
	  return;
      if (opaque->gujin_attr.prefer_enddisk_mbr) {
	  if (bootsector_lba == 0)
	      return;
	  bootsector_lba = 0;
	  }
	else {
	  if (bootsector_lba != 0)
	      return;
	  bootsector_lba = DI.param[opaque->disk].nbtotalsector - 1;
	  }
      }

  if (STREQL ((char *)&bootsect, "DOSEMU"))
      return;
    else if (STRBEGINEQL (bootsect.before.String, "DOSEMU"))
      return;
    else if (STRBEGINEQL ((char *)&bootsect + 0x78, "Non-System disk"))
      return;
    else if (STREQL (bootsect.before.String, "mkdosfs"))
      return;
    else if (STRBEGINEQL (&bootsect.before.String[2], "zLILO"))
      ptr = "zLILO";
    else if (STRBEGINEQL (bootsect.before.String, "lbaLILO"))
      ptr = "lbaLILO";
    else if (STRBEGINEQL (&bootsect.before.String[3], "LILO"))
      ptr = "LILO";
    else if (STRBEGINEQL ((char *)&bootsect + 2, "LILO"))
      ptr = "oldLILO";
    else if (STRBEGINEQL ((char *)&bootsect + 0x181, "XOSLLOAD"))
      ptr = "XoslLoad";
    else {
      for (ptr = bootsect.before.String;
	      ptr < &bootsect.before.String[sizeof (bootsect.before.String)]
	   && *ptr == 0;
	   ptr++) ;
      if (ptr >= &bootsect.before.String[sizeof (bootsect.before.String)])
	  return;
      if (*ptr <= ' ' || *ptr > '~')
	  ptr = "unknown";
      }

  if (opaque->nb >= NB_DESC_ARRAY) {
      UDBG ((" [%s: search finnished early]\r\n", __FUNCTION__));
      return;
      }

  elem = &opaque->array[opaque->nb];

  elem->disk = opaque->disk;
  elem->partition = opaque->part;
  elem->inode = bootsector_lba;
  elem->is_kernel = 0; /* marker for boot record */
  elem->has_vmlinuz_header = 0;
  elem->filesize = 0; /* marker for boot record */
  for (cpt = 0; cpt < nbof (elem->filename); cpt++) {
      elem->filename[cpt] = *ptr;
      if (*ptr <= ' ' || *ptr > '~')
	  break;
      ptr++;
      }
  elem->filename[cpt] = '\0';

  opaque->nb += 1;
  }}
#endif /* BOOTSECT_PROBE */

SEGMENT_EXTRA(count_and_sort_partition) static inline void
count_and_sort_partition (struct desc_str *desc_array, unsigned nb)
  {
  unsigned i;

  UDBG (("\r\n%s: counting images/MBR & initrd: ", __FUNCTION__));
  boot_system.nb = 0;
  boot_system.nb_initrd = 0;
  for (i = 0; i < nb; i++) {
      if (desc_array[i].filesize != 0 && !desc_array[i].is_kernel)
	  boot_system.nb_initrd++;
	else
	  boot_system.nb++;
      }
  UDBG (("%u, %u\r\n", boot_system.nb, boot_system.nb_initrd));

  UDBG (("%s: Sorting array to have the initrd at the end: ", __FUNCTION__));
  /*
   * We need to keep the order of MBR & kernels, the order of initrd
   * is not important.
   */
  i = 0;
  while (i < boot_system.nb) {
      if (desc_array[i].filesize != 0 && !desc_array[i].is_kernel) {
	  struct desc_str save_initrd = desc_array[i];
	  unsigned j = i;

	  do {
	      desc_array[j] = desc_array[j + 1];
	      } while (++j < nb - 1);
	  desc_array[j] = save_initrd;
	  }
	else
	  i++;
      }
  UDBG (("array sorted.\r\n"));
  }

/*
 * External functions:
 */
GENERATE_VOID_CODE2EXTRA_STUB(disk_analyse, struct gujin_param_attrib attr)

SEGMENT_EXTRA(disk_analyse) void
disk_analyse (struct gujin_param_attrib attr)
  {bound_stack();{
#ifdef UTIL_USE_MALLOC
  struct desc_str desc_array[NB_DESC_ARRAY];
#else
  struct desc_str *desc_array = boot_system.desc;
#endif
  struct treat_directory_str opaque = {
      array: desc_array,
      gujin_attr: attr
      };

#if NB_DISK > 1
  for (opaque.disk = 0; opaque.disk < DI.nbdisk; opaque.disk++) {
#else
  if (DI.nbdisk != 0) {
      opaque.disk = 0;
#endif
      UDBG (("\r\n%s disk %u: ", __FUNCTION__, opaque.disk));
#if DISK_SUPPORT & DOS_SUPPORT
      if (DI.param[opaque.disk].access == dos_part) {
	  UDBG (("add DOSFS "));
	  DOS_analyse (&opaque);
	  UDBG (("add DOSFS done, "));
	  }
	else
#endif /* DOS_SUPPORT */
	  {
#if DISK_SUPPORT & BOOTSECT_PROBE
	  if (!BOOT1_DOS_running()) {
	      /* NOTE: cannot load at 0x7C00 and exit to DOS... */
	      opaque.slashboot_inode = 0;
	      opaque.part = 0xFF;
	      do {
		  if (attr.search_disk_mbr || opaque.part != 0xFF)
		      add_bootsector (&opaque);
		  if (!attr.search_part_mbr)
		      break;
		  for (;;) {
		      unsigned char parttype;

		      opaque.part = (opaque.part + 1) & 0xFF;
		      if (opaque.part >= DI.param[opaque.disk].nbpartition)
			  break;
		      opaque.slashboot_inode = DI.param[opaque.disk]
					.partition[opaque.part].start;
		      if (opaque.slashboot_inode == 0) {
			  /* Already added as the MBR, floppy no-part trick */
			  continue;
			  }
		      parttype = DI.param[opaque.disk]
					.partition[opaque.part].type;
		      if (   parttype != extended
			  && parttype != EMPTY
			  && parttype != Linux_Swap)
			  break;
		      }
		  } while (opaque.part < DI.param[opaque.disk].nbpartition);
	      }
#endif /* BOOTSECT_PROBE */
	  for (opaque.part = 0;
	       opaque.part < DI.param[opaque.disk].nbpartition;
	       opaque.part++) {
	      unsigned char type = DI.param[opaque.disk].partition[opaque.part].type;
#if DISK_SUPPORT & E2FS_PROBE
	      if (type == Linux) {
		  UDBG (("add E2FS "));
		  opaque.slashboot_inode = 0;
		  e2fs_analyse (&opaque);
		  UDBG (("add E2FS done, "));
		  }
		else
#endif /* E2FS_PROBE */
#if DISK_SUPPORT & (FAT12_PROBE|FAT16_PROBE|FAT32_PROBE)
		  if (      0
#if DISK_SUPPORT & FAT12_PROBE
			 || type == DOS_FAT12     || type == HiddenDOS_FAT12
#endif
#if DISK_SUPPORT & FAT16_PROBE
			 || type == DOS_FAT16     || type == HiddenDOS_FAT16
			 || type == BIG_DOS_FAT16 || type == Win95_FAT16LBA
#endif
#if DISK_SUPPORT & FAT32_PROBE
			 || type == HiddenDOS_FAT32
			 || type == Win95_FAT32 || type == Win95_FAT32LBA
#endif
			) {
		  UDBG (("add FATxx "));
		  opaque.slashboot_inode = 0;
		  FAT_analyse (&opaque);
		  UDBG (("add FATxx done, "));
		  }
		else
#endif /* FAT12_PROBE|FAT16_PROBE|FAT32_PROBE */
		  if (type != EMPTY)
		      UDBG (("Do not know how to analyse disk %u, part %u,"
			 " type 0x%X\r\n", opaque.disk, opaque.part, type));
	      }
	  }
      }
  UDBG (("%s end adding.", __FUNCTION__));

  count_and_sort_partition (desc_array, opaque.nb);

#ifdef UTIL_USE_MALLOC
  if (boot_system.desc != 0) {
      free (boot_system.desc);
      boot_system.desc = 0;
      }
  if (opaque.nb != 0) {
      boot_system.desc = malloc (opaque.nb * sizeof (struct desc_str));
      if (boot_system.desc == 0) {
	  boot_system.nb = 0;
	  boot_system.nb_initrd = 0;
	  DDBG (("%s: malloc failed!\r\n", __FUNCTION__));
	  }
	else {
	  while (opaque.nb--)
	      boot_system.desc[opaque.nb] = desc_array[opaque.nb];
	  }
      }
    else {
      boot_system.desc = 0;
      }
#endif

  UDBG (("%s end.\r\n", __FUNCTION__));
  }}

/*
 * start_address is only input.
 * size is input (maximal size) and output (real _uncompressed_ size).
 */
GENERATE_CODE2EXTRA_STUB(system_file_load, unsigned,
	unsigned index, struct system_load_str *loader)

SEGMENT_EXTRA(system_file_load) unsigned
system_file_load (unsigned index, struct system_load_str *loader)
  {bound_stack();{
  struct desc_str *elem = &boot_system.desc[index];
  z_stream gzlib = { mode: INIT, extra: loader }; /* everything else null */
  unsigned returned;
  unsigned char disk = elem->disk;
  unsigned char buffer[DEFAULT_BUFFER_SIZE]; /* _not_ malloced (DOS) */

  /* assert (index < boot_system.nb) */

  loader->compressed_size = 0;

  gzlib.avail_out = gzlib.extra->max_size;
  if (BOOT1_DOS_running() && loader->XMSentry)
      gzlib.next_out = 0;
    else {
#if ASSEMBLY_TYPE == ASSEMBLY_DSES
      gzlib.next_out =  (void *)linear2dsrelative(loader->load_address);
#else
      gzlib.next_out = (void *)loader->load_address;
#endif
      }

#if DISK_SUPPORT & DOS_SUPPORT
  if (DI.param[disk].access == dos_part) {
      returned = DOS_system_file_load (disk, elem->filename,
				  &loader->compressed_size, &gzlib, buffer);
      if (returned)
	  returned += 0x2000;
	else if (elem->filesize != loader->compressed_size) {
	  returned = 0x2E00;
	  UDBG ((" [difference in filesizes: %u instead of %u] ",
			loader->compressed_size, elem->filesize));
	  }
      }
    else
#endif
      {
#if DISK_SUPPORT & (E2FS_PROBE|FAT12_PROBE|FAT16_PROBE|FAT32_PROBE)
      unsigned char part = elem->partition,
		    type = DI.param[disk].partition[part].type;

#if DISK_SUPPORT & E2FS_PROBE
      if (type == Linux) {
	  returned = e2fs_system_file_load (disk, part, elem->inode,
				&loader->compressed_size, &gzlib, buffer);

	  if (returned)
	      returned += 0x3000;
#ifdef NEW_E2FS_READ
	    else if (   elem->filesize > loader->compressed_size
		     || elem->filesize < loader->compressed_size
						- EXT2_MAX_BLOCK_SIZE) {
	      returned = 0x3E00;
	      UDBG ((" [difference in filesizes: %u instead of %u] ",
			loader->compressed_size, elem->filesize));
	      }
	    else {
	      UDBG ((" [E2FS info: blocksize is > %u] ",
			loader->compressed_size - elem->filesize));
	      /* correct it with uncompressed size: */
	      loader->compressed_size = elem->filesize;
	      }
#else /* NEW_E2FS_READ */
	    else if (elem->filesize != loader->compressed_size) {
	      returned = 0x3E00;
	      UDBG ((" [difference in filesizes: %u instead of %u] ",
			loader->compressed_size, elem->filesize));
	      }
#endif /* NEW_E2FS_READ */
	  }
	else
#endif /* E2FS_PROBE */
#if DISK_SUPPORT & (FAT12_PROBE|FAT16_PROBE|FAT32_PROBE)
	  if (     0
#if DISK_SUPPORT & FAT12_PROBE
		|| type == DOS_FAT12     || type == HiddenDOS_FAT12
#endif
#if DISK_SUPPORT & FAT16_PROBE
		|| type == DOS_FAT16     || type == HiddenDOS_FAT16
		|| type == BIG_DOS_FAT16 || type == Win95_FAT16LBA
#endif
#if DISK_SUPPORT & FAT32_PROBE
		|| type == DOS_FAT32 || type == HiddenDOS_FAT32
		|| type == Win95_FAT32 || type == Win95_FAT32LBA
#endif
		) {
	  returned = FAT_system_file_load (disk, part, elem->inode,
			&loader->compressed_size, &gzlib, buffer);
	  /* here elem->filesize is the real file size,
		filesize is the size rounded up to cluster */
	  if (returned)
	      returned += 0x4000;
	    else if (elem->filesize > loader->compressed_size) {
	      returned = 0x4E00;
	      UDBG ((" [difference in filesizes: %u instead of %u] ",
			loader->compressed_size, elem->filesize));
	      }
	    else {
	      UDBG ((" [FAT info: cluster size is > %u] ",
			loader->compressed_size - elem->filesize));
	      /* correct it with uncompressed size: */
	      loader->compressed_size = elem->filesize;
	      }
	  }
	else
#endif /* FAT12_PROBE|FAT16_PROBE|FAT32_PROBE */
	  {
	  returned = 0x5000;
	  UDBG ((" [%s: Do not know partition type 0x%X] ",
			__FUNCTION__, type));
	  }
#else /* E2FS_PROBE|FAT12_PROBE|FAT16_PROBE|FAT32_PROBE */
      returned = 0x5000;
      UDBG (("Only DOS_SUPPORT compiled in, cannot load!\r\n"));
#endif /* E2FS_PROBE|FAT12_PROBE|FAT16_PROBE|FAT32_PROBE */
      }
  loader->uncompressed_size = gzlib.total_out;

  return returned;
  }}

/*
 * The main structure of this file in .BSS
 */
struct boot_system_str boot_system;

