/* vmlinuz.h */
#ifndef VMLINUZ_H
#define VMLINUZ_H

/* Gujin is a bootloader, it loads a Linux kernel from cold boot or DOS.
 * Copyright (C) 1999-2001 Etienne Lorrain, fingerprint (2D3AF3EA):
 *   2471 DF64 9DEE 41D4 C8DB 9667 E448 FF8C 2D3A F3EA
 * E-Mail: etienne.lorrain@masroudeau.com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

/*
 * Description of vmlinuz files, and of memory parameters
 * passed to kernel (they are similar):
 *  (see linux/Documentation/i386/zero-page.txt)
 */
struct screen_info { /* stolen from linux/include/linux/tty.h */
    unsigned char  orig_x;
    unsigned char  orig_y;
    unsigned short EXT_MEM_K;		/* general info */
    unsigned char  undefined;
    unsigned char  orig_video_page;
    unsigned char  orig_video_mode;
    unsigned char  orig_video_cols;
    unsigned short unused2;
    unsigned short orig_video_ega_bx;
    unsigned short unused3;
    unsigned char  orig_video_lines;
    unsigned char  orig_video_isVGA;
    unsigned short orig_font_heigh;
    /* VESA graphic mode -- linear frame buffer */
    unsigned short lfb_width;
    unsigned short lfb_height;
    unsigned short lfb_depth;
    unsigned       lfb_base;
    unsigned       lfb_size;
    unsigned short CL_MAGIC;	/* 0xA33F */	/* general info */
    unsigned short CL_OFFSET;			/* general info */
    unsigned short lfb_linelength;
    unsigned char  red_size;
    unsigned char  red_pos;
    unsigned char  green_size;
    unsigned char  green_pos;
    unsigned char  blue_size;
    unsigned char  blue_pos;
    unsigned char  rsvd_size;
    unsigned char  rsvd_pos;
    unsigned short vesapm_seg;
    unsigned short vesapm_off;
    unsigned short lfb_pages;
    /* 0x34 -- 0x3f reserved for future expansion */
    unsigned char reserved[12];
    } __attribute__ ((packed));

struct apm_bios_info {
    unsigned short   version;
    unsigned short   cseg;
    unsigned         offset;
    unsigned short   cseg_16;
    unsigned short   dseg;
    unsigned short   flags;
    unsigned short   cseg_len;
    unsigned short   unused; /* cseg_16_len */
    unsigned short   dseg_len;
    } __attribute__ ((packed));

/* See probe_cmos_for_drives() in linux/drivers/ide/ide-geometry.c */
/* rarely used, also check for CMOS address 0x12 (int 0x41, 0x46) */
struct drive_info {
    unsigned short cyl;
    unsigned char  head;
    unsigned short start_reduce_write_current;	/* PC XT only */
    unsigned short start_write_precomp;
    unsigned char  max_ECC_data;		/* PC XT only */
    unsigned char  ctl;				/* different for XT and AT */
    unsigned char  standart_timeout;		/* PC XT only */
    unsigned char  formating_timeout;		/* PC XT only */
    unsigned char  check_drive_timeout;		/* PC XT only */
    unsigned short landing_zone_cylinder;	/* PC AT+ */
    unsigned char  sect;			/* PC AT+ */
    unsigned char  reserved;
    } __attribute__ ((packed));

struct sys_desc_table_info {
    unsigned short length;
    unsigned char table[14];
    } __attribute__ ((packed));

struct in_out_info {
    unsigned char  setup_sects; /* default 4 */
    struct root_flags_str {
	unsigned short read_only : 1;
	unsigned short unused    : 15;
	} root_flags;
    unsigned short kernel_compressed_size; /* 16 bytes unit, rounded up */
    unsigned short swap_dev;    /* 0 */
    struct ramdisk_str {
	unsigned short	size   : 11;
	unsigned short	unused : 3;
	unsigned short	load   : 1;
	unsigned short	prompt : 1;
	} __attribute__ ((packed)) ramdisk;	/* RAMDISK, default 0 */
    unsigned short vid_mode;    /* SVGA_MODE, default ASK_VGA = 0xFFFD */
    union {
	struct {
	    unsigned char  minor_root;  /* ROOT_DEV */
	    unsigned char  major_root;  /* ROOT_DEV */
	    } __attribute__ ((packed)) fields;
	unsigned short minor_major;
	} __attribute__ ((packed)) root;
    unsigned char  AUX_DEVICE_INFO; /* ignored */
    unsigned char  jmp_trampoline[3];
    union header_signature_union {
	unsigned char  bytes[4];   /* "HdrS" */
	unsigned       longword;
	} __attribute__ ((packed)) header_signature;
    unsigned short header_version_number; /* 0x0202 */
    unsigned short realmode_swtch[2]; /* hook to plug in a A20-open func */
    unsigned short start_sys_seg;  /* SYSSEG */
    unsigned short kernel_version; /* char pointer */
    struct {
	enum { LILO = 0, Loadlin, bootsect,
	       SYSLX, ETHERBOOT, GUJIN } type : 4;
	unsigned char version : 4;
	} __attribute__ ((packed)) type_of_loader;
    struct {
	unsigned char LOADED_HIGH  : 1;
	unsigned char unused       : 6;
	unsigned char CAN_USE_HEAP : 1; /* heap_end_ptr is set */
	} __attribute__ ((packed)) loadflags;
    unsigned short setup_move_size; /* 0x8000 */
    unsigned       code32_start;
    unsigned       ramdisk_image;
    unsigned       ramdisk_size; /* in bytes */
    unsigned short bootsect_kludge[2];
    unsigned short heap_end_ptr; /* modelist+1024, version >= 0x0201 */
    unsigned short pad1;
    unsigned       cmd_line_ptr; /* default 0x90800, version >= 0x0202 */
    } __attribute__ ((packed));

struct e820map_info {
    long long addr;	/* start of memory segment */
    long long size;	/* size of memory segment */
    long type;		/* type of memory segment */
    } __attribute__ ((packed));

struct linux_param {
    struct screen_info		screen_info;
    struct apm_bios_info	apm_bios_info;
    unsigned char		free1[44];
    struct drive_info		drive1, drive2;
    struct sys_desc_table_info	sys_desc_table;
    unsigned char		free2[304];
    unsigned			ALT_MEM_K;
    unsigned char		free3[4];
    unsigned char		nb_E820_entries;
    unsigned char		free4[8];
    /* 512 bytes bootsector limit */
    /* "in_out" is mostly init'ed in vmlinuz file: */
    struct in_out_info		in_out;
    unsigned char		free5[164];
    struct e820map_info		e820map[32];
    unsigned char		free6[688];
    unsigned char		command_line[2*1024];
    } __attribute__ ((packed));

extern inline void test_linux_param_struct (void)
  {
  if (offsetof(struct linux_param, screen_info.EXT_MEM_K) != 0x02)
      __ERROR();
  if (offsetof(struct linux_param, apm_bios_info) != 0x40)
      __ERROR();
  if (offsetof(struct linux_param, drive1) != 0x80)
      __ERROR();
  if (offsetof(struct linux_param, drive2) != 0x80 + 16)
      __ERROR();
  if (offsetof(struct linux_param, sys_desc_table) != 0xA0)
      __ERROR();
  if (offsetof(struct linux_param, ALT_MEM_K) != 0x1E0)
      __ERROR();
  if (offsetof(struct linux_param, nb_E820_entries) != 0x1E8)
      __ERROR();
  if (offsetof(struct linux_param, in_out.setup_sects) != 0x1F1)
      __ERROR();
  if (offsetof(struct linux_param, in_out.root_flags) != 0x1F2)
      __ERROR();
  if (offsetof(struct linux_param, in_out.kernel_compressed_size) != 0x1F4)
      __ERROR();
  if (offsetof(struct linux_param, in_out.swap_dev) != 0x1F6)
      __ERROR();
  if (offsetof(struct linux_param, in_out.header_version_number) != 0x206)
      __ERROR();
  if (offsetof(struct linux_param, in_out.code32_start) != 0x214)
      __ERROR();
  if (offsetof(struct linux_param, in_out.heap_end_ptr) != 0x224)
      __ERROR();
  if (offsetof(struct linux_param, e820map) != 0x2D0)
      __ERROR();
  if (offsetof(struct linux_param, command_line) != 0x800)
      __ERROR();
  }

/*
 * The structure exchanged between util.c::system_file_load()
 * and vmlinuz.c::menu_load_system()
 */
struct system_load_str {
	/* input: */
    unsigned load_address, max_size;
    farptr   XMSentry;
    unsigned short handle;
    unsigned short KBalloced;
	/* output: */
    unsigned compressed_size, uncompressed_size;
    unsigned short minor_major_root;
    unsigned short license_compliant;
	/* Initialise from the "comment" line of the GZIP file */
    union {
	unsigned array[7];
	struct {
	    unsigned late_relocation_address;
	    unsigned runadr;
	    unsigned maskcpu;
#define MASKCPU_8086	(1 << 0)
#define MASKCPU_8186	(1 << 1)
#define MASKCPU_8286	(1 << 2)
#define MASKCPU_386	(1 << 3) /* if bit 3 set, cannot run on 386 */
#define MASKCPU_486	(1 << 4) /* most 486s, AMD 5x86, Cyrix 5x86 */
#define MASKCPU_586	(1 << 5) /* Intel P5, P54C, P55C, P24T,
					NextGen Nx586, Cyrix M1, AMD K5, K6,
					Centaur C6, C2, C3, Rise mP6 */
#define MASKCPU_686	(1 << 6) /* Intel P6, P2, P3, AMD K7,
					Cyrix M2, VIA Cyrix III */
#define MASKCPU_786	(1 << 7) /* extended family 0xF, Intel P4 only */

#define MASKCPU_REALMODE (1 << 8) /* start in realmode instead */
#define MASKCPU_IA64MODE (1 << 9) /* start in IA64 instead */
#define MASKCPU_AA64MODE (1 << 10) /* start in AA64 (AMD Long Mode) instead */

	    unsigned maskflag;	/* request those bits set, cpuid 0x00000001 */
	    unsigned maskextra;	/* request those bits set, cpuid 0x80000001 */
	    unsigned maskvesa;     /* if bit (8-1) set, VESA 8BPP supported */
#define MASKVESA_1BPP	(1 << 0)
#define MASKVESA_2BPP	(1 << 1)
#define MASKVESA_4BPP	(1 << 3)
#define MASKVESA_8BPP	(1 << 7)
#define MASKVESA_15BPP	(1 << 14)
#define MASKVESA_16BPP	(1 << 15)
#define MASKVESA_24BPP	(1 << 23)
#define MASKVESA_32BPP	(1 << 31)
#define MASKVESA_TEXT	(1 << 16)	/* able to boot in text mode */
#define MASKVESA_WIN	(1 << 17)	/* able to boot in VESA1 mode */
#define MASKVESA_LIN	(1 << 18)	/* able to boot in VESA2 linear mode */
	    unsigned maskresolution;	/* bit set => exclude those modes */
#define MASKRESOLUTION_40cols		(1 << 0)
#define MASKRESOLUTION_80cols		(1 << 1)
#define MASKRESOLUTION_100cols		(1 << 2)
#define MASKRESOLUTION_132cols		(1 << 3)
#define MASKRESOLUTION_160cols		(1 << 4)

#define MASKRESOLUTION_24rows		(1 << 5)	/* mostly VT/serial */
#define MASKRESOLUTION_25rows		(1 << 6)
#define MASKRESOLUTION_30rows		(1 << 7)
#define MASKRESOLUTION_36rows		(1 << 8)	/* mostly VT/serial */
#define MASKRESOLUTION_43rows		(1 << 9)
#define MASKRESOLUTION_44rows		(1 << 10)
#define MASKRESOLUTION_48rows		(1 << 11)	/* mostly VT/serial */
#define MASKRESOLUTION_50rows		(1 << 12)
#define MASKRESOLUTION_60rows		(1 << 13)
#define MASKRESOLUTION_64rows		(1 << 14)

#define MASKRESOLUTION_320x200		(1 << 15)
#define MASKRESOLUTION_320x400		(1 << 16)
#define MASKRESOLUTION_640x200		(1 << 17)
#define MASKRESOLUTION_640x350		(1 << 18)
#define MASKRESOLUTION_640x400		(1 << 19)
#define MASKRESOLUTION_640x480		(1 << 20)
#define MASKRESOLUTION_720x400		(1 << 21)
#define MASKRESOLUTION_768x1024		(1 << 22)
#define MASKRESOLUTION_800x600		(1 << 23)
#define MASKRESOLUTION_832x624		(1 << 24)
#define MASKRESOLUTION_1024x768		(1 << 25)
#define MASKRESOLUTION_1152x864		(1 << 26)
#define MASKRESOLUTION_1280x960		(1 << 27)
#define MASKRESOLUTION_1280x1024	(1 << 28)
#define MASKRESOLUTION_1600x1200	(1 << 29)
#define MASKRESOLUTION_2048x1540	(1 << 30)
#define MASKRESOLUTION_UNLISTED		(1 << 31)	/* not listed invalid */
	    } elem;
	} comment;
	/* related to LINUX version: */
    unsigned params_farptr;
    char     cmdline_extraparam[128];
#if defined (LINUZNAME1) || defined (LINUZNAME2) || defined (LINUZNAME3)
	/* saved temporary: */
    int	treat_start_bytes; /* vmlinuz header skipping */
    void *gzip_start;
#endif
    };

/*
 * The structure given to GPL compliant kernel:
 */
struct gpl_compliant_str {
    unsigned signature;
    unsigned version;
    unsigned feature;

    setjmp_buf jmpbuf;
    unsigned initrd_index;
    struct system_load_str *kloader, *iloader;
    /* those are linear address: */
    unsigned instboot, UI, DI, boot_system;
    };

#if (SETUP & EXTRA_SEGMENT) && !defined (CALLING_FROM_EXTRASEG)
#define ASMNAME(x)	asm (STRING(codeseg_ ## x))
#else
#define ASMNAME(x)	/* call in same segment */
#endif

unsigned initrd_file (unsigned index) ASMNAME(initrd_file);

void menu_load_system (unsigned system_index,
			unsigned extendmem,
			unsigned freemem,
			farptr   extra_cmdline,
			farptr   instboot_info,
			unsigned short DefaultTextVideoMode,
			struct gujin_param_attrib gujin_attr)
		ASMNAME(menu_load_system);

void treat_gzip_comment (struct system_load_str *loader, const char *ptr)
		ASMNAME(treat_gzip_comment);


void treat_gzip_name (struct system_load_str *loader, const char *ptr)
		ASMNAME(treat_gzip_name);

#if defined (LINUZNAME1) || defined (LINUZNAME2) || defined (LINUZNAME3)
unsigned vmlinuz_header_treat (struct system_load_str *loader,
				unsigned char *buffer,
				unsigned buflen)
				ASMNAME(vmlinuz_header_treat);
#endif

#undef ASMNAME

/*
 * For gzlib.c:
 */
#define GZLIB_TREAT_NAME(loader, ptr) { \
    void treat_gzip_name (struct system_load_str *, const char *);	\
    treat_gzip_name (loader, ptr);					\
    }

#define GZLIB_TREAT_COMMENT(loader, ptr) { \
    void treat_gzip_comment (struct system_load_str *, const char *);	\
    treat_gzip_comment (loader, ptr);					\
    }

#endif /* VMLINUZ_H */

